﻿# -*- coding: utf-8 -*-

from brownie import *
import ui
import resource

#=============================================================================

class Item(object):
    __slots__ = ('name', 'execute', 'children', 'icon', 'menu')
    def __init__(self, name = None, execute = None, children = None, icon = None, menu = None):
        self.name = name
        self.execute = execute
        self.children = children
        self.icon = icon
        self.menu = menu

#=============================================================================

class PathItem(Path):
    __slots__ = ('name', 'icon', 'children')
    def __init__(self, path, name = None):
        super(PathItem, self).__init__(path)
        self.name = name if name != None else self.basename
        self.icon =  super(PathItem, self).icon
        if self.has_subfolder:
            self.children = self.get_subfolders

    def execute(self):
        resolved = resource.resolve(self)
        if resolved.browsable > 1:
            ui.form.open(resolved)
        else:
            super(PathItem, self).execute()

    def get_basename(self):
        return super(PathItem, self).name
    basename = property(get_basename)

    def get_subfolders(self):
        folders = self.subfolders
        if folders:
            for i in folders:
                yield PathItem(i)

    def get_children(self):
        for i in super(PathItem, self).children:
            yield PathItem(i)

#=============================================================================

class TabItem(object):
    __slots__ = ('target')
    def __init__(self, target):
        self.target = target

    def execute(self):
        if modifiers() & MBUTTON:
            self.target.close()
        else:
            self.target.activate()

    def get_name(self):
        return self.target.path.name
    name = property(get_name)

#=============================================================================

class Trash(object):
    __slots__ = ('icon')
    def __init__(self):
        self.icon = Path(FOLDER_TRASH).icon

    def execute(self):
        trash.clear()

    def get_name(self):
        return 'ごみ箱を空にする\t' + trash.format
    name = property(get_name)

    def get_enabled(self):
        return trash.size > 0
    enabled = property(get_enabled)

    def get_menu(self):
        return (PathItem(FOLDER_TRASH, 'ごみ箱を開く'),)
    menu = property(get_menu)

#=============================================================================

class FavoriteItem(object):
    __slots__ = ('name', 'icon')
    def __init__(self, name, icon = None):
        self.name = name
        self.icon = icon

    def get_children(self):
        yield Item('お気に入りに追加(*A)', self.add)
        yield Item('お気に入りの整理(*E)', lambda: ui.form.open(config.PATH_FAVORITE))
        yield None # separator
        favorite = Path(config.PATH_FAVORITE)
        for i in favorite.children:
            n = PathItem(i)
            if n.browsable:
                n.children = n.get_children
            yield n
    children = property(get_children)

    def add(self):
        target = ui.current()
        if target:
            path = target.path
            path.symlink('%s\\%s' % (config.PATH_FAVORITE, path.name))
            ui.form.notify('"%s" をお気に入りに追加しました。' % path, INFO)
        else:
            ui.form.notify('アクティブなフォルダがありません。', WARNING)

#=============================================================================

class Current(object):
    __slots__ = ('name', 'execute', 'children', '__enabled', '__checked', 'icon', 'menu')
    def __init__(self, name, execute = None, enabled = None, checked = None, children = None, icon = None, menu = None):
        self.name = name
        if execute:
            def __execute():
                target = ui.current()
                if target:
                    execute(target)
            self.execute = __execute
        if children:
            def __children():
                target = ui.current()
                if target:
                    return children(target)
                else:
                    return []
            self.children = __children
        self.__enabled = enabled
        self.__checked = checked
        self.icon = icon
        self.menu = menu

    def get_enabled(self):
        target = ui.current()
        if not target:
            return False
        elif self.__enabled:
            return self.__enabled(target)
        else:
            return True
    enabled = property(get_enabled)

    def get_checked(self):
        if self.__checked:
            target = ui.current()
            if target:
                return self.__checked(target)
        return False
    checked = property(get_checked)

#=============================================================================

class Grouping(Current):
    def __init__(self, name, icon = None):
        def execute(v):
            v.grouping = not v.grouping
        def checked(v):
            return v.grouping
        super(Grouping, self).__init__(name, execute, checked = checked, icon = icon)

#=============================================================================

class Go(Current):
    def __init__(self, name, destination, *args, **kwds):
        def execute(v):
            v.open(destination)
        def enabled(v):
            return v.parse(destination)
        super(Go, self).__init__(name, execute, enabled, *args, **kwds)

#=============================================================================

class GoUp(Current):
    def __init__(self, name, icon = None):
        def execute(v):
            v.open(FOLDER_PARENT)
        def enabled(v):
            return not v.path.has_parent
        def children(v):
            folders = []
            path = v.parse(FOLDER_PARENT)
            while path:
                n = PathItem(path)
                delattr(n, 'children')
                folders.append(n)
                path = path.parent
            return folders
        super(GoUp, self).__init__(name, execute, enabled, children = children, icon = icon)

#=============================================================================

class GoHistory(object):
    __slots__ = ('target', 'name', 'destination', 'icon')
    def __init__(self, target, name, destination, icon):
        self.target = target
        self.name = name
        self.destination = destination
        self.icon = icon

    def execute(self):
        self.target.open(self.destination)

#=============================================================================

class GoBack(Go):
    def __init__(self, name, icon = None):
        def children(v):
            folders = []
            for i in v.history(FOLDER_BACK):
                folders.append(GoHistory(v, i.name, -len(folders) - 1, i.icon))
            return folders
        super(GoBack, self).__init__(name, FOLDER_BACK, children = children, icon = icon)

#=============================================================================

class GoForward(Go):
    def __init__(self, name, icon = None):
        def children(v):
            folders = []
            for i in v.history(FOLDER_FORWARD):
                folders.append(GoHistory(v, i.name, len(folders) + 1, i.icon))
            return folders
        super(GoForward, self).__init__(name, FOLDER_FORWARD, children = children, icon = icon)

#=============================================================================

class Close(Current):
    def __init__(self, name, icon = None):
        super(Close, self).__init__(name, lambda v: v.close(), icon = icon)

#=============================================================================

class Mode(Current):
    def __init__(self, name, icon = None):
        super(Mode, self).__init__(name, children = self.modelist, icon = icon)

    @staticmethod
    def modelist(view):
        if config.WINDOWS_VERSION < WINDOWS_VISTA:
            r = [
                Item('縮小版(*H)'    , lambda: view.mode(ICON,  256), icon=32),
                Item('並べて表示(*S)', lambda: view.mode(TILE,   48), icon=33),
                Item('アイコン(*N)'  , lambda: view.mode(ICON,   32), icon=34),
                Item('一覧(*L)'      , lambda: view.mode(ICON,   16), icon=35),
                Item('詳細(*D)'      , lambda: view.mode(DETAIL, 16), icon=36),
            ]
        else:
            r = [
                Item('Icon 256'  , lambda: view.mode(ICON, 256), icon=32),
                Item('Icon 96'   , lambda: view.mode(ICON,  96), icon=32),
                Item('Icon 48'   , lambda: view.mode(ICON,  48), icon=32),
                Item('Icon 32'   , lambda: view.mode(ICON,  32), icon=34),
                Item('Icon 16'   , lambda: view.mode(ICON,  16), icon=35),
                Item('Details 96', lambda: view.mode(DETAIL, 96), icon=36),
                Item('Details 16', lambda: view.mode(DETAIL, 16), icon=36),
                Item('Tile 96'   , lambda: view.mode(TILE, 96), icon=33),
                Item('Tile 48'   , lambda: view.mode(TILE, 48), icon=33),
            ]
        return r

#=============================================================================

class ConfigBool(object):
    def __init__(self, name):
        self.name = name

    def execute(self):
        var = self.__var()
        v = getattr(config, var)
        setattr(config, var, not v)

    def get_checked(self):
        return getattr(config, self.__var())
    checked = property(get_checked)

    def __var(self):
        var = str(type(self))
        return var[var.rfind('.')+1:var.rfind("'")]

#=============================================================================

class LISTVIEW_LOOP(ConfigBool):
    pass

#=============================================================================

class LISTVIEW_WHEELSELECT(ConfigBool):
    pass

#=============================================================================

class RENAME_EXTENSION(ConfigBool):
    pass

#=============================================================================

LISTVIEW_AUTOADJUST_DEFAULT = config.LISTVIEW_AUTOADJUST

class LISTVIEW_AUTOADJUST(object):
    def __init__(self, name):
        self.name = name
    def execute(self):
        if config.LISTVIEW_AUTOADJUST == -1:
            config.LISTVIEW_AUTOADJUST = LISTVIEW_AUTOADJUST_DEFAULT
        else:
            config.LISTVIEW_AUTOADJUST = -1
    def get_checked(self):
        return config.LISTVIEW_AUTOADJUST != -1
    checked = property(get_checked)

#=============================================================================

LISTVIEW_BACKGROUND_DEFAULT = config.LISTVIEW_BACKGROUND

class LISTVIEW_BACKGROUND(object):
    def __init__(self, name):
        self.name = name
    def execute(self):
        def adjust_color(c):
            if c < 128:
                return 255 - ((255 - c) * 15 / 16)
            else:
                return c * 15 / 16
        if config.LISTVIEW_BACKGROUND == LISTVIEW_BACKGROUND_DEFAULT:
            config.LISTVIEW_BACKGROUND = [adjust_color(i) for i in config.LISTVIEW_BACKGROUND]
        else:
            config.LISTVIEW_BACKGROUND = LISTVIEW_BACKGROUND_DEFAULT
    def get_checked(self):
        return config.LISTVIEW_BACKGROUND != LISTVIEW_BACKGROUND_DEFAULT
    checked = property(get_checked)

#=============================================================================

class LISTVIEW_AUTOUNIT(ConfigBool):
    def get_enabled(self):
        return config.LISTVIEW_BACKGROUND != LISTVIEW_BACKGROUND_DEFAULT
    enabled = property(get_enabled)

#=============================================================================
