// abstractbar.cpp

#include "stdafx.h"
#include "abstractbar.hpp"
#include "math.hpp"
#include "node.hpp"
#include "path.hpp"

using namespace brownie;

//==================================================================================================================================================================

void AbstractBar::__init__(Tuple* args, Dict* kw, DIRECTION dock, DWORD style, DWORD styleEx)
{
	m_pushed = m_hot = -1;
	set_items(null);
	__super::__init__(args, kw, dock, style, styleEx | WS_EX_COMPOSITED);
	DisableIME();
	m_drop.Register(this);
}

void AbstractBar::__init__(Tuple* args, Dict* kw)
{
	abstract(typeof<AbstractBar>(), self);
	__init__(args, kw, NORTH, 0, 0);
}

void AbstractBar::set_Dock(DIRECTION value)
{
	__super::set_Dock(value);
	Update(true);
}

SIZE AbstractBar::get_DefaultSize() const throw()
{
	SIZE	sz = __super::get_DefaultSize();
	ssize_t count = GetItemCount();
	int		ln = GetItem(count)->offset;
	if (HFONT font = GetFont())
		ln = math::max(ln, FontHeight(font));
	switch (Dock)
	{
	case NORTH:
	case SOUTH:
		sz.cx = ln;
		break;
	case EAST:
	case WEST:
		sz.cy = ln;
		break;
	}
	return sz;
}

Icons* AbstractBar::GetIcons() const throw()
{
	return cast<Icons>(m_icons);
}

void AbstractBar::SetIcons(Icons* value) throw()
{
	if (m_icons == value)
		return;
	m_icons = value;
	Invalidate();
}

AbstractBar::Item* AbstractBar::GetItem(ssize_t index) throw()
{
	if (index < 0 || index >= (ssize_t)m_items.size())
		return null;
	return &m_items[index];
}

AbstractBar::Item* AbstractBar::GetItems() throw()
{
	return &m_items[0];
}

bool AbstractBar::GetItemRect(ssize_t index, RECT* rc) const throw()
{
	ASSERT(rc);
	if (index < 0 || GetItemCount() <= index)
		return false;
	SetRect(rc, m_items[index].offset, 0, m_items[index+1].offset, this->ClientSize.cy);
	return true;
}

void AbstractBar::SetPushedItem(ssize_t index) throw()
{
	if (m_pushed == index)
		return;
	RECT rc;
	if (m_pushed >= 0)
	{
		GetItemRect(m_pushed, &rc);
		Invalidate(rc);
	}
	m_pushed = index;
	if (m_pushed >= 0)
	{
		GetItemRect(m_pushed, &rc);
		Invalidate(rc);
	}
}

void AbstractBar::SetHotItem(ssize_t index) throw()
{
	if (m_hot == index)
		return;
	RECT rc;
	if (m_hot >= 0)
	{
		GetItemRect(m_hot, &rc);
		Invalidate(rc);
	}
	const Item* item = GetItem(index);
	if (item && item->enabled)
	{
		m_hot = index;
		GetItemRect(m_hot, &rc);
		Invalidate(rc);
	}
	else
		m_hot = -1;
}

ssize_t AbstractBar::HitTest(int x, int y) const throw()
{
	SIZE sz = ClientSize;
	if (y < 0 || sz.cy <= y)
		return -1;
	ssize_t count = GetItemCount();
	for (ssize_t i = 0; i < count; ++i)
	{
		if (m_items[i].offset <= x && x < m_items[i+1].offset)
			return i;
	}
	return -1;
}

void AbstractBar::GetStatus(int* mode, STATUS* status, ssize_t index, const int modes[]) const throw()
{
	// 0:NORMAL, 1:HOT, 2:PRESSED, 3:DISABLED, 4:CHECKED, 5:HOTCHECKED
	if (index == m_pushed)
	{
		*status = SELECTED;
		*mode = modes[2];
	}
	else if (index == m_hot)
	{
		*status = HOT;
		*mode = (m_items[index].checked ? modes[5] : modes[1]);
	}
	else if (m_items[index].enabled)
	{
		*status = NORMAL;
		*mode = (m_items[index].checked ? modes[4] : modes[0]);
	}
	else
	{
		*status = DISABLED;
		*mode = modes[3];
	}
}

LRESULT AbstractBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_PAINT:
	{
		PAINTSTRUCT ps;
		HDC dc = BeginPaint(m_hwnd, &ps);
		OnDraw(dc, ps.rcPaint);
		EndPaint(m_hwnd, &ps);
		return 0;
	}
	case WM_PRINTCLIENT:
		OnDraw((HDC)wParam, ClientArea);
		return 0;
	case WM_MOUSELEAVE:
		SetHotItem(-1);
		return 0;
	case WM_MOUSEMOVE:
		switch (GET_KEYSTATE_WPARAM(wParam) & MK_BUTTONS)
		{
		case 0:
			TrackMouseEvent(TME_LEAVE, 0);
			SetHotItem(HitTest(GET_XY_LPARAM(lParam)));
			break;
		case MK_LBUTTON:
			// {^ォ̃hbOŃ|bvAbv
			DropDown(m_hot, POINT_FROM_LPARAM(lParam));
			break;
		}
		return 0;
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
		if ((GET_KEYSTATE_WPARAM(wParam) & MK_BUTTONS) == MSG2MK(msg, wParam))
		{	// {^ЂƂĂ
			HRESULT hr = E_FAIL;
			ssize_t index = HitTest(GET_XY_LPARAM(lParam));
			if (const Item* item = GetItem(index))
				if (!item->executable && index == GetHotItem())
					hr = DropDown(index);
			if FAILED(hr)
			{
				SetPushedItem(index);
				SetCapture();
			}
		}
		return 0;
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
		OnMouseUp(MSG2VK(msg, wParam), wParam, lParam);
		return 0;
	case WM_MOUSEWHEEL:
		DropDown(HitTest(ScreenToClient(POINT_FROM_LPARAM(lParam))));
		return 0;
	case WM_KILLFOCUS:
		SetHotItem(-1);
		break;
	case WM_GETFONT:
		return (LRESULT)m_font.Handle;
	case WM_SETFONT:
		m_font = (HFONT)wParam;
		if (LOWORD(lParam))
			Invalidate();
		return 0;
	case WM_DESTROY:
		m_drop.Revoke();
		{
			py::Lock lock;
			m_items.clear(); // TODO: zGCɑΉׂB
		}
		break;
	case WM_CREATE:
		m_font = GetStockFont(DEFAULT_GUI_FONT);
		break;
	case WM_SETTINGCHANGE:
		m_font = GetStockFont(DEFAULT_GUI_FONT);
		OnLayout();
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

void AbstractBar::OnUpdate()
{
	py::Lock lock;
	try
	{
		ssize_t count = GetItemCount();
		for (ssize_t i = 0; i < count; ++i)
		{
			Item&	item = m_items[i];
			bool	modified = false;
			bool	enabled = node_enabled(item.node);
			bool	checked = node_checked(item.node);
			if (item.enabled != enabled)
			{
				item.enabled = enabled;
				modified = true;
			}
			if (item.checked != checked)
			{
				item.checked = checked;
				modified = true;
			}
			if (modified)
			{
				RECT rc;
				GetItemRect(i, &rc);
				Invalidate(rc);
			}
		}
	}
	catch (Exception&)
	{
		Exception::clear();
	}
	__super::OnUpdate();
}

bool AbstractBar::OnGesture(const MSG& msg)
{
	UINT vk = MSG2VK(msg);
	if (vk == VK_LBUTTON || vk == VK_MBUTTON)
	{
		POINT pt = { GET_XY_LPARAM(msg.lParam) };
		if (msg.hwnd != m_hwnd)
			MapWindowPoints(msg.hwnd, m_hwnd, &pt, 1);
		if (HitTest(pt) >= 0)
			return false;
	}
	return __super::OnGesture(msg);
}

void AbstractBar::OnMouseUp(UINT vk, WPARAM wParam, LPARAM lParam)
{
	ssize_t index = HitTest(GET_XY_LPARAM(lParam));
	if (index >= 0 && index == m_pushed)
	{
		if (PyObject* o = m_items[index].node)
			NodeExecute(o, vk);
	}
	SetPushedItem(-1);
	if ((GET_KEYSTATE_WPARAM(wParam) & MK_BUTTONS) == 0)
		ReleaseCapture();
}

void AbstractBar::OnDraw(HDC dc, const RECT& rcPaint)
{
}

HRESULT AbstractBar::DropDown(ssize_t index) throw()
{
	Item* item = GetItem(index);
	if (!item || !item->node)
		return E_FAIL;

	ReleaseCapture();
	SetHotItem(-1);
	SetPushedItem(index);
	HRESULT hr = NodePopupChildren(this, item->node, (int)index);
	SetPushedItem(-1);
	return hr;
}

// pt : J[\ʒuiNCAgWj
HRESULT AbstractBar::DropDown(ssize_t index, POINT pt) throw()
{
	Item* item = GetItem(index);
	if (!item || !item->node)
		return E_FAIL;
	RECT rc;
	GetItemRect(m_hot, &rc);
	bool popup = false;
	switch (this->Dock)
	{
	case NORTH:
		popup = (rc.left <= pt.x && pt.x < rc.right && rc.bottom <= pt.y);
		break;
	case SOUTH:
		popup = (rc.left <= pt.x && pt.x < rc.right && pt.y < rc.top);
		break;
	case WEST:
		popup = (rc.top <= pt.y && pt.y < rc.bottom && rc.right <= pt.x);
		break;
	case EAST:
		popup = (rc.top <= pt.y && pt.y < rc.bottom && pt.x < rc.left);
		break;
	}
	if (!popup)
		return E_FAIL;

	return DropDown(m_hot);
}

namespace
{
	class DropTargetWrapper : public Unknown< implements<IDropTarget> >
	{
	private:
		ref<IDropTarget>	m_target;
		AbstractBar*		m_owner;
		ssize_t				m_index;

	public:
		DropTargetWrapper(IDropTarget* target, AbstractBar* owner, ssize_t index) : 
			m_target(target), m_owner(owner), m_index(index)
		{ }
	public: // IDropTarget
		STDMETHODIMP DragEnter(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
		{
			HRESULT hr;
			if SUCCEEDED(hr = m_target->DragEnter(data, keys, pt, effect))
				m_owner->SetHotItem(m_index);
			return hr;
		}
		STDMETHODIMP DragOver(DWORD keys, POINTL pt, DWORD* effect)
		{
			return m_target->DragOver(keys, pt, effect);
		}
		STDMETHODIMP DragLeave()
		{
			HRESULT hr = m_target->DragLeave();
			m_owner->SetHotItem(-1);
			return hr;
		}
		STDMETHODIMP Drop(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
		{
			HRESULT hr = m_target->Drop(data, keys, pt, effect);
			m_owner->SetHotItem(-1);
			return hr;
		}
	};
}

void AbstractBar::OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw()
{
	ssize_t index = HitTest(pt);
	Item* item = GetItem(index);
	if (item && is<Path>(item->node))
	{
		if (rc)
			GetItemRect(index, rc);
		ref<IDropTarget> target;
		if SUCCEEDED(cast<Path>(item->node)->GetObject(BHID_SFUIObject, &target))
			*pp = new DropTargetWrapper(target, this, index);
	}
}

//==================================================================================================================================================================

void AbstractBar::set_icons(PyObject* value)
{
	SetIcons(cast<Icons>(value));
}

object AbstractBar::get_items()
{
	ssize_t count = GetItemCount();
	tuple r = Tuple::newobj(count);
	for (ssize_t i = 0; i < count; ++i)
		setitem(r, i, m_items[i].node);
	return r;
}

void AbstractBar::set_items(PyObject* value)
{
	m_items.clear();
	if (is<Null>(value))
		m_items.resize(1);
	else
	{
		array	items = Array::newobj(value);
		ssize_t	ln = len(items);
		m_items.resize(ln+1);
		for (ssize_t i = 0; i < ln; ++i)
		{
			Item& item = m_items[i];
			item.node = items[i];
			item.name = node_get_name(item.node);
			item.enabled = node_enabled(item.node);
			item.checked = node_checked(item.node);
			item.executable = node_executable(item.node);
			item.icon = node_get_icon(item.node);
		}
	}
	ASSERT(m_items.size() > 0);
	m_items[m_items.size()-1].icon = I_IMAGENONE;
	Update(true);
}

//==================================================================================================================================================================

PYTHON_MEMBERS(AbstractBar)
	PYTHON_MEMBER(icons, SET )
	PYTHON_MEMBER(items, BOTH)
PYTHON_END()
