// clipboard.cpp

#include "stdafx.h"
#include "clipboard.hpp"
#include "stream.hpp"
#include "shell.hpp"

using namespace brownie;

//========================================================================================================================

namespace
{
	inline bool OpenClipboard() throw()	{ return ::OpenClipboard(null) != 0; }

	template < typename T > struct ClipFormat;
	template <> struct ClipFormat<PCWSTR> { enum { value = CF_UNICODETEXT }; };

	class Lock
	{
	private:
		bool	m_locked;
		bool	m_mustRelease;
	public:
		Lock(bool locked)
		{
			m_locked = locked;
			if (!m_locked)
			{
				m_locked = OpenClipboard();
				m_mustRelease = true;
			}
		}
		~Lock()
		{
			if (m_locked && m_mustRelease)
				::CloseClipboard();
		}
		operator bool () const throw()	{ return m_locked; }
	};

	template < typename T >
	class Data
	{
	public:
		T	ptr;
	private:
		HGLOBAL	m_mem;
	public:
		Data(CLIPFORMAT format = ClipFormat<T>::value)
		{
			m_mem = ::GetClipboardData(format);
			ptr = (T)::GlobalLock(m_mem);
		}
		~Data()
		{
			if (ptr)
				::GlobalUnlock(m_mem);
		}
		operator T () const throw()	{ return ptr; }
	};
}

//========================================================================================================================

Clipboard::Clipboard() : m_locked(FALSE)
{
}

Clipboard::~Clipboard()
{
	__exit__(null);
}

void Clipboard::__enter__()
{
	if (!m_locked)
		m_locked = OpenClipboard();
}

void Clipboard::__exit__(Tuple* args)
{
	if (m_locked)
	{
		::CloseClipboard();
		m_locked = FALSE;
	}
}

void Clipboard::clear()
{
	Lock lock(m_locked);
	if (lock)
		::EmptyClipboard();
}

void Clipboard::copy(PyObject* value)
{
	set_text(value);
}

void Clipboard::cut(PyObject* value)
{
	set_text(value);
}

object Clipboard::get_text()
{
	Lock lock(m_locked);
	if (lock)
	{
		Data<PCWSTR> data;
		if (data)
			return Unicode::newobj(data);
	}
	return null;
}

void Clipboard::set_Text(PCWSTR text)
{
	ssize_t ln = len(text);
	size_t sz = sizeof(WCHAR) * (ln + 1);
	HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, sz);
	void* p = ::GlobalLock(hGlobal);
	memcpy(p, text, sz);
	::GlobalUnlock(hGlobal);
	Lock lock(m_locked);
	if (lock)
	{
		::EmptyClipboard();
		if (::SetClipboardData(CF_UNICODETEXT, hGlobal))
			return; // ok
	}
	::GlobalFree(hGlobal);
}

void Clipboard::set_text(PyObject* value)
{
	set_Text(str(cast<unicode>(value)));
}

tuple Clipboard::get_file()
{
	return null;
}

void Clipboard::set_file(PyObject* files, bool copy)
{
}

//==================================================================================================================================================================

PYTHON_METHODS(Clipboard)
	PYTHON_METHOD(__enter__)
	PYTHON_METHOD(__exit__)
	PYTHON_METHOD(clear)
	PYTHON_METHOD(copy)
	PYTHON_METHOD(cut)
PYTHON_END()

PYTHON_MEMBERS(Clipboard)
	PYTHON_MEMBER(text		, BOTH)
PYTHON_END()

//========================================================================================================================

#if 0

//================================================================================

namespace
{
	inline tuple DropToTuple(HDROP hDrop)
	{
		UINT count = ::DragQueryFile(hDrop, 0xFFFFFFFF, null, 0);
		tuple files = Tuple::newobj(count);
		for (UINT i = 0; i < count; ++i)
		{
			WCHAR path[MAX_PATH];
			::DragQueryFile(hDrop, i, path, MAX_PATH);
			setitem(files, i, Unicode::newobj(path));
		}
		return files;
	}
}

tuple Clipboard::get_file()
{
	if (::IsClipboardFormatAvailable(CF_HDROP))
	{
		Data<HDROP> cb(CF_HDROP);
		if (cb)
		{
			return newtuple(DropToTuple(cb.ptr), IsCopy());
		}
	}
	return null;
}

//================================================================================

namespace
{
	inline HGLOBAL TupleToDrop(PyObject* files)
	{
		size_t charlen = DoubleNullTextLength(files);
		if (charlen == 0)
			return null;
		//
		DROPFILES dropfiles = { sizeof(DROPFILES), 0, 0, TRUE, TRUE };
		size_t bytelen = sizeof(DROPFILES) + charlen * sizeof(WCHAR);
		//
		HGLOBAL hDrop = GlobalAlloc(GMEM_MOVEABLE, bytelen);
		BYTE* p = (BYTE*)GlobalLock(hDrop);
		memcpy(p, &dropfiles, sizeof(DROPFILES)); p += sizeof(DROPFILES);
		DoubleNullTextFill(files, (PWSTR)p);
		GlobalUnlock(hDrop);
		return hDrop;
	}

	inline HGLOBAL CreateDropEffect(bool copy)
	{
		HGLOBAL hDropEffect = GlobalAlloc(GMEM_MOVEABLE, sizeof(DWORD));
		DWORD* pdw = static_cast<DWORD*>(GlobalLock(hDropEffect));
		*pdw = (copy ? DROPEFFECT_COPY : DROPEFFECT_MOVE);
		GlobalUnlock(hDropEffect);
		return hDropEffect;
	}
}

void Clipboard::set_file(PyObject* files, bool copy)
{
	if (HGLOBAL hDrop = TupleToDrop(files))
	{
		static UINT CF_PREFERREDDROPEFFECT = RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);

		HGLOBAL hDropEffect = CreateDropEffect(copy);

		// Nbv{[hɃf[^[Zbg
		Lock cb;
		if (!cb)
		{
			::GlobalFree(hDrop);
			::GlobalFree(hDropEffect);
			throw WindowsError("OpenClipboard");
		}
		::EmptyClipboard();
		::SetClipboardData(CF_HDROP, hDrop);
		::SetClipboardData(CF_PREFERREDDROPEFFECT, hDropEffect);
	}
}

//================================================================================



namespace avesta
{
	class EnumSequence
	{
	private:
		Sequence*	m_seq;
		int			m_iter;
		ref<Object>	m_item;
	public:
		EnumSequence(PyObject* o);
		~EnumSequence();
		bool next();
		// item
		const ref<Object>& get_item() const throw()	{ return m_item; }
		__declspec(property(get=get_item)) ref<Object> item;
		// index
		int get_index() const throw()			{ return m_iter-1; }
		__declspec(property(get=get_index)) int index;
	};

	//================================================================================

	size_t DoubleNullTextLength(PyObject* seq);
	void DoubleNullTextFill(PyObject* seq, WCHAR buffer[]);
}

//================================================================================

EnumSequence::EnumSequence(Object* o)
{
	m_iter = 0;
	if (is<Tuple>(o) || is<List>(o))
	{
		m_seq = (Sequence*)o;
		m_item = null; 
		Py_INCREF(m_seq);
	}
	else
	{
		m_seq = null;
		m_item = o;
		Py_INCREF(m_item);
	}
}

EnumSequence::~EnumSequence()
{
	decref(m_seq);
}

bool EnumSequence::next()
{
	if (m_seq)
	{
		if (m_iter >= len(m_seq))
			return false;
		m_item = getitem(m_seq, m_iter++);
		return true;
	}
	else
	{
		if (m_iter != 0)
			return false;
		m_iter = 1;
		return true;
	}
}


namespace avesta
{
	size_t DoubleNullTextLength(PyObject* seq);
	void DoubleNullTextFill(PyObject* seq, WCHAR buffer[]);
}

//================================================================================

size_t avesta::DoubleNullTextLength(PyObject* seq)
{
	size_t charlen = 0;
	// foreach (iterator i, seq)
	for (EnumSequence i(seq); i.next();)
	{
		if (PCWSTR text = cast<PCWSTR>(i.item))
		{
			int ln = lstrlenW(text);
			if (ln > 0)
				charlen += ln+1;
		}
	}
	if (charlen == 0)
		return 0;
	else
		return charlen+1;
}

void avesta::DoubleNullTextFill(PyObject* seq, WCHAR buffer[])
{
	BYTE* p = (BYTE*)buffer;
	for (EnumSequence i(seq); i.next();)
	{
		if (PCWSTR text = cast<PCWSTR>(i.item))
		{
			int ln = lstrlenW(text);
			if (ln > 0)
			{
				size_t blen = (ln+1) * sizeof(WCHAR);
				memcpy(p, text, blen); p += blen;
			}
		}
	}
	// footer
	memcpy(p, L"\0", sizeof(WCHAR));
}

#endif

namespace
{
	static HWND	theNextClipboardViewer;
}

bool Clipboard::OnMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT& lResult)
{
	switch (msg)
	{
	case WM_CHANGECBCHAIN:
	{
		HWND hwndRemove = (HWND)wParam;
		HWND hwndNext = (HWND)lParam;
		if (theNextClipboardViewer == hwndRemove)
			theNextClipboardViewer = hwndNext;
		else if (theNextClipboardViewer)
			::SendMessage(theNextClipboardViewer, WM_CHANGECBCHAIN, wParam, lParam);
		return 0;
	}
	case WM_DRAWCLIPBOARD:
	{
		HRESULT hr = S_OK;
		HGLOBAL hGlobal = null;
		try
		{
			UINT cbFormat = CF_HDROP;
			if (GetPriorityClipboardFormat(&cbFormat, 1) == CF_HDROP && ::OpenClipboard(hwnd))
			{
				if (HDROP drop = (HDROP)GetClipboardData(CF_HDROP))
				{
					ref<IStream> stream = StreamCreate(&hGlobal);
					if SUCCEEDED(hr = HDropToString(drop, stream))
					{
						stream = null;
						if (!::SetClipboardData(CF_UNICODETEXT, hGlobal))
							hr = HRESULT_FROM_WIN32(GetLastError());
					}
				}
				::CloseClipboard();
			}
		}
		catch (HRESULT e) { hr = e; }

		if (FAILED(hr) && hGlobal)
			::GlobalFree(hGlobal);
		if (theNextClipboardViewer)
			::SendMessage(theNextClipboardViewer, msg, wParam, lParam);
		return 0;
	}
	case WM_CREATE:
		theNextClipboardViewer = SetClipboardViewer(hwnd);
		break;
	case WM_DESTROY:
		if (theNextClipboardViewer)
		{
			ChangeClipboardChain(hwnd, theNextClipboardViewer);
			theNextClipboardViewer = null;
		}
		break;
	}

	return false;
}
