// exception.cpp

#include "stdafx.h"
#include "alert.hpp"
#include "exception.hpp"
#include "system.hpp"

using namespace brownie;

//==================================================================================================================================================================
// Exception

Exception::Exception()
{
	ASSERT(py::HasLock());
	ASSERT(PyErr_Occurred());
}

Exception::Exception(PyObject* type, PyObject* what)
{
	ASSERT(py::HasLock());
	PyErr_SetObject(type, what);
}

Exception::Exception(PyObject* type, PCSTR name)
{
	ASSERT(py::HasLock());
	PyErr_SetString(type, name);
}

Exception::Exception(PyObject* type, PCWSTR name)
{
	ASSERT(py::HasLock());
	PyErr_SetObject(type, cast<unicode>(name));
}

void Exception::raise()
{
	ASSERT(py::HasLock());

	PyObject* e = PyErr_Occurred();
	ASSERT(e);

	if (!e)
		throw SystemError("Unknown internal error");
#define MATCH(name)	(PyErr_GivenExceptionMatches(PyExc_##name, e))
#define MATCH_THROW(name)	if (PyErr_GivenExceptionMatches(PyExc_##name, e)) { throw name(); }

	if MATCH(Exception)
	{
		MATCH_THROW(TypeError)
		MATCH_THROW(AttributeError)
		if MATCH(EnvironmentError)
		{
			MATCH_THROW(IOError)
			if MATCH(OSError)
			{
				MATCH_THROW(WindowsError)
				throw OSError();
			}
			throw EnvironmentError();
		}
		MATCH_THROW(EOFError)
		if MATCH(RuntimeError)
		{
			MATCH_THROW(NotImplementedError)
			throw RuntimeError();
		}
		if MATCH(NameError)
		{
			MATCH_THROW(UnboundLocalError)
			throw NameError();
		}
		if MATCH(SyntaxError)
		{
			if MATCH(IndentationError)
			{
				MATCH_THROW(TabError)
				throw IndentationError();
			}
			throw SyntaxError();
		}
		if MATCH(LookupError)
		{
			MATCH_THROW(IndexError)
			MATCH_THROW(KeyError)
			throw LookupError();
		}
		if MATCH(ArithmeticError)
		{
			MATCH_THROW(OverflowError)
			MATCH_THROW(ZeroDivisionError)
			MATCH_THROW(FloatingPointError)
			throw ArithmeticError();
		}
		if MATCH(ValueError)
		{
			if MATCH(UnicodeError)
			{
				MATCH_THROW(UnicodeEncodeError)
				MATCH_THROW(UnicodeDecodeError)
				MATCH_THROW(UnicodeTranslateError)
				throw UnicodeError();
			}
			throw ValueError();
		}
		MATCH_THROW(ReferenceError)
		MATCH_THROW(SystemError)
		MATCH_THROW(MemoryError)
		MATCH_THROW(ImportError)
		MATCH_THROW(AssertionError)
		MATCH_THROW(KeyboardInterrupt)
		throw Exception();
	}

#undef MATCH
#undef MATCH_THROW

	throw Exception();
}

//==========================================================================================================================

namespace
{
	inline void fetch_and_normalize(PyObject** e, PyObject** v, PyObject** t) throw()
	{
		PyErr_Fetch(e, v, t);
		PyErr_NormalizeException(e, v, t);
	}

	inline tuple fetch_error() throw()
	{
		object exception, value, traceback;
		fetch_and_normalize(&exception, &value, &traceback);
		PyErr_Clear();
		if (!exception)
			return null;
		tuple err = Tuple::newobj(3);
		setitem(err, 0, exception);
		setitem(err, 1, value);
		setitem(err, 2, traceback);
		return err;
	}

	inline bool hook_error(Tuple* err) throw()
	{
		PyObject* hook = PySys_GetObject("excepthook");
		if (!hook)
			return false;
		try
		{
			return cast<bool>(apply(hook, err));
		}
		catch (Exception&)
		{	// excepthook ŁAɃG[B
			if (PyErr_ExceptionMatches(PyExc_SystemExit))
			{
				System::quit();
				return true;
			}
			PyErr_Clear();
			return false;
		}
	}
/*
def func(type, value, traceback):
    while traceback.tb_next:
        traceback = traceback.tb_next
    
    filename = traceback.tb_frame.f_code.co_filename
    lineno = traceback.tb_lineno
    sys.stderr.write("%s(%d) : %s : %s\n" % (filename, lineno, type, value))
 
sys.excepthook = func
*/
}

void Exception::notify() throw()
{
	if (!PyErr_Occurred())
		return;
	if (PyErr_ExceptionMatches(PyExc_SystemExit))
	{
		System::quit();
		return;
	}
	if (tuple err = fetch_error())
	{
		if (hook_error(err))
			return;
		alert(getitem<unicode>(err, 1), (Unicode*)null, MB_OK | MB_ICONERROR);
	}
}

void Exception::clear() throw()
{
#ifdef _DEBUG
	if (PyObject* err = PyErr_Occurred())
	{
		try
		{
			unicode s = cast<unicode>(err);
			TRACE(str(s));
		}
		catch (Exception&)
		{
			PyErr_Clear();
		}
	}
#endif
	PyErr_Clear();
}

#ifdef _DEBUG
void TRACE(PCWSTR format, ...) throw()
{
	WCHAR buf[1024];
	va_list va;
	va_start(va, format);
	wvsprintf(buf, format, va);
	va_end(va);

	OutputDebugString(buf);
	OutputDebugString(L"\n");
}
#endif
