// export.cpp

#include "stdafx.h"
#include <thumbcache.h>
#include "alert.hpp"
#include "clipboard.hpp"
#include "config.hpp"
#include "abstractbar.hpp"
#include "dialog.hpp"
#include "folderview.hpp"
#include "fileop.hpp"
#include "headerbar.hpp"
#include "listview.hpp"
#include "node.hpp"
#include "path.hpp"
#include "rebar.hpp"
#include "statusbar.hpp"
#include "system.hpp"
#include "tabbar.hpp"
#include "textview.hpp"
#include "toolbar.hpp"
#include "treeview.hpp"
#include "xp/xp_class.hpp"

using namespace brownie;

extern PyObject* brownie_system_execute(PyObject* fn, PyObject* args);

namespace
{
	struct REGCLASS
	{
		CLSID	clsid;
		HRESULT (*create)(REFINTF pp);
	};

#define DEF_CLASS(name)		{ CLSID_##name, Xp##name##Create }

	const REGCLASS CLASS[] =
	{
		DEF_CLASS( ExplorerBrowser ),
		DEF_CLASS( KnownFolderManager ),
		DEF_CLASS( FileOperation ),
		DEF_CLASS( FileOpenDialog ),
		DEF_CLASS( FileSaveDialog ),
		DEF_CLASS( SharedBitmap ),
		DEF_CLASS( LocalThumbnailCache ),
	};
}

HRESULT brownie::CreateInstance(REFCLSID clsid, DWORD clsctx, REFINTF pp)
{
	HRESULT hr = ::CoCreateInstance(clsid, null, clsctx, pp.iid, pp.pp);
	if SUCCEEDED(hr)
		return hr;
	else if (hr != REGDB_E_CLASSNOTREG)
		return hr;

	for (size_t i = 0; i < lengthof(CLASS); ++i)
		if (CLASS[i].clsid == clsid)
			return CLASS[i].create(pp);
	return REGDB_E_CLASSNOTREG;
}

namespace
{
	static PyObject* brownie_exception(PyObject*, PyObject*) throw()
	{
		PyObject* r = PyErr_Occurred();
		if (!r)
			Py_RETURN_NONE;
		Py_INCREF(r);
		return r;
	}

	static PyObject* brownie_modifiers(PyObject*, PyObject*) throw()
	{
		return Integer::newobj(GetModifiers()).detach();
	}

	static PyObject* brownie_now(PyObject*, PyObject*) throw()
	{
		return Integer::newobj(::timeGetTime()).detach();
	}
}

PyMethodDef brownie_methods[] =
{
	{ "alert"		, brownie::alert	, METH_VARARGS },
	{ "beep"		, brownie::beep		, METH_VARARGS },
	{ "exception"	, brownie_exception	, METH_NOARGS  },
	{ "modifiers"	, brownie_modifiers	, METH_NOARGS  },
	{ "now"			, brownie_now		, METH_NOARGS  },
	{ "popup"		, (PyCFunction)brownie::popup, METH_KEYWORDS},
	{ null }
};

PyMODINIT_FUNC
initbrownie(void)
{
	ref<Module> m = Module::newobj("brownie", "Brownie core module", brownie_methods);
#define __def(name)	m->def(#name, name)

	config_export(m);

	// mods
	m->def("SHIFT"		, MK_SHIFT);
	m->def("CONTROL"	, MK_CONTROL);
	m->def("ALT"		, MK_ALTKEY);
	m->def("WIN"		, MK_WINKEY);
	m->def("LBUTTON"	, MK_LBUTTON);
	m->def("RBUTTON"	, MK_RBUTTON);
	m->def("MBUTTON"	, MK_MBUTTON);
	m->def("XBUTTON"	, MK_XBUTTON1);
	m->def("YBUTTON"	, MK_XBUTTON2);

	m->def("KEYS"		, MK_KEYS);
	m->def("BUTTONS"	, MK_BUTTONS);

	// alert
	m->def("LOG"		, LOG);
	m->def("INFO"		, INFO);
	m->def("WARNING"	, WARNING);
	m->def("ERROR"		, ERROR);

	// direction
	__def(NORTH);
	__def(SOUTH);
	__def(WEST);
	__def(EAST);
	__def(CENTER);

	// special parse
	__def(FOLDER_DRIVE);
	__def(FOLDER_HOME);
	__def(FOLDER_DOCUMENT);
	__def(FOLDER_CONTROL);
	__def(FOLDER_NETWORK);
	__def(FOLDER_TRASH);
	__def(FOLDER_PRINTER);
	__def(FOLDER_PARENT);
	__def(FOLDER_BACK);
	__def(FOLDER_FORWARD);

	// view mode
	m->def("ICON"  , FVM_ICON);
	m->def("DETAIL", FVM_DETAILS);
	m->def("TILE"  , FVM_TILE);

	// Windows versions
	__def(WINDOWS_2000);
	__def(WINDOWS_XP);
	__def(WINDOWS_2003);
	__def(WINDOWS_VISTA);

	// class : system
	m->def<System>();

	// class : windows
	m->def<Window>();
	m->def<FolderView>();
	m->def<HeaderBar>();
	m->def<Label>();
	m->def<ListView>();
	m->def<ReBar>();
	m->def<StatusBar>();
	m->def<TabBar>();
	m->def<TextBox>();
	m->def<TextView>();
	m->def<ToolBar>();
	m->def<TreeView>();

	// class : dialog
	m->def<Dialog>();
	m->def<FileDialog>();

	// class : etc.
	m->def<FileOperation_>();
	m->def<Icons>();
	m->def<Clipboard>();
	m->def<Path>();

	m.detach();
}
