// file.cpp

#include "stdafx.h"
#include "class.hpp"
#include "stream.hpp"
#include "unknown.hpp"
#include <C:/bin/Python30/include/fileobject.h>

using namespace brownie;

//==================================================================================================================================================================
// File

void File::close()
{
	if (PyObject* r = PyObject_CallMethod(this, "close", null))
		decref(r);
	else
		Exception::raise();
}

object File::read(ssize_t size)
{
	if (PyObject* r = PyObject_CallMethod(this, "read", (size >= 0 ? "i" : null), (int)size))
		return object::from(r);
	else
		Exception::raise();
}

object File::readline(int size)
{
	if (PyObject* r = PyFile_GetLine(this, size))
		return object::from(r);
	else
		Exception::raise();
}

void File::write(PyObject* obj, int flags)
{
	if (PyFile_WriteObject(this, obj, flags) < 0)
		Exception::raise();
}

void File::write(const char* s)
{
	if (PyFile_WriteString(s, this) < 0)
		Exception::raise();
}

void File::seek(int offset, int whence)
{
	if (PyObject* r = PyObject_CallMethod(this, "seek", "ii", offset, whence))
		decref(r);
	else
		Exception::raise();
}

int File::tell()
{
	if (object r = object::from(PyObject_CallMethod(this, "tell", null)))
		return cast<int>(r);
	else
		Exception::raise();
}

void File::truncate(int size)
{
	if (PyObject* r = PyObject_CallMethod(this, "truncate", (size >= 0 ? "i" : null), (int)size))
		decref(r);
	else
		Exception::raise();
}

//==================================================================================================================================================================
// FileOnStream

namespace
{
	class FileOnStream : public Object
	{
	public:
		ref<IStream>	m_stream;
		bool			m_text;

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.File"; }
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

	public:
		void	close();
		object	read(PyObject* o);
		object	readline();
		void	write(PyObject* o);
		void	seek(Tuple* args);
		object	tell();
		void	truncate(Tuple* args);
		object	get_name();
		object	__iter__() throw();
		object	__next__() throw();
	};
}

object File::newobj(IStream* stream, bool text)
{
	PyType_Ready(typeof<FileOnStream>());
	ref<FileOnStream> o = brownie::newobj<FileOnStream>();
	o->m_stream = stream;
	o->m_text = text;
	return o;
}

void FileOnStream::close()
{
	m_stream = null;
}

object FileOnStream::read(PyObject* o)
{
	int ln = cast<int>(o);
	if (ln <= 0)
		return null;
	std::vector<char> buf;
	buf.resize(ln);
	ULONG done = 0;
	if FAILED(m_stream->Read(&buf[0], (ULONG)ln, &done))
		throw IOError("FileOnStream::read()");
	return String::newobj(&buf[0], done);
}

// FIXME: ƂĂȂB
object FileOnStream::readline()
{
	std::vector<char>	buf;
	buf.reserve(256);

	char c;
	for (;;)
	{
		if (m_stream->Read(&c, 1, null) != S_OK)
		{
			if (buf.empty())
				return null;
			break;
		}
		if (c == '\n')
			break;
		if (c == '\r' && m_text)
			continue;
		buf.push_back(c);
	}

	if (buf.empty())
		return String::newobj("");
	return String::newobj(&buf[0], buf.size());
}

void FileOnStream::write(PyObject* o)
{
	string s = cast<string>(o);
	if FAILED(m_stream->Write(str(s), (ULONG)len(s), null))
		throw IOError("FileOnStream::write()");
}

void FileOnStream::seek(Tuple* args)
{
	int offset, whence = SEEK_SET;
	parse(args, "i|i", &offset, whence);
	try
	{
		StreamSeek(m_stream, offset, whence);
	}
	catch (HRESULT)
	{
		throw IOError("FileOnStream::seek()");
	}
}

object FileOnStream::tell()
{
	try
	{
		UINT64 cur = StreamSeek(m_stream, 0, SEEK_CUR);
		UINT64 ln = StreamSeek(m_stream, 0, SEEK_END);
		StreamSeek(m_stream, cur, SEEK_SET);
		return Integer::newobj((int)ln);
	}
	catch (HRESULT)
	{
		throw IOError("FileOnStream::tell()");
	}
}

void FileOnStream::truncate(Tuple* args)
{
	try
	{
		ULARGE_INTEGER size;
		int ln = -1;
		parse(args, "|i", &ln);
		if (ln < 0)
			size.QuadPart = StreamSeek(m_stream, 0, SEEK_END);
		else
			size.QuadPart = ln;
		m_stream->SetSize(size);
	}
	catch (HRESULT)
	{
		throw IOError("FileOnStream::truncate()");
	}
}

object FileOnStream::get_name()
{
	WCHAR path[MAX_PATH] = L"";
	STATSTG stat = { 0 };
	stat.pwcsName = path;
	if SUCCEEDED(m_stream->Stat(&stat, STATFLAG_DEFAULT))
		return String::newobj(path);
	else
		return null;
}

object FileOnStream::__iter__() throw()	{ return object(self); }
object FileOnStream::__next__() throw()	{ return readline(); }

//==================================================================================================================================================================

PYTHON_METHODS(FileOnStream)
	PYTHON_METHOD(close)
	PYTHON_METHOD(read)
	PYTHON_METHOD(readline)
	PYTHON_METHOD(write)
	PYTHON_METHOD(seek)
	PYTHON_METHOD(tell)
	PYTHON_METHOD(truncate)
PYTHON_END()

PYTHON_MEMBERS(FileOnStream)
	PYTHON_MEMBER(name		, GET)
PYTHON_END()

//========================================================================================================================

namespace
{
	class StreamOnFile : public Unknown< implements<IStream, ISequentialStream>, mixin<StreamImpl, DynamicLife> >
	{
	private:
		ref<File>		m_file;

	public:
		StreamOnFile(File* file) : m_file(file)
		{
		}

		~StreamOnFile()
		{
			py::Lock lock;
			m_file = null;
		}

		HRESULT __stdcall Read(void* data, ULONG size, ULONG* done)
		{
			py::Lock lock;
			try
			{
				if (string s = cast<string>(m_file->read(size)))
				{
					ULONG ln = (ULONG)len(s);
					if (ln > size)
						ln = size;
					memcpy(data, str(s), ln);
					if (done)
						*done = ln;
					return (ln < size) ? S_FALSE : S_OK;
				}
				return S_FALSE;
			}
			catch (Exception&)
			{
				Exception::clear();
				return E_FAIL;
			}
		}

		HRESULT __stdcall Write(const void* data, ULONG size, ULONG* done)
		{
			py::Lock lock;
			try
			{
				m_file->write(String::newobj((PCSTR)data, size));
				return S_OK;
			}
			catch (Exception&)
			{
				Exception::clear();
				return E_FAIL;
			}
		}

		HRESULT __stdcall SetSize(ULARGE_INTEGER sz)
		{
			py::Lock lock;
			try
			{
				m_file->truncate((int)sz.QuadPart);
				return S_OK;
			}
			catch (Exception&)
			{
				Exception::clear();
				return E_FAIL;
			}
		}

		DWORD GetAccessMode() const
		{
			return STGM_READWRITE;
		}

		UINT64 GetSize() const
		{
			py::Lock lock;
			try
			{
				int pos = m_file->tell();
				m_file->seek(0, SEEK_END);
				int sz = m_file->tell();
				m_file->seek(pos);
				return sz;
			}
			catch (Exception&)
			{
				Exception::clear();
				return 0;
			}
		}

		UINT64 GetPosition() const
		{
			py::Lock lock;
			try
			{
				return m_file->tell();
			}
			catch (Exception&)
			{
				Exception::clear();
				return 0;
			}
		}

		HRESULT SeekTo(UINT64 pos)
		{
			py::Lock lock;
			try
			{
				m_file->seek((int)pos);
				return S_OK;
			}
			catch (Exception&)
			{
				Exception::clear();
				return E_FAIL;
			}
		}
	};
}

// ԂIStream̃V[N|C^ʒũt@CɘA邩͎ˑłB
// {ׂ͂StreamOnFileɃ}bv΂悢̂A܂ɔIȂ̂ŁccB
ref<IStream> brownie::StreamCreate(PyObject* o) throw(HRESULT)
{
	if (FileOnStream* s = cast<FileOnStream>(o))
		return s->m_stream;
	else
		return ref<IStream>::from(new StreamOnFile(static_cast<File*>(o)));
}
