// folderstate.cpp

#include "stdafx.h"
#include "autoptr.hpp"
#include "buffer.hpp"
#include "map.hpp"
#include "shell.hpp"
#include "struct.hpp"
#include "thread.hpp"

using namespace brownie;

//==================================================================================================================================================================

BROWNIE_POD(FolderState)

namespace
{
	class __declspec(uuid("95810FAE-EF3D-47E3-BBEE-9C078422D6D6")) FolderStateVersion;

	const int DEFAULT_VIEWMODE	= FVM_DETAILS;
	const int DEFAULT_FLAGS		= FWF_SHOWSELALWAYS | FWF_NOWEBVIEW;
	const int LOCKWAIT_LOOPS	= 10;
	const int LOCKWAIT_TIMEOUT	= 50;
	const FILETIME INVALID_FILETIME = { 0 };

	inline bool STGM_IS_WRITE(DWORD stgm) throw()
	{
		switch (stgm & (STGM_READ | STGM_WRITE | STGM_READWRITE))
		{
		case STGM_WRITE:
		case STGM_READWRITE:
			return true;
		default:
			return false;
		}
	}

	const __time64_t DEFAULT_SETTING_EXPIRE = 7*24*60*60;

	class StateStream;

	class StateBuffer
	{
		friend class StateStream;
	private:
		UINT			m_locks;
		__time64_t		m_time;
		FolderState		m_state;
		BufferT<BYTE>	m_buffer;
	public:
		StateBuffer() : m_locks(0), m_time(0)
		{
			m_state.ViewMode = DEFAULT_VIEWMODE;
			m_state.fFlags = DEFAULT_FLAGS;
			m_state.mtime = INVALID_FILETIME;
		}

		HRESULT	Lock(IStream** pp, FolderState** fs, bool write) throw();
		HRESULT	Unlock(IStream* stream) throw();

		bool CanDelete(__time64_t threshold = _I64_MAX) const throw()
		{
			return m_locks == 0 && m_time < threshold;
		}

		friend IStream& operator >> (IStream& stream, StateBuffer& v) throw(HRESULT)
		{
			v.m_locks = 0;
			return stream >> v.m_time >> v.m_state >> v.m_buffer;
		}

		friend IStream& operator << (IStream& stream, const StateBuffer& v) throw(HRESULT)
		{
			return stream << v.m_time << v.m_state << v.m_buffer;
		}
	};

	class StateStream : public BufferStream
	{
	private:
		StateBuffer&	m_owner;

	public:
		StateStream(StateBuffer& owner) : BufferStream(owner.m_buffer), m_owner(owner)		{ }
		~StateStream()	{ m_owner.Unlock(this); }

	private: // non-copyable
		StateStream(const StateStream&);
		StateStream& operator = (const StateStream&);
	};

	HRESULT StateBuffer::Lock(IStream** pp, FolderState** fs, bool write) throw()
	{
		ASSERT(pp);
		*pp = null;
		if (write)
		{
			if (m_locks != 0)
				return STG_E_SHAREVIOLATION;
			m_locks = UINT_MAX;
			m_buffer.clear();
		}
		else
		{
			if (m_locks == UINT_MAX)
				return STG_E_SHAREVIOLATION;
			++m_locks;
		}
		*pp = new StateStream(*this);
		if (fs)
			*fs = &m_state;
		m_time = _time64(null);
		return S_OK;
	}

	HRESULT StateBuffer::Unlock(IStream* stream) throw()
	{
		ASSERT(m_locks != 0);
		if (m_locks == 0)
			return E_UNEXPECTED;
		if (m_locks == UINT_MAX)
			m_locks = 0;
		else
			--m_locks;
		return S_OK;
	}
}

//==================================================================================================================================================================

namespace
{
	const WCHAR DELIM = (WCHAR)1;

	struct PathCmp
	{
		bool operator () (PCWSTR lhs, PCWSTR rhs) const throw()
		{
			return lstrcmp(lhs, rhs) < 0;
		}
	};

	bool PathRemoveLeaf(PWSTR path) throw()
	{
		if (empty(path))
			return false;
		PWSTR p = wcsrchr(path, DELIM);
		if (!p)
			return false;
		*p = 0;
		return true;
	}

	void PathReplaceSep(PWSTR path) throw()
	{
		ASSERT(path);
		PWSTR p;
		for (p = path; *p; ++p)
			if (*p == L'\\')
				*p = DELIM;
		for (--p; p > path && *p == DELIM; --p)
			*p = 0;
	}

	typedef AutoPtr<WCHAR, CoTaskMem> PathPtr;

	PWSTR ILParsePath(const ITEMIDLIST* pidl) throw()
	{
		if (!pidl)
			return null;

		PWSTR path = null;

		ref<IShellItem> item;
		if SUCCEEDED(ItemCreate(&item, pidl))
		{
			if FAILED(item->GetDisplayName(SIGDN_DESKTOPABSOLUTEPARSING, &path))
				item->GetDisplayName(SIGDN_NORMALDISPLAY, &path);
			if (path)
			{
				PathReplaceSep(path);
				return path;
			}
		}

		path = (PWSTR)::CoTaskMemAlloc(MAX_PATH + sizeof(WCHAR));
		if SUCCEEDED(ILPath(pidl, path))
		{
			PathReplaceSep(path);
			return path;
		}

		::CoTaskMemFree(path);
		return null;
	}

	typedef std::map<PCWSTR, StateBuffer, PathCmp>	Map;
	typedef Map::iterator							Iter;

	Map				theMap;
	CritialSection	theLock;

	Iter Delete(Iter i) throw()
	{
		PCWSTR key = i->first;
		i = theMap.erase(i);
		delete[] key;
		return i;
	}
}

namespace
{
	HRESULT InternalGetFolderStream(PCWSTR path, bool write, IStream** pp, FolderState** fs) throw()
	{
		if (!empty(path))
		{
			HRESULT hr;
			for (int c = 0; c < LOCKWAIT_LOOPS; ++c)
			{
				Lock lock(theLock);
				Iter i = theMap.find(path);
				if (i != theMap.end())
				{
					hr = i->second.Lock(pp, fs, write);
					if (hr == STG_E_SHAREVIOLATION)
					{
						Sleep(LOCKWAIT_TIMEOUT);
						continue; // ̃XbhbNۂ܂ő҂
					}
					return hr;
				}
				else if (write)
				{
					size_t ln = len(path) + 1;
					PWSTR clone = new WCHAR[ln];
					memcpy(clone, path, ln * sizeof(WCHAR));
					theMap[clone].Lock(pp, fs, write);
					return S_OK;
				}
				break; // read and not found
			}
		}

		// TODO: ftHgԃXg[ԂB
		if (fs)
			*fs = null;
		return E_NOTIMPL;
	}
}

//==================================================================================================================================================================

// ̊֐̌ďóAstream  item pXg[ɃZbgꂽ܂܂ɂȂB
HRESULT brownie::FolderStateLoad(const ITEMIDLIST* item, FolderState* state, IStream** stream) throw()
{
	if (!item || !state)
		return E_INVALIDARG;

	ref<IStream> tmp;
	if (stream)
		*stream = null;
	else
		stream = &tmp;
	FolderState* fs;

	PathPtr path = ILParsePath(item);

	// ̃tH_͊ɐݒ肪ۑĂ邩H
	if SUCCEEDED(InternalGetFolderStream(path, false, stream, &fs))
	{
		*state = *fs;
		return S_OK;
	}

	// 傤ǂ̃tH_̐ݒ肪Ȃ΁A킹͕K{B
	state->mtime = INVALID_FILETIME;

	// Ȃꍇ͐e̐ݒgB
	// Rs[͍ŏ1݂̂ɗ}Bȍ~͖B
	if (path)
	{
		while (PathRemoveLeaf(path))
		{
			if SUCCEEDED(InternalGetFolderStream(path, false, stream, &fs))
			{
				state->ViewMode = fs->ViewMode;
				state->fFlags = fs->fFlags;
				return S_OK;
			}
		}
	}

	// Ȃ̂ŁAftHg̐ݒgB
	state->ViewMode = DEFAULT_VIEWMODE;
	state->fFlags = DEFAULT_FLAGS;
	return S_FALSE;
}

HRESULT	brownie::FolderStateSave(const ITEMIDLIST* item, IShellView* view, IStream** stream) throw()
{
	HRESULT hr;

	if (!item || !view || !stream)
		return E_INVALIDARG;

	*stream = null;
	FolderState* fs;

	PathPtr path = ILParsePath(item);

	if FAILED(hr = InternalGetFolderStream(path, true, stream, &fs))
		return hr;

	FOLDERSETTINGS state;
	if SUCCEEDED(view->GetCurrentInfo(&state))
	{
		fs->ViewMode = state.ViewMode;
		fs->fFlags = state.fFlags;
	}
	else
	{
		fs->ViewMode = DEFAULT_VIEWMODE;
		fs->fFlags = DEFAULT_FLAGS;
	}

	// ̃tH_̍ŏIXVۑB
	WIN32_FIND_DATA info;
	if SUCCEEDED(ILInfo(item, &info))
		fs->mtime = info.ftLastWriteTime;
	else
		fs->mtime = INVALID_FILETIME;

	if FAILED(hr = view->SaveViewState())
	{
		ASSERT(!"FolderStateSave failed");
		ULARGE_INTEGER zero = { 0 };
		(*stream)->SetSize(zero);
	}

	(*stream)->Release();
	*stream = null;
	return hr;
}

void brownie::FolderStateReset(const ITEMIDLIST* item)
{
	PathPtr path = ILParsePath(item);
	if (!path)
		return;

	ssize_t ln = len(path);

	Lock lock(theLock);

	Iter i = theMap.lower_bound(path);
	if (i == theMap.end())
		return;

	do
	{
		if (i->second.CanDelete())
			i = Delete(i);
		else
			++i;
	} while (i != theMap.end() && len(i->first) > ln && i->first[ln] == L'\\');
}

HRESULT brownie::FolderStateRestore(PCWSTR path) throw()
{
	FolderStateDispose();

	try
	{
		GUID version;
		ref<IStream> stream = StreamCreate(path, STGM_READ);
		Lock lock(theLock);
		stream >> version;
		if (version != __uuidof(FolderStateVersion))
			return E_FAIL;
		stream >> theMap;
		return S_OK;
	}
	catch (HRESULT hr)
	{
		// r[ȏԂɂȂ邭炢Ȃ΁AׂďB
		FolderStateDispose();
		return hr;
	}
}

namespace
{
	void FolderStateVacuum(__time64_t threshold)
	{
		for (Iter i = theMap.begin(); i != theMap.end();)
		{
			if (i->second.CanDelete(threshold))
				i = Delete(i);
			else
				++i;
		}
	}
}

HRESULT brownie::FolderStateDump(PCWSTR path) throw()
{
	try
	{
		Lock lock(theLock);
		FolderStateVacuum(_time64(null) - DEFAULT_SETTING_EXPIRE);
		ref<IStream> stream = StreamCreate(path, STGM_WRITE | STGM_CREATE);
		stream << __uuidof(FolderStateVersion) << theMap;
	}
	catch (HRESULT hr)
	{
		return hr;
	}
	return S_OK;
}

void brownie::FolderStateDispose() throw()
{
	Lock lock(theLock);
	for (int c = 0; c < LOCKWAIT_LOOPS; c++)
	{
		for (Iter i = theMap.begin(); i != theMap.end();)
		{
			if (i->second.CanDelete())
				i = Delete(i);
			else
				++i;
		}
		Sleep(LOCKWAIT_TIMEOUT);
	}
	ASSERT(theMap.empty());
	theMap.clear();
}
