// graphics.cpp

#include "stdafx.h"
#include "config.hpp"
#include "drawing.hpp"
#include "graphics.hpp"
#include "ref.hpp"
#include "struct.hpp"
#include <vssym32.h>
#pragma comment(lib, "uxtheme")
#pragma comment(lib, "gdi32")

using namespace brownie;

BOOL __stdcall FontDispose(HFONT handle) throw()		{ return ::DeleteObject(handle); }
BOOL __stdcall PenDispose(HPEN handle) throw()			{ return ::DeleteObject(handle); }
BOOL __stdcall BrushDispose(HBRUSH handle) throw()		{ return ::DeleteObject(handle); }
BOOL __stdcall BitmapDispose(HBITMAP handle) throw()	{ return ::DeleteObject(handle); }

void brownie::FillRect(HDC dc, const RECT& rc, COLORREF color) throw()
{
	ASSERT(dc);

	COLORREF old = ::SetBkColor(dc, color);
	if (old != CLR_INVALID)
	{
		VERIFY( ::ExtTextOut(dc, 0, 0, ETO_OPAQUE, &rc, null, 0, null) );
		::SetBkColor(dc, old);
	}
}

void brownie::DrawLine(HDC dc, int x1, int y1, int x2, int y2) throw()
{
	ASSERT(dc);

	VERIFY( ::MoveToEx(dc, x1, y1, null) );
	VERIFY( ::LineTo(dc, x2, y2) );
}

void brownie::GetNonClientMetrics(NONCLIENTMETRICS* metrics)
{
	if (WINDOWS_VERSION < WINDOWS_VISTA)
		metrics->cbSize = offsetof(NONCLIENTMETRICS, iPaddedBorderWidth);
	else
		metrics->cbSize = sizeof(NONCLIENTMETRICS);
	VERIFY( ::SystemParametersInfo(SPI_GETNONCLIENTMETRICS, metrics->cbSize, metrics, 0) );
}

//==================================================================================================================================================================

namespace
{
	IGraphics*	theGraphics;
	HINSTANCE	theGraphicsDll;

	typedef IGraphics* (*GraphicsCreateFn)();
}

HRESULT brownie::GraphicsCreate(PCWSTR name) throw()
{
	GraphicsDispose();

	WCHAR path[MAX_PATH];
#ifdef _DEBUG
	wnsprintf(path, MAX_PATH, L"%s\\%s_d.dll", PATH_BIN, name);
#else
	wnsprintf(path, MAX_PATH, L"%s\\%s.dll", PATH_BIN, name);
#endif

	theGraphicsDll = LoadLibrary(path);
	if (theGraphicsDll)
		if (GraphicsCreateFn fn = (GraphicsCreateFn)::GetProcAddress(theGraphicsDll, "GraphicsCreate"))
			theGraphics = fn();
	return theGraphics ? S_OK : HRESULT_FROM_WIN32(GetLastError());
}

void brownie::GraphicsDispose() throw()
{
	if (theGraphics)
	{
		theGraphics->Dispose();
		theGraphics = null;
		::FreeLibrary(theGraphicsDll);
		theGraphicsDll = null;
	}
}

HBITMAP brownie::BitmapLoad(IStream* stream) throw()
{
	ASSERT(theGraphics);
	return theGraphics->BitmapLoad(stream);
}

HBITMAP brownie::BitmapLoad(PCWSTR filename) throw()
{
	ASSERT(theGraphics);
	return theGraphics->BitmapLoad(filename);
}

HBITMAP brownie::BitmapLoad(IShellItem* item, SIZE sz)
{
	if (!item)
		return null;

	ref<IShellItemImageFactory> factory;
	if SUCCEEDED(item->QueryInterface(&factory))
	{
		HBITMAP	bitmap;
		// SIIGBF_BIGGERSIZEOK  : sz傫ȃTCỸrbg}bvԂ邩ȂB
		// SIIGBF_MEMORYONLY    : fBXNANZXȂBUIXbhĂԏꍇɂ͂𐄏B
		// SIIGBF_ICONONLY      : ɃACRԂBTlC͕ԂȂB
		// SIIGBF_THUMBNAILONLY : ɃTlCԂBTlCT|[gȂꍇ͎sB
		// SIIGBF_INCACHEONLY   : LbVĂCX^X摜ԂBꍇ̓NX摜ԂB
		if FAILED(factory->GetImage(sz, 0, &bitmap))
			return null;
		return bitmap;
	}
	else
	{
		// TODO: for XP
		return null;
	}
}

//========================================================================================================================
// Font

SIZE brownie::FontSize(HFONT font, PCWSTR text, HDC dc) throw()
{
	ASSERT(dc);
	ASSERT(font);

	SIZE sz = { 0, 0 };
	if (empty(text))
		return sz;
	HFONT old = SelectFont(dc, font);
	::GetTextExtentPoint32(dc, text, (int)wcslen(text), &sz);
	SelectFont(dc, old);
	return sz;
}

int brownie::FontHeight(HFONT font, HDC dc) throw()
{
	LOGFONT info;
	GetObject(font, sizeof(info), &info);
	if (info.lfHeight < 0)
		return -info.lfHeight;

	int h;
	if (dc)
		h = -MulDiv(info.lfHeight, 72, GetDeviceCaps(dc, LOGPIXELSY));
	else
	{
		dc = ::GetDC(null);
		h = -MulDiv(info.lfHeight, 72, GetDeviceCaps(dc, LOGPIXELSY));
		::ReleaseDC(null, dc);
	}

	return info.lfHeight;
}

//========================================================================================================================
// ImageList

HIMAGELIST brownie::ImageListLoad(IStream* stream) throw()
{
	ASSERT(theGraphics);
	return theGraphics->ImageListLoad(stream);
}

HIMAGELIST brownie::ImageListLoad(PCWSTR filename) throw()
{
	ASSERT(theGraphics);
	return theGraphics->ImageListLoad(filename);
}

//==========================================================================================================================

Theme::Theme() : m_theme(null)
{
}

Theme::~Theme()
{
	Dispose();
}

bool Theme::Create(HWND hwnd, PCWSTR scheme)
{
	ASSERT(!m_theme);
	m_theme = ::OpenThemeData(hwnd, scheme);
	return m_theme != null;
}

void Theme::Dispose() throw()
{
	if (m_theme)
	{
		::CloseThemeData(m_theme);
		m_theme = null;
	}
}

void Theme::Background(HDC dc, int part, int state, const RECT* rc, const RECT* clip)
{
	if (m_theme)
	{
		DrawThemeBackground(m_theme, dc, part, state, rc, clip);
	}
	else
	{
		UINT s = DFCS_BUTTONPUSH;
		switch (state)
		{
		case TIS_DISABLED:	s |= DFCS_INACTIVE;	break;
		case TIS_HOT:		s |= DFCS_HOT;		break;
		case TIS_SELECTED:	s |= DFCS_PUSHED;	break;
		}
		::DrawFrameControl(dc, const_cast<RECT*>(rc), DFC_BUTTON, s);
	}
}

void Theme::Edge(HDC dc, int part, int state, const RECT* rc, UINT uEdge, UINT uFlags, RECT* rcAdjust)
{
	ASSERT(rc);
	if (m_theme)
	{
		if (rcAdjust)
			uFlags |= BF_ADJUST;
		DrawThemeEdge(m_theme, dc, part, state, rc, uEdge, uFlags, rcAdjust);
	}
	else
	{
		if (rcAdjust)
			*rcAdjust = *rc;
	}
}

void Theme::Text(HDC dc, int part, int state, const RECT* rc, PCWSTR text, DWORD dwFlagsDT, STATUS status)
{
	if (empty(text))
		return;
	::SetBkMode(dc, TRANSPARENT);
	::SetTextColor(dc, GetSysColor(status & DISABLED ? COLOR_GRAYTEXT : COLOR_WINDOWTEXT));
	::DrawText(dc, text, -1, const_cast<RECT*>(rc), dwFlagsDT);
	// DrawThemeText ͐F𐧌łȂ悤Ȃ̂ŁAgȂB
}

/*
	enum WidgetScheme
	{
        BUTTON,
		CLOCK,
		COMBOBOX,
		EDIT,
		EXPLORERBAR,
		GLOBALS,
		HEADER, 
		LISTVIEW,
		MENU,
		MENUBAND,
		PAGE,
		PROGRESS,
		REBAR,
		SCROLLBAR,
		SPIN,
		STARTPANEL,
		STATUS,
		TAB,
		TASKBAND,
		TASKBAR,
		TOOLBAR,
		TOOLTIP,
		TRACKBAR,
		TRAYNOTIFY,
		TREEVIEW,
		WINDOW,
	};
*/
