// layout.cpp

#include "stdafx.h"
#include "math.hpp"
#include "window.hpp"

using namespace brownie;

namespace
{
	class DWP
	{
	private:
		HDWP	m_dwp;
	public:
		DWP(int n)		{ m_dwp = ::BeginDeferWindowPos(n); }
		~DWP()			{ ::EndDeferWindowPos(m_dwp); }
		void operator () (HWND hwnd, const RECT& r)
		{
			m_dwp = ::DeferWindowPos(m_dwp, hwnd, null, RECT_XYWH(r), SWP_NOACTIVATE | SWP_NOZORDER);
		}
	};

	class LayoutBase : public ILayout
	{
	public:
		SIZE		m_margin;
		bool		m_visibleOnly;

		LayoutBase()	{ m_margin.cx = m_margin.cy = 0; m_visibleOnly = true; }
		SIZE	get_Margin() const throw()		{ return m_margin; } 
		void	set_Margin(SIZE value) throw()	{ m_margin = value; }
	};

	class LayoutAll : public LayoutBase
	{
	private:
		struct Child
		{
			HWND		hwnd;
			DIRECTION	dock;

			struct Sort
			{
				bool operator () (const Child& lhs, const Child& rhs) const throw()
				{
					return lhs.dock < rhs.dock;
				}
			};
		};

		typedef std::vector<Child> Children;

	private:
		Children	m_children;

	public:
		void	Insert(HWND hwnd, DIRECTION dock);
		void	GetCenterArea(RECT* bounds) const throw();
		void	SortByDock() throw();
		HWND	Arrange(RECT bounds) const throw();
		void	Copy(std::vector<HWND>& children) const throw();
	};

	class LayoutCenter : public LayoutBase
	{
	private:
		typedef std::vector<HWND> Children;

	private:
		Children	m_children;

	public:
		void	Insert(HWND hwnd, DIRECTION dock);
		HWND	Top() const throw();
		void	Swap(Children& children) throw();
		HWND	GetFocus(HWND focus, DIRECTION dir) const throw();
	};

	template < typename T, typename U >
	inline ssize_t indexof(const T& container, const U& o) throw()
	{
		T::const_iterator i = std::find(container.begin(), container.end(), o);
		if (i == container.end())
			return -1;
		else
			return i - container.begin();
	}
}

//==================================================================================================================================================================

void LayoutAll::Insert(HWND hwnd, DIRECTION dock)
{
	if (!m_visibleOnly || IsVisible(hwnd))
	{
		Child c = { hwnd, dock };
		m_children.push_back(c);
	}
}

void LayoutAll::GetCenterArea(RECT* bounds) const throw()
{
	InflateRect(bounds, -m_margin.cx, -m_margin.cy);

	// Child=㉺E
	for (Children::const_iterator i = m_children.begin(); i != m_children.end(); ++i)
	{
		RECT r;
		::GetWindowRect(i->hwnd, &r);
		SIZE sz = { RECT_WH(r) };

		switch (i->dock)
		{
		case NORTH: bounds->top    += sz.cy; break;
		case SOUTH: bounds->bottom -= sz.cy; break;
		case WEST : bounds->left   += sz.cx; break;
		case EAST : bounds->right  -= sz.cx; break;
		}
	}
}

void LayoutAll::SortByDock() throw()
{
	std::stable_sort(m_children.begin(), m_children.end(), Child::Sort());
}

HWND LayoutAll::Arrange(RECT bounds) const throw()
{
	InflateRect(&bounds, -m_margin.cx, -m_margin.cy);

	DWP dwp((int)m_children.size());
	Children::const_iterator i;

	// Child=㉺E
	for (i = m_children.begin(); i != m_children.end(); ++i)
	{
		if (i->dock == CENTER)
			break;

		RECT r1, r2;
		::GetWindowRect(i->hwnd, &r1);
		int w = RECT_W(r1);
		int h = RECT_H(r1);
		if (Window* o = Window::From(i->hwnd))
		{
			SIZE sz = o->DefaultSize;
			w = math::max<int>(w, sz.cx);
			h = math::max<int>(h, sz.cy);
		}

		switch (i->dock)
		{
		case NORTH:
			SetRect(&r2, bounds.left, bounds.top, bounds.right, bounds.top + h);
			bounds.top += h;
			break;
		case SOUTH:
			SetRect(&r2, bounds.left, bounds.bottom - h, bounds.right, bounds.bottom);
			bounds.bottom -= h;
			break;
		case WEST:
			SetRect(&r2, bounds.left, bounds.top, bounds.left + w, bounds.bottom);
			bounds.left += w;
			break;
		case EAST:
			SetRect(&r2, bounds.right - w, bounds.top, bounds.right, bounds.bottom);
			bounds.right -= w;
			break;
		default:
			ASSERT(0);
			continue;
		}

		if (r1 != r2)
			dwp(i->hwnd, r2);
	}

	// Child=
	int numCenters = (int)(m_children.end() - i);
	if (numCenters == 0)
		return null;

	int horz = numCenters;
	int vert = 1;

	const int PADDING_W = 0;
	const int PADDING_H = 0;
	const int BORDER_W = ::GetSystemMetrics(SM_CXSIZEFRAME);
	const int BORDER_H = ::GetSystemMetrics(SM_CYSIZEFRAME);

	int width  = (RECT_W(bounds) - PADDING_W*2) - BORDER_W * (horz-1);
	int height = (RECT_H(bounds) - PADDING_H*2) - BORDER_H * (vert-1);
	int h = 0, v = 0;

	HWND center = i->hwnd;
	for (size_t n = 0; i != m_children.end(); ++i, ++n)
	{
		ASSERT(i->dock == CENTER);

		const int x = bounds.left + h * width / horz + h * BORDER_W + PADDING_W;
		const int y = bounds.top + v * height / vert + v * BORDER_H + PADDING_H;

		RECT r = { x, y, x+width/horz, y+height/vert };
		if (h == horz-1)
		{	// ԉE[́Ag҂
			r.right = bounds.right - PADDING_W;
		}
		if (v == vert-1)
		{	// ԉ[́Ag҂
			r.bottom = bounds.bottom - PADDING_H;
		}
		dwp(i->hwnd, r);
		if (++h == horz)
		{
			h = 0;
			++v;
		}
	}
	return center;
}

void LayoutAll::Copy(std::vector<HWND>& children) const throw()
{
	children.clear();
	for (Children::const_iterator i = m_children.begin(); i != m_children.end(); ++i)
		children.push_back(i->hwnd);
}

//==================================================================================================================================================================

HWND LayoutCenter::Top() const throw()
{
	return m_children.empty() ? null : m_children.front();
}

void LayoutCenter::Swap(Children& children) throw()
{
	m_children.swap(children);
}

void LayoutCenter::Insert(HWND hwnd, DIRECTION dock)
{
	if (dock == CENTER && (!m_visibleOnly || IsVisible(hwnd)))
		m_children.push_back(hwnd);
}

HWND LayoutCenter::GetFocus(HWND focus, DIRECTION dir) const throw()
{
	size_t count = m_children.size();
	if (count == 0)
	{	// qȂ̂ŎtH[JX
		return null;
	}
	
	size_t i = 0;
	if (::IsWindow(focus) && (i = indexof(m_children, focus)) >= 0)
	{
		switch (dir)
		{
		case EAST:
		case SOUTH:
			if (++i == count)
				i = 0;
			break;
		case WEST:
		case NORTH:
			if (i == 0)
				i = count-1;
			else
				--i;
			break;
		}
	}
	else
	{
		switch (dir)
		{
		case EAST:
		case SOUTH:
			i = 0;
			break;
		case WEST:
		case NORTH:
			i = count - 1;
			break;
		}
	}
	return m_children[i];
}

//==================================================================================================================================================================

HWND brownie::LayoutGetCenter(const Window* w)
{
	LayoutCenter layout;
	w->QueryLayout(layout);
	return layout.Top();
}

void brownie::LayoutEnumCenter(const Window* w, std::vector<HWND>& children)
{
	LayoutCenter layout;
	w->QueryLayout(layout);
	layout.Swap(children);
}

void brownie::LayoutEnumAll(const Window* w, std::vector<HWND>& children)
{
	LayoutAll layout;
	layout.m_visibleOnly = false;
	w->QueryLayout(layout);
	layout.SortByDock();
	layout.Copy(children);
}

void brownie::LayoutCenterArea(const Window* w, RECT* bounds)
{
	LayoutAll layout;
	w->QueryLayout(layout);
	layout.GetCenterArea(bounds);
}

HWND brownie::LayoutArrange(Window* w, RECT bounds)
{
	LayoutAll layout;
	w->QueryLayout(layout);
	layout.SortByDock();
	return layout.Arrange(bounds);
}

HWND brownie::LayoutFocus(Window* w, DIRECTION dir)
{
	LayoutCenter layout;
	w->QueryLayout(layout);
	HWND focus = w->Focus;
	return layout.GetFocus(focus == w->Handle ? null : focus, dir);
}
