// main.cpp

#include "stdafx.h"
#include "clipboard.hpp"
#include "drawing.hpp"
#include "shell.hpp"
#include "stream.hpp"
#include "system.hpp"
#include "thread.hpp"
#include "window.hpp"
#include <map>

#pragma comment(lib, "shlwapi")
#pragma comment(lib, "imm32")

using namespace brownie;

namespace
{
	int					theExitCode = EXIT_ERROR_BUSY;
	PyInterpreterState*	theInterpreter;

	DWORD	TLS_THREAD = INVALID_TLS;
	DWORD	TLS_PYTHON = INVALID_TLS;

	void Save()
	{
		FolderStateDump(PATH_FOLDER_STATE);
	}

	const UINT AUTOSAVE_ID = 1234;
	const UINT AUTOSAVE_INTERVAL = 5 * 60 * 1000; // 5 min.
	void CALLBACK OnAutoSave(HWND hwnd, UINT msg, UINT_PTR idEvent, DWORD dwTime)
	{
		Save();
	}
}

//==================================================================================================================================================================

Thread::Thread() throw() :
	m_hwnd(null), m_wndprocOriginal(null), m_hwndLastWheel(null),
	m_hookMSG(null), m_needsUpdate(false), m_mouseState(MouseNormal)
{
	ASSERT(!Get());

	PyThreadState* pyThread  = PyThreadState_New(theInterpreter);
	::TlsSetValue(TLS_PYTHON, pyThread);

	HINSTANCE hInstance = ::GetModuleHandle(null);
	m_ID = ::GetCurrentThreadId();
	m_hookMSG = ::SetWindowsHookEx(WH_GETMESSAGE, HookMSG, hInstance, m_ID);

	::TlsSetValue(TLS_THREAD, this);
}

namespace
{
	inline PyThreadState* GetPyThread() throw()
	{
		ASSERT(TLS_PYTHON != INVALID_TLS);
		return (PyThreadState*)::TlsGetValue(TLS_PYTHON);
	}
}

Thread::~Thread()
{
	ASSERT(Get() == this);

	::TlsSetValue(TLS_THREAD, null);
	::UnhookWindowsHookEx(m_hookMSG);

	PyThreadState* pyThread = GetPyThread();

	// Kick garbage collector.
	py::CollectGarbage();

	// Release thread descriptor.
	PyEval_AcquireLock();
	if (_PyThreadState_Current == pyThread)
		PyThreadState_Swap(null);
	PyThreadState_Clear(pyThread);
	PyEval_ReleaseLock();

	PyThreadState_Delete(pyThread);
}

Thread* Thread::Get() throw()
{
	if (TLS_THREAD != INVALID_TLS)
		return (Thread*)::TlsGetValue(TLS_THREAD);
	return null;
}

//==================================================================================================================================================================

#ifdef _DEBUG
//#define PYTHON_DEBUG_TIME
#endif

#ifdef PYTHON_DEBUG_TIME
namespace
{
	class PythonDebugTimer
	{
	public:
		PythonDebugTimer()		{ ::timeBeginPeriod(1); }
		~PythonDebugTimer()		{ ::timeEndPeriod(1); }
	} PythonDebugTimer_;

	const DWORD PYTHON_DEBUG_LIMIT = 400; // ms
	DWORD	PythonDebugTime;
}
#endif

void brownie::py::Enter() throw()
{
	ASSERT(!py::HasLock());
	PyEval_AcquireThread(GetPyThread());
#ifdef PYTHON_DEBUG_TIME
	PythonDebugTime = ::timeGetTime();
#endif
}

void brownie::py::Leave() throw()
{
	ASSERT(py::HasLock());
	PyEval_ReleaseThread(GetPyThread());
#ifdef PYTHON_DEBUG_TIME
	if (PythonDebugTime != 0)
	{
		DWORD dt = ::timeGetTime() - PythonDebugTime;
		if (dt > PYTHON_DEBUG_LIMIT)
		{
			if (::IsDebuggerPresent())
			{
				TRACE(L"PYTHON_DEBUG_LIMIT : %u", dt);
				__debugbreak();
			}
			else
			{
				WCHAR msg[1024];
				wsprintf(msg, L"PYTHON_DEBUG_LIMIT : %u", dt);
				::MessageBox(NULL, msg, NULL, MB_OK | MB_ICONWARNING);
			}
		}
	}
#endif
}

bool brownie::py::HasLock() throw()
{
	return _PyThreadState_Current == GetPyThread();
}

void brownie::py::CollectGarbage() throw()
{
	PyThreadState* pyThread = GetPyThread();
	if (_PyThreadState_Current != pyThread)
		PyEval_AcquireThread(pyThread);
	PyGC_Collect();
	PyEval_ReleaseThread(pyThread);
}

namespace
{
	//==================================================================================================================================================================

	unsigned STDMETHODCALLTYPE OnFolk(void* arg)
	{
		Thread	thread;
		HWND	hwnd = null;

		::OleInitialize(null);

		{
			py::Lock lock;
			// XbhOɎQƃJEg𑝂₵Ă̂ŁAł̓A^b`̂݁B
			object fn = object::from((PyObject*)arg);
			try
			{
				object r = apply(fn);
				if (is<Window>(r))
					hwnd = cast<Window>(r)->Handle;
			}
			catch (Exception&)
			{
				Exception::notify();
			}
		}

		// Xbh֐ Window ܂͂̔hNX̃CX^XԂꍇ́A
		// ꂪj܂ŃbZ[W[vɓBȂ΂̂܂܏IB
		if (::IsWindow(hwnd))
			System::MessageLoop(hwnd);

		::OleUninitialize();

		return 0;
	}

	LRESULT CALLBACK System_OnMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
		LRESULT lResult;

		switch (msg)
		{
		case WM_CREATE:
			if ((lResult = ::DefWindowProc(hwnd, msg, wParam, lParam)) != 0)
				return lResult;
			SHObserveBegin(hwnd, WM_SHELL);
			Clipboard::OnMessage(hwnd, msg, wParam, lParam, lResult);
			SetTimer(hwnd, AUTOSAVE_ID, AUTOSAVE_INTERVAL, OnAutoSave);
			return 0;
		case WM_DESTROY:
			if (theExitCode == EXIT_OK)
				theExitCode = EXIT_SHUTDOWN;
			KillTimer(hwnd, AUTOSAVE_ID);
			SHObserveEnd();
			break;
		case WM_ENDSESSION:
			if (!wParam)
				break;
			// fall through
		case WM_CLOSE:
			if (theExitCode == EXIT_OK)
				theExitCode = EXIT_SHUTDOWN;
			// VXeEBhEɑ΂Iv́AׂĂUIEBhEɓ]B
			System::Broadcast(msg, wParam, lParam);
			return 0;
		case WM_FORK:
			System::Fork(OnFolk, (PyObject*)lParam);
			return 0;
		case WM_SHELL:
			SHObserveNotify(wParam, lParam);
			return 0;
		case WM_ACTIVATEWINDOW:
			System::ActivateWindow((int)wParam);
			return 0;
		case WM_RESTART:
			theExitCode = EXIT_RESTART;
			System::Broadcast(WM_CLOSE, wParam, lParam);
			return 0;
		case WM_COPYDATA:
			if (theExitCode != EXIT_OK)
				return FALSE;
			// ܂VXeEBhE󂯎AׂẴEBhEɓ]B
			if (((COPYDATASTRUCT*)lParam)->dwData == COPYDATA_MAGIC)
			{
				return TRUE;
			}
			break;
		case WM_HOTKEY:
			System::OnHotKey(wParam, lParam);
			return 0;
		}

		if (Clipboard::OnMessage(hwnd, msg, wParam, lParam, lResult))
			return lResult;

		return ::DefWindowProc(hwnd, msg, wParam, lParam);
	}
}

//==================================================================================================================================================================
// DLL

namespace
{
	typedef std::map<PCWSTR, HINSTANCE, Less>	DllCache;

	CritialSection	theDllLock;
	DllCache		theDlls;
}

HINSTANCE brownie::DllLoad(PCWSTR name) throw()
{
	Lock lock(theDllLock);
	HINSTANCE& value = theDlls[name];
	if (!value)
		value = ::LoadLibrary(name);
	return value;
}

namespace
{
	void UnloadDlls() throw()
	{
		Lock lock(theDllLock);
		for (DllCache::const_iterator i = theDlls.begin(); i != theDlls.end(); ++i)
			if (i->second)
				::FreeLibrary(i->second);
		theDlls.clear();
	}
}

//==================================================================================================================================================================
// Window : Xbhˑ

namespace
{
	const UINT_PTR	ID_SUBCLASS = 1;
}

Window* Window::From(HWND hwnd) throw()
{
	DWORD_PTR data;
	if (hwnd
	&& ::GetWindowSubclass(hwnd, OnMessage_Main, ID_SUBCLASS, &data)
	&& ((Window*)data)->m_hwnd == hwnd)
		return (Window*)data;
	return null;
}

void Window::create(WNDCLASSEX wc, DWORD style, DWORD styleEx, HWND parent)
{
	if (m_hwnd)
	{	// ̃EBhEɃA^b`
		ASSERT(::IsWindow(m_hwnd));
		m_size.cx = m_size.cy = -1;
		::SetWindowSubclass(m_hwnd, OnMessage_Main, ID_SUBCLASS, (DWORD_PTR)this);
		incref(self);
	}
	else
	{	// VK쐬
		Thread* thread = Thread::Get();
		ASSERT(!thread->m_wndprocOriginal);
		thread->m_wndprocOriginal = wc.lpfnWndProc;
		wc.lpfnWndProc = OnMessage_Bootstrap;

		WCHAR classname[MAX_PATH];
		MultiByteToWideChar(CP_UTF8, 0, self->ob_type->tp_name, -1, classname, MAX_PATH);
		wc.lpszClassName = classname;

		::RegisterClassEx(&wc); // TODO: 2xڈȍ~̌Ăяo͎ŝŒ

		// X^C̒
		if (!(style & WS_POPUP))
		{
			if (parent)
				style |= WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
			else
				style |= WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN;
		}

		// TCY
		RECT rc = { 0, 0, SIZE_WH(m_size) };
		if (parent)
			AdjustWindowRectEx(&rc, style, false, styleEx);

		HWND hwnd = ::CreateWindowEx(styleEx, wc.lpszClassName, null, style,
			CW_USEDEFAULT, CW_USEDEFAULT, RECT_WH(rc), parent, null, wc.hInstance, this);
		UNREFERENCED_PARAMETER(hwnd);
		ASSERT(hwnd == m_hwnd);

		thread->m_wndprocOriginal = null;
	}

	if (!m_hwnd)
		throw WindowsError(wc.lpszClassName);
}

LRESULT CALLBACK Window::OnMessage_Bootstrap(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	Thread* thread = Thread::Get();

	ASSERT(thread->m_wndprocOriginal);
	WNDPROC wndproc = thread->m_wndprocOriginal;

	if (msg == WM_NCCREATE)
	{
		CREATESTRUCT* cs = (CREATESTRUCT*)lParam;
		Window* me = (Window*)cs->lpCreateParams;
		thread->m_wndprocOriginal = null;
		SetWindowProc(hwnd, wndproc);
		::SetWindowSubclass(hwnd, OnMessage_Main, ID_SUBCLASS, (DWORD_PTR)me);
		me->m_hwnd = hwnd;
		me->m_size.cx = me->m_size.cy = -1;
		incref(me);
	}

	return ::CallWindowProc(wndproc, hwnd, msg, wParam, lParam);
}

LRESULT CALLBACK Window::OnMessage_Main(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, UINT_PTR id, DWORD_PTR data)
{
	ASSERT(id == ID_SUBCLASS);

	Window* w = (Window*)data;
	ASSERT(w);
	ASSERT(w->m_hwnd == hwnd);

	LRESULT lResult = 0;
	try
	{
		lResult = w->OnMessage(msg, wParam, lParam);
	}
	catch (Exception&)
	{
		Exception::notify();
		py::Leave(); // ̏ԂPythonbN擾ł邽߁AB
	}

	if (msg == WM_ACTIVATE)
	{
		System::OnActivate(wParam, lParam);
	}
	else if (msg == WM_NCDESTROY)
	{
		::RemoveWindowSubclass(hwnd, OnMessage_Main, ID_SUBCLASS);
		w->m_hwnd = null;

		py::Lock lock;
		decref(w);
	}

	return lResult;
}

void Window::Update(bool needsLayout) throw()
{
	if (needsLayout)
		m_rearrange = true;
	Thread::Get()->Update();
}

//==================================================================================================================================================================

BOOL APIENTRY DllMain(HMODULE module, DWORD what, LPVOID)
{
	switch (what)
	{
	case DLL_PROCESS_ATTACH:
		System::Module = module;
		break;
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
    return TRUE;
}

namespace
{
//#define		DEBUG_NTH_ALLOC		995

#ifdef DEBUG_NTH_ALLOC
	int YourAllocHook(int allocType, void *userData, size_t size, int blockType, long requestNumber, const unsigned char *filename, int lineNumber)
	{
		if (requestNumber == DEBUG_NTH_ALLOC)
			__debugbreak();
		return TRUE;
	}
#endif

	inline PCWSTR main_init(DWORD version)
	{
		// debug options
		_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF | _CRTDBG_CHECK_ALWAYS_DF);
#ifdef DEBUG_NTH_ALLOC
		_CrtSetAllocHook(YourAllocHook);
#endif

		theExitCode = EXIT_ERROR_BUSY;

		// epX̏̂ŁAKԍŏɍsB
		ConfigInit();

		// o[W`FbN
		if (version != PROTOCOL_VERSION)
			return _W("Error.VersionMismatch");

		// Set process DPI aware for VISTA
		AwareDPI();

		// init python
		CHAR path[MAX_PATH];
		::GetModuleFileNameA(null, path, MAX_PATH);
		Py_SetProgramName(path);
		::PathRemoveFileSpecA(path);
		TLS_PYTHON = ::TlsAlloc();
		Py_Initialize();
		PyEval_InitThreads();
		PyThreadState* thread = PyThreadState_Get();
		::TlsSetValue(TLS_PYTHON, thread);
		theInterpreter = thread->interp;
		PyUnicode_SetDefaultEncoding("utf-8");

		// COM
		CoInitializeEx(null, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

		// Graphics subsystem
		if FAILED(GraphicsCreate(L"graphics.gdiplus"))
			return _W("Error.GraphicsCreate");

		// 
		FolderStateRestore(PATH_FOLDER_STATE);

		// VXeEBhE
		TLS_THREAD = ::TlsAlloc();
		System::Init(System_OnMessage);

		return null;
	}

	inline void main_term()
	{
		theExitCode = EXIT_ERROR_BUSY;

		// 
		Save();
		FolderStateDispose();

		// Graphics subsystem
		GraphicsDispose();

		// VXeEBhE
		System::Term();
		::TlsFree(TLS_THREAD);
		TLS_THREAD = INVALID_TLS;

		// term python
		if (PyErr_Occurred())
			PyErr_Clear();
		Py_Finalize();
		theInterpreter = null;
		::TlsFree(TLS_PYTHON);
		TLS_PYTHON = INVALID_TLS;

		// unload dlls
		UnloadDlls();

		CoUninitialize();
	}

	// Entry point is /bin/main.py.
	inline int main_run()
	{
		theExitCode = EXIT_OK;

		// bin, tools tH_ftHgT[`pXɉB
		PyObject* sys_path = PySys_GetObject("path");
		PyList_Insert(sys_path, 1, Unicode::newobj(PATH_BIN));
		PyList_Insert(sys_path, 1, Unicode::newobj(PATH_TOOLS));

		// Gg[|Cg main.py ̓ǂݍ
		CHAR path[MAX_PATH];
		PSTR main;
		::GetModuleFileNameA(System::Module, path, MAX_PATH);
		::PathRemoveFileSpecA(path);
		main = ::PathFindFileNameA(path);
		strcat_s(main, lengthof(path) - (main - path), ".main");
		if (ref<Module> m = Module::import(main))
		{	// C[v
			py::Unlock unlock;
			System::Run();
		}
		else
		{	// ǂݍݎsBԂ񕶖@G[
			Exception::notify();
		}

		return theExitCode;
	}
}

extern "C"
__declspec(dllexport) int main(DWORD version)
{
	int	ret;

	if (PCWSTR error = main_init(version))
	{
		::MessageBox(null, error, APPNAME, MB_OK | MB_ICONERROR);
		return 2;
	}
	ret = main_run();
	main_term();
	return ret;
}
