// node.cpp

#include "stdafx.h"
#include "hwnd.hpp"
#include "node.hpp"
#include "path.hpp"

using namespace brownie;

//========================================================================================================================

namespace
{
	inline object get_attr_or_self(PyObject* self, PCSTR attr)
	{
		if (!self)
			return null;
		else if (object o = getattr(self, attr))
			return o;
		else
			return object(self);
	}

	template < typename T >
	inline T get_value(PyObject* o, T defval) throw()
	{
		if (is<Null>(o))
			return defval;
		try
		{
			if (is_callable(o))
				return cast<T>(apply(o));
			else
				return cast<T>(o);
		}
		catch (Exception&)
		{
			ASSERT(! __FUNCTION__);
			Exception::clear();
		}
		return defval;
	}

	inline iterator get_iter(PyObject* o) throw()
	{
		if (is<Null>(o))
			return null;
		try
		{
			if (!is_iterable(o) && is_callable(o))
			{
				object r = apply(o);
				return is<Null>(r) ? null : Iterator::newobj(r);
			}
			else
				return Iterator::newobj(o);
		}
		catch (Exception&)
		{
			ASSERT(! __FUNCTION__);
			Exception::clear();
		}
		return null;
	}
}

//========================================================================================================================

// unicode(node.name or node)
unicode brownie::node_get_name(PyObject* node) throw()
{
	try
	{
		if (object o = get_attr_or_self(node, "name"))
			return Unicode::newobj(o);
	}
	catch (Exception&)
	{
		ASSERT(! __FUNCTION__);
		Exception::clear();
	}
	return null;
}

// iterable(node.children or node)
bool brownie::node_has_children(PyObject* node) throw()
{
	if (object o = get_attr_or_self(node, "children"))
		return (is_callable(o) || is_iterable(o))
			&& !(is<String>(o) || is<Unicode>(o));
	return false;
}

// callable(node.execute or node)
bool brownie::node_executable(PyObject* node) throw()
{
	return is_callable(get_attr_or_self(node, "execute"));
}

// int(node.icon)
int brownie::node_get_icon(PyObject* node) throw()
{
	return get_value<int>(getattr(node, "icon"), I_IMAGENONE);
}

// iter(node.children or node)
iterator brownie::node_enum_children(PyObject* node)
{
	return get_iter(get_attr_or_self(node, "children"));
}

// iter(node.menu)
iterator brownie::node_enum_menu(PyObject* node, IContextMenu** pp) throw()
{
	if (is<Path>(node))
		cast<Path>(node)->get_menu(pp);
	return get_iter(getattr(node, "menu"));
}

// node.execute() or node()
object brownie::node_execute(PyObject* node, Tuple* args, Dict* kw) throw()
{
	try
	{
		if (object o = get_attr_or_self(node, "execute"))
			if (is_callable(o))
				return apply(o, args, kw);
	}
	catch (Exception&)
	{
		Exception::notify();
	}
	return null;
}

// bool(node.enabled)
bool brownie::node_enabled(PyObject* node) throw()
{
	return get_value<bool>(getattr(node, "enabled"), true);
}

// bool(node.checked)
bool brownie::node_checked(PyObject* node) throw()
{
	return get_value<bool>(getattr(node, "checked"), false);
}

HRESULT brownie::NodeExecute(PyObject* node, UINT vk, UINT mk) throw()
{
	if (is<Null>(node))
		return E_INVALIDARG;

	HRESULT hr;
	if (vk)
		SetKeyState(vk, GetModifiers() | mk);
	do
	{
		py::Lock lock;
		hr = node_execute(node) ? S_OK : E_NOTSUPPORT;
	} while (0);
	if (vk)
		ResetKeyState(vk);
	return hr;
}
