// path.cpp

#include "stdafx.h"
#include "path.hpp"
#include "shell.hpp"
#include "stream.hpp"
#include "system.hpp"

using namespace brownie;
//========================================================================================================================

namespace
{
	template < typename T >
	HRESULT QI(T* self, REFIID iid, void** pp)
	{
		if (!pp)
			return E_POINTER;
		if (iid == __uuidof(T) || iid == __uuidof(IUnknown))
		{
			*pp = self;
			self->AddRef();
			return S_OK;
		}
		else
		{
			*pp = null;
			return E_NOINTERFACE;
		}
	}

#define IMPLEMENT_UNKNOWN(class, interface)																	\
	STDMETHODIMP class::QueryInterface(REFIID iid, void** pp)	{ return QI<interface>(this, iid, pp); }	\
	STDMETHODIMP_(ULONG) class::AddRef()						{ Object::AddRef(); return 1; }				\
	STDMETHODIMP_(ULONG) class::Release()						{ Object::Release(); return 1; }
}

//========================================================================================================================

namespace
{
	class IteratorOnPaths : public Object, public IEnumShellItems
	{
	private:
		ref<Paths>	m_paths;
		size_t		m_index;

	public:
		static ref<IteratorOnPaths> newobj(Paths* paths, size_t index = 0) throw();

	public: // IUnknown
		STDMETHODIMP QueryInterface(REFIID iid, void** pp);
		STDMETHODIMP_(ULONG) AddRef();
		STDMETHODIMP_(ULONG) Release();

	public: // IEnumShellItems
		STDMETHODIMP Next(ULONG celt, IShellItem **rgelt, ULONG *pceltFetched);
		STDMETHODIMP Skip(ULONG celt);
		STDMETHODIMP Reset();
		STDMETHODIMP Clone(IEnumShellItems **ppenum);

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Paths.Iterator"; }
		object	__iter__() throw();
		object	__next__() throw();
	};

	class IteratorOnEnumIDList : public Object
	{
	private:
		ref<Path>			m_parent;
		ref<IEnumIDList>	m_enum;

	public:
		static ref<IteratorOnEnumIDList> newobj(Path* parent, IEnumIDList* e) throw();

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Path.Iterator"; }
		object	__iter__() throw();
		object	__next__() throw();
	};
}

//========================================================================================================================

Path::~Path()
{
	ILDelete(m_id);
}

ref<Path> Path::newobj(PyObject* o) throw()
{
	if (is<Path>(o))
		return ref<Path>(cast<Path>(o));
	unicode u = Unicode::newobj(o);
	return newobj(str(u));
}

ref<Path> Path::newobj_nocopy(ITEMIDLIST* id) throw()
{
	if (!id)
		return null;
	ref<Path> self = brownie::newobj<Path>();
	self->m_id = id;
	return self;
}

ref<Path> Path::newobj(PCWSTR path) throw()
{
	return newobj_nocopy(ILCreate(path));
}

ref<Path> Path::newobj(const ITEMIDLIST* id) throw()
{
	return newobj_nocopy(ILClone(id));
}

ref<Path> Path::newobj(const ITEMIDLIST* parent, const ITEMIDLIST* leaf) throw()
{
	return newobj_nocopy(ILCombine(parent, leaf));
}

ref<Path> Path::newobj(IShellFolder* folder, const ITEMIDLIST* parent, const ITEMIDLIST* leaf) throw()
{
	ref<Path> self = newobj_nocopy(ILCombine(parent, leaf));
	if (folder)
	{
		//  folder 킩Ă̂ŁAłShellItemLbVĂ܂B
		SHCreateShellItem(parent, folder, leaf, &self->m_item);
	}
	return self;
}

ref<Path> Path::newobj(IShellItem* item) throw()
{
	ITEMIDLIST* id = ILCreate(item);
	if (!id)
		return null;
	ref<Path> self = newobj_nocopy(id);
	self->m_item = item;
	return self;
}

void Path::__init__(Tuple* args, Dict* kw) throw(...)
{
	PyObject*	path = null;
	PCSTR kwds[] = { "path", null };
	brownie::parse(args, kw, "|O", kwds, &path);
	if (is<Path>(path))
	{
		Path* rhs = cast<Path>(path);
		m_id = ILClone(rhs->m_id);
		m_item = rhs->m_item;
	}
	else if (path)
	{
		unicode u = cast<unicode>(path);
		PCWSTR s = str(u);
		ITEMIDLIST* id = ILCreate(s);
		if (!id)
			throw IOError(u);
		m_id = id;
	}
}

HRESULT Path::Verify()
{
	if (m_item)
		return S_OK;
	if (!m_id)
	{	// Python AȂ Path() ̌`ō쐬Ă܂悤B
		::SHGetSpecialFolderLocation(null, CSIDL_DESKTOP, &m_id);
	}
	return ItemCreate(&m_item, m_id);
}

//==================================================================================================================================================================

HRESULT Path::GetObject(REFGUID rbhid, REFINTF pp) throw()
{
	HRESULT hr;
	if FAILED(hr = Verify())
		return hr;
	return m_item->BindToHandler(null, rbhid, pp.iid, pp.pp);
}

PWSTR Path::GetDisplayName(SIGDN sigdn) throw()
{
	PWSTR name = null;
	if (FAILED(Verify() || FAILED(m_item->GetDisplayName(sigdn, &name))))
		return null;
	return name;
}

unicode Path::get_displayname(SIGDN sigdn) throw(...)
{
	PWSTR name;
	{
		py::Unlock unlock;
		name = GetDisplayName(sigdn);
	}
	if (!name)
		throw IOError("IShellItem has no name");
	unicode r = Unicode::newobj(name);
	CoTaskMemFree(name);
	return r;
}

bool Path::HasAttribute(SFGAOF sfgaof) throw()
{
	SFGAOF flags = 0;
	if FAILED(Verify())
		return false;
	if FAILED(m_item->GetAttributes(sfgaof, &flags))
		return false;
	return (flags & sfgaof) != 0;
}

HRESULT Path::get_menu(IContextMenu** pp) throw()
{
	py::Unlock unlock;

	HRESULT hr;
	if FAILED(hr = Verify())
		return hr;
	return m_item->BindToHandler(null, BHID_SFUIObject, IID_PPV_ARGS(pp));
}

void Path::execute(Tuple* args, Dict* kw)
{
	PyObject* o[3] = { null };
	PCWSTR s[3] = { null };
	unicode u[3];
	PCSTR kwds[] = { "verb", "argv", "dir", null };
	brownie::parse(args, kw, "|OOO", kwds, &o[0], &o[1], &o[2]);

	for (int i = 0; i < 3; ++i)
	{
		if (o[i])
		{
			u[i] = cast<unicode>(o[i]);
			s[i] = str(u[i]);
		}
	}

	py::Unlock unlock;
	if FAILED(ILExecute(m_id, s[0], s[1], s[2]))
		throw WindowsError();
}

ref<IStream> Path::Open(DWORD stgm) throw()
{
	// zACeł邽߂ɏLsꍇɂ́AoRȂ΃t@Cǂݏł\B
	// Vistał̃oCfoCX(PDA)̃ACeȂǂɑ΂Ċ҂łB
	if FAILED(Verify())
		return null;

	ref<IBindCtx> bind;
	if SUCCEEDED(CreateBindCtx(0, &bind))
	{
		BIND_OPTS tools = { sizeof(BIND_OPTS), 0, stgm, 0 };
		bind->SetBindOptions(&tools);
	}
	ref<IStream> stream;
	if SUCCEEDED(m_item->BindToHandler(bind, BHID_Stream, IID_PPV_ARGS(&stream)))
		return stream;

	if (PWSTR name = GetDisplayName(SIGDN_DESKTOPABSOLUTEPARSING))
	{
		try { stream = StreamCreate(name, stgm); } catch(HRESULT) {}
		CoTaskMemFree(name);
		if (stream)
			return stream;
	}

	return null; // error
}

namespace
{
	DWORD ModeStringToSTGM(PCSTR mode) throw()
	{
		if (strcmp(mode, "r") == 0 || strcmp(mode, "rb") == 0)
			return STGM_READ;
		else if (strcmp(mode, "w") == 0 || strcmp(mode, "wb") == 0)
			return STGM_WRITE;
		else
			return STGM_READWRITE;
	}
}

object Path::open(Tuple* args)
{
	PCSTR mode = null;
	parse(args, "|s", &mode);
	if (empty(mode))
		mode = "r";

	ref<IStream> stream;
	{
		py::Unlock unlock;
		stream = Open(ModeStringToSTGM(mode));
	}
	if (stream)
		return File::newobj(stream, strchr(mode, 'r') != null);

	throw IOError("Path.open() is not supported.");
}

// Vista: BOOL CreateSymbolicLink(srcpath, dstpath, SYMLINK_FLAG_DIRECTORY or 0);
void Path::symlink(PyObject* o)
{
	WCHAR dst[MAX_PATH];
	wcscpy_s(dst, str(cast<unicode>(o)));

	// gq͋I .lnk
	if (!HasAttribute(SFGAO_FOLDER))
	{
		PathRemoveExtension(dst);
		PathAddExtension(dst, L".lnk");
	}
	else if (_wcsicmp(PathFindExtension(dst), L".lnk") != 0)
	{
		wcscat_s(dst, L".lnk");
	}

	HRESULT hr = LinkCreate(m_id, dst);
	if FAILED(hr)
		throw IOError("path.symlink()");
}

// ʏ́u΃pXvu΃pXvuURLv3ނ݂̂ۂB
// .lnk Ȃǂ̏ꍇ́APARSING  EDITING ňႢ邩B
// SIGDN_NORMALDISPLAY == SIGDN_PARENTRELATIVEPARSING == SIGDN_PARENTRELATIVEEDITING
// SIGDN_PARENTRELATIVEFORADDRESSBAR V[gJbg .lnk ̂ႢH
// SIGDN_DESKTOPABSOLUTEPARSING == SIGDN_DESKTOPABSOLUTEEDITING == SIGDN_FILESYSPATH
// SIGDN_URL

object Path::get_name()	{ return get_displayname(SIGDN_PARENTRELATIVE); }
object Path::get_path()	{ return get_displayname(SIGDN_DESKTOPABSOLUTEPARSING); }
//object Path::get_url()	{ return get_displayname(SIGDN_URL); }

BROWSE Path::GetBrowsable()
{
	if (!HasAttribute(SFGAO_FOLDER))
		return BROWSE_NG;

	AutoPtr<WCHAR, CoTaskMem> path(GetPath());
	if (!path || PathIsDirectory(path))
		return BROWSE_PREFERRED;	// ztH_ or fBNg

	PCWSTR ext = PathFindExtension(path);
	WCHAR key[MAX_PATH], def[MAX_PATH], command[MAX_PATH], buffer[MAX_PATH];
	if FAILED(RegGetString(HKEY_CLASSES_ROOT, ext, NULL, key))
		return BROWSE_PREFERRED; // o^
	wsprintf(buffer, L"%s\\shell", key);
	if FAILED(RegGetString(HKEY_CLASSES_ROOT, buffer, NULL, def))
		return BROWSE_PREFERRED; // G[
	if (_wcsicmp(def, L"explore") == 0)
		return BROWSE_PREFERRED; // GNXv[Ɋ֘At
	if (empty(def))
		lstrcpy(def, L"open");
	wsprintf(buffer, L"%s\\shell\\%s\\command", key, def);
	if FAILED(RegGetString(HKEY_CLASSES_ROOT, buffer, NULL, command))
		return BROWSE_PREFERRED;

	// AvP[VɊ֘AtĂB
	// {͏o邪AftHgsƂĂ̓t@CƂĈׂB
	return BROWSE_OK;
}

object Path::get_browsable()
{
	int r;
	{
		py::Unlock unlock;
		r = GetBrowsable();
	}
	return cast<integer>(r);
}

object Path::get_has_parent()
{
	return Boolean::newobj(ILIsRoot(m_id));
}

object Path::get_has_subfolder()
{
	bool r;
	{
		py::Unlock unlock;
		r = HasAttribute(SFGAO_HASSUBFOLDER);
	}
	return cast<boolean>(r);
}

ref<Path> Path::parent()
{
	if (ILIsRoot(m_id))
		return null;

	ref<Path> parent = brownie::newobj<Path>();
	parent->m_id = ILCloneParent(m_id);
	if (m_item)
	{
		py::Unlock unlock;
		m_item->GetParent(&parent->m_item);
	}
	return parent;
}

object Path::get_parent()
{
	if (ref<Path> o = parent())
		return object(o->self);
	else
		return null;
}

object Path::__iter__()
{
	return get_children();
}

object Path::get_children(DWORD flags)
{
	Verify();
	ref<IEnumIDList> iter;

	// hCuɃfBAĂȂƂ́AŎԂ悤B
	{

		py::Unlock unlock;
		ref<IShellFolder> folder;
		if (m_item)
			m_item->BindToHandler(null, BHID_SFObject, IID_PPV_ARGS(&folder));
		if (!folder)
			ILFolder(&folder, m_id);
		if (!folder)
			return null;

		//SHCONTF_INCLUDEHIDDEN
		if FAILED(folder->EnumObjects(System::GetWindow(), flags, &iter))
			return null;
	}
	return IteratorOnEnumIDList::newobj(this, iter);
}

object Path::get_children()
{
	return get_children(SHCONTF_FOLDERS | SHCONTF_NONFOLDERS);
}

object Path::get_subfolders()
{
	return get_children(SHCONTF_FOLDERS);
}

object Path::get_linked()
{
	ITEMIDLIST* linked = null;

	{
		py::Unlock unlock;
		if (PWSTR name = GetDisplayName(SIGDN_DESKTOPABSOLUTEPARSING))
		{
			LinkResolve(name, &linked);
			CoTaskMemFree(name);
		}
	}

	if (linked)
		return Path::newobj_nocopy(linked);
	else
		return object(self);	// .link łȂꍇ͎gԂ
}

object Path::get_icon()
{
	int index;
	{
		py::Unlock unlock;
		index = ILIconIndex(m_id);
	}
	return (index < 0) ? null : Integer::newobj(MAKELONG(index, 1));
}

//==================================================================================================================================================================

PYTHON_METHODS(Path)
	PYTHON_METHOD_(__unicode__,	get_path)
	PYTHON_METHOD(execute)
	PYTHON_METHOD(open)
	PYTHON_METHOD(symlink)
PYTHON_END()

PYTHON_MEMBERS(Path)
	PYTHON_MEMBER(name			, GET)
	PYTHON_MEMBER(path			, GET)
	PYTHON_MEMBER(browsable		, GET)
	PYTHON_MEMBER(has_subfolder	, GET)
	PYTHON_MEMBER(has_parent		, GET)
	PYTHON_MEMBER(parent		, GET)
	PYTHON_MEMBER(children		, GET)
	PYTHON_MEMBER(subfolders	, GET)
	PYTHON_MEMBER(linked		, GET)
	PYTHON_MEMBER(icon			, GET)
PYTHON_END()

//========================================================================================================================

ref<Paths> Paths::newobj(const CIDA* items, IShellFolder* folder, bool iscopy)
{
	PyType_Ready(typeof<Paths>());
	ref<Paths> self = brownie::newobj<Paths>();
	self->m_items = CIDAClone(items);
	self->m_object = folder;
	self->m_iscopy = iscopy;
	return self;
}

ref<Paths> Paths::newobj(IShellItemArray* items, bool iscopy) throw()
{
	PyType_Ready(typeof<Paths>());
	ref<Paths> self = brownie::newobj<Paths>();
	self->m_object = items;
	self->m_iscopy = iscopy;
	return self;
}

inline IShellItemArray* Paths::get_array() const throw()
{
	return static_cast<IShellItemArray*>(m_object.ptr);
}

inline IShellFolder* Paths::get_folder() const throw()
{
	return static_cast<IShellFolder*>(m_object.ptr);
}

//========================================================================================================================
// Paths.IShellItemArray

IMPLEMENT_UNKNOWN(Paths, IShellItemArray)

STDMETHODIMP Paths::BindToHandler(IBindCtx *pbc, REFGUID rbhid, REFIID riid, void **pp)
{
	if (m_items)
		return E_NOTIMPL;
	else
		return m_array->BindToHandler(pbc, rbhid, riid, pp);
}

STDMETHODIMP Paths::GetPropertyStore(GETPROPERTYSTOREFLAGS flags, REFIID riid, void **ppv)
{
	if (m_items)
		return E_NOTIMPL;
	else
		return m_array->GetPropertyStore(flags, riid, ppv);
}

STDMETHODIMP Paths::GetPropertyDescriptionList(REFPROPERTYKEY keyType, REFIID riid, void **ppv)
{
	if (m_items)
		return E_NOTIMPL;
	else
		return m_array->GetPropertyDescriptionList(keyType, riid, ppv);
}

STDMETHODIMP Paths::GetAttributes(SIATTRIBFLAGS dwAttribFlags, SFGAOF sfgaoMask, SFGAOF *psfgaoAttribs)
{
	if (m_items)
		return E_NOTIMPL;
	else
		return m_array->GetAttributes(dwAttribFlags, sfgaoMask, psfgaoAttribs);
}

STDMETHODIMP Paths::GetCount(DWORD* pdwNumItems)
{
	if (m_items)
	{
		*pdwNumItems = (DWORD)CIDACount(m_items);
		return S_OK;
	}
	else
		return m_array->GetCount(pdwNumItems);
}

STDMETHODIMP Paths::GetItemAt(DWORD dwIndex, IShellItem **ppsi)
{
	if (m_items)
	{
		const ITEMIDLIST* parent = CIDAParent(m_items);
		if (!m_object)
			InitFolder(parent);
		return ItemCreate(ppsi, m_folder, parent, CIDAChild(m_items, dwIndex));
	}
	else
		return m_array->GetItemAt(dwIndex, ppsi);
}

STDMETHODIMP Paths::EnumItems(IEnumShellItems** pp)
{
	if (m_items)
	{
		*pp = IteratorOnPaths::newobj(this).detach();
		return S_OK;
	}
	else
		return m_array->EnumItems(pp);
}

void Paths::InitFolder(const ITEMIDLIST* parent)
{
	ref<IShellFolder> folder;
	ILFolder(&folder, parent);
	m_object = folder;
}

//========================================================================================================================

object Paths::__getitem__(ssize_t index) throw(...)
{
	size_t ln = __len__();
	size_t i = (index >= 0 ? index : ln + index + 1);
	if (i >= ln)
		throw IndexError("Paths index out of bounds");

	if (m_items)
	{
		const ITEMIDLIST* parent = CIDAParent(m_items);
		if (!m_object)
		{
			py::Unlock unlock;
			InitFolder(parent);
		}
		return object(Path::newobj(m_folder, parent, CIDAChild(m_items, i)));
	}
	else
	{
		ref<IShellItem> item;
		if FAILED(m_array->GetItemAt((DWORD)i, &item))
			throw IOError("Paths invalid item");
		return object(Path::newobj(item));
	}
}

ssize_t Paths::__len__() throw()
{
	if (m_items)
		return (int)CIDACount(m_items);
	else
	{
		DWORD count = 0;
		m_array->GetCount(&count);
		return (ssize_t)count;
	}
}

//========================================================================================================================

object Paths::__iter__() throw()				{ return IteratorOnPaths::newobj(this); }
ssize_t Paths::__len__(PyObject* self) throw()	{ return cast<Paths>(self)->__len__(); }

PyObject* Paths::__getitem__(PyObject* self, ssize_t index) throw()
{
	try
	{
		if (PyObject* r = cast<Paths>(self)->__getitem__(index).detach())
			return r;
		else
			Py_RETURN_NONE;
	}
	catch (Exception&)
	{
		return null;
	}
}

PySequenceMethods* Paths::__sequence__() throw()
{
	static PySequenceMethods seq =
	{
		&Paths::__len__,		// sq_length;
		null,		// sq_concat;
		null,		// sq_repeat;
		&Paths::__getitem__,		// sq_item;
		null,		// sq_slice;
		null,		// sq_ass_item;
		null,		// sq_ass_slice;
		null,		// sq_contains;
		null,		// sq_inplace_concat;
		null,		// sq_inplace_repeat;
	};
	return &seq;
} PySequenceMethods;

//========================================================================================================================

IMPLEMENT_UNKNOWN(IteratorOnPaths, IEnumShellItems)

ref<IteratorOnPaths> IteratorOnPaths::newobj(Paths* paths, size_t index) throw()
{
	PyType_Ready(typeof<IteratorOnPaths>());
	ref<IteratorOnPaths> self = brownie::newobj<IteratorOnPaths>();
	self->m_paths = paths;
	self->m_index = index;
	return self;
}

STDMETHODIMP IteratorOnPaths::Next(ULONG celt, IShellItem **rgelt, ULONG *pceltFetched)
{
	DWORD count = 0;
	m_paths->GetCount(&count);
	ULONG i;
	for (i = 0; i < celt && m_index < count; i++, m_index++)
		m_paths->GetItemAt(i, rgelt + i);
	if (pceltFetched)
		*pceltFetched = i;
	if (i == celt)
		return S_OK;
	else if (i > 0)
		return S_FALSE;
	else
		return E_FAIL;
}

STDMETHODIMP IteratorOnPaths::Skip(ULONG celt)
{
	m_index += celt;
	return S_OK;
}

STDMETHODIMP IteratorOnPaths::Reset()
{
	m_index = 0;
	return S_OK;
}

STDMETHODIMP IteratorOnPaths::Clone(IEnumShellItems **ppenum)
{
	if (!ppenum)
		return E_POINTER;
	*ppenum = newobj(m_paths, m_index).detach();
	return S_OK;
}

object IteratorOnPaths::__iter__() throw()	{ return object(self); }

object IteratorOnPaths::__next__() throw()
{
	if ((ssize_t)m_index < m_paths->__len__())
		return m_paths->__getitem__(m_index++);
	return null;
}

void Paths::execute(Tuple* args, Dict* kw)
{
	ssize_t ln = __len__();
	if (ln == 0)
		return;

	PCWSTR s_verb = null;
	PCWSTR s_args = null;
	unicode u_verb;
	unicode u_args;
	PyObject* o_verb = null;
	PyObject* o_args = null;
	PCSTR kwds[] = { "verb", "argv", null };
	brownie::parse(args, kw, "|OO", kwds, &o_verb, &o_args);

	if (o_verb)
	{
		u_verb = cast<unicode>(o_verb);
		s_verb = str(u_verb);
	}
	if (o_args)
	{
		u_args = cast<unicode>(o_args);
		s_args = str(u_args);
	}

	py::Unlock unlock;

	if (m_items)
	{
		for (ssize_t i = 0; i < ln; ++i)
		{
			ILPtr item = ILCreate(m_items, i);
			ILExecute(item, s_verb, s_args);
		}
	}
	else
	{
		for (ssize_t i = 0; i < ln; ++i)
		{
			ref<IShellItem> item;
			if SUCCEEDED(m_array->GetItemAt((DWORD)i, &item))
				ILExecute(ILPtr(ILCreate(item)), s_verb, s_args);
		}
	}
}

PYTHON_METHODS(Paths)
	PYTHON_METHOD(execute)
PYTHON_END()

//========================================================================================================================

ref<IteratorOnEnumIDList> IteratorOnEnumIDList::newobj(Path* parent, IEnumIDList* e) throw()
{
	PyType_Ready(typeof<IteratorOnEnumIDList>());
	ref<IteratorOnEnumIDList> self = brownie::newobj<IteratorOnEnumIDList>();
	self->m_parent = parent;
	self->m_enum = e;
	return self;
}

object IteratorOnEnumIDList::__iter__() throw()	{ return object(self); }

object IteratorOnEnumIDList::__next__() throw()
{
	ITEMIDLIST* leaf = null;
	ULONG fetched = 0;
	if (m_enum->Next(1, &leaf, &fetched) != S_OK)
		return null;
	ref<Path> child = Path::newobj(m_parent->ID, leaf);
	ILDelete(leaf);
	return child;
}
