// py.cpp

#include "stdafx.h"
#include "class.hpp"
#include "exception.hpp"

using namespace brownie;

//==================================================================================================================================================================

bool brownie::same(IUnknown* lhs, IUnknown* rhs) throw()
{
	if (lhs == rhs)
		return true;
	if (!lhs || !rhs)
		return false;
	ref<IUnknown> lu, ru;
	lhs->QueryInterface(&lu);
	rhs->QueryInterface(&ru);
	return lu.ptr == ru.ptr;
}

//==================================================================================================================================================================

Object* brownie::detail::ObjectFrom(PyObject* o, PyTypeObject* type) throw(...)
{
	if (is<Null>(o))
		return null;
	if (!PyObject_TypeCheck(o, type))
		throw TypeError(o);
	return (Object*)((BYTE*)o + sizeof(PyObject));
}

namespace
{
#ifdef _DEBUG
	void TRACE_OBJECT(PCWSTR message, PyObject* self)
	{
		static int line = 0;
		TRACE(L"PYTHON[%4d]: %S [obj=0x%08X, refs=%d] : %s()",
			line++, self->ob_type->tp_name, self, self->ob_refcnt, message);
	}
#else
#	define TRACE_OBJECT	__noop
#endif

	inline size_t GetMemberOffset(const PyGetSetDef* m) throw()
	{
		size_t offset = (size_t)m->closure;
		ASSERT(offset <= 0xFFFF);
		return offset;
	}

	inline PyObject** ObjectMember(PyObject* self, const PyGetSetDef* m) throw()
	{
		size_t offset = GetMemberOffset(m);
		if (offset > 0)
			return (PyObject**)((BYTE*)self + offset);
		else
			return null;
	}

	inline bool HasMember(const PyGetSetDef members[]) throw()
	{
		if (!members)
			return false;
		for (const PyGetSetDef* m = members; m->name; ++m)
			if (GetMemberOffset(m))
				return true;
		return false;
	}

	// ЂƂ̌^Kwɑ΂traverse
	inline int GenericTraverseOne(PyObject* self, PyTypeObject* type, visitproc visit, void* arg) throw()
	{
		if (const PyGetSetDef* members = type->tp_getset)
			for (const PyGetSetDef* m = members; m->name; ++m)
				if (PyObject** p = ObjectMember(self, m))
					Py_VISIT(*p);
		return 0;
	}

	// ЂƂ̌^Kwɑ΂clear
	inline int GenericClearOne(PyObject* self, PyTypeObject* type) throw()
	{
		if (const PyGetSetDef* members = type->tp_getset)
			for (const PyGetSetDef* m = members; m->name; ++m)
				if (PyObject** p = ObjectMember(self, m))
					Py_CLEAR(*p);
		return 0;
	}
}

int py::GenericTraverse(PyObject* self, visitproc visit, void* arg) throw()
{
	for (PyTypeObject* type = self->ob_type; type; type = type->tp_base)
		if (type->tp_traverse == GenericTraverse)
			if (int r = GenericTraverseOne(self, type, visit, arg))
				return r;
	return 0;
}

int py::GenericClear(PyObject* self) throw()
{
	for (PyTypeObject* type = self->ob_type; type; type = type->tp_base)
		if (type->tp_clear == GenericClear)
			if (int r = GenericClearOne(self, type))
				return r;
	return 0;
}

PyObject* py::GenericNew(PyTypeObject* type) throw()
{
	PyObject* self = type->tp_alloc(type, 0);
	TRACE_OBJECT(L"new", self);
	return self;
}

void py::GenericDelete(PyObject* self) throw()
{
	if (self->ob_type->tp_clear)
		self->ob_type->tp_clear(self);
	TRACE_OBJECT(L"delete", self);
}

int py::GenericTypeFlags(const PyGetSetDef members[], PyTypeObject* base) throw()
{
	int flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE;
	if ((base && PyType_IS_GC(base)) || HasMember(members))
		flags |= Py_TPFLAGS_HAVE_GC;
    return flags;
}

namespace
{
	PyObject* GetUnicodeToString(PyObject* self) throw()
	{
		if (PyObject* s = PyObject_CallMethod(self, "__unicode__", ""))
		{
			PyObject* u = PyUnicode_FromObject(s);
			decref(s);
			if (u)
				return u;
		}
		return PyObject_Repr(self);
	}
}

reprfunc py::GenericStr(const PyMethodDef methods[]) throw()
{
	if (!methods)
		return null;
	for (const PyMethodDef* m = methods; m->ml_name; ++m)
	{
		if (strcmp(m->ml_name, "__unicode__") == 0)
			return GetUnicodeToString;
	}
	return null;
}

//==================================================================================================================================================================
// Boolean

namespace
{
	static bool parse_boolean(PCWSTR value) throw(...)
	{
		const PCWSTR TRUES[] = { L"true", L"on", L"yes", L"1" };
		const PCWSTR FALSES[] = { L"false", L"off", L"no", L"0" };

		for (size_t i = 0; i < lengthof(TRUES); ++i)
			if (_wcsicmp(value, TRUES[i]) == 0)
				return true;
		for (size_t i = 0; i < lengthof(FALSES); ++i)
			if (_wcsicmp(value, FALSES[i]) == 0)
				return false;

		WCHAR message[MAX_PATH];
		swprintf_s(message, L"Invalid boolean expression. %s", value);
		throw ValueError(message);
	}
}

bool Boolean::value(PyObject* o) throw(...)
{
	if (is<Null>(o))
		return false;
	if (is<String>(o) || is<Unicode>(o))
		return parse_boolean(str(cast<unicode>(o)));
	int r = PyObject_IsTrue(o);
	if (r < 0)
		Exception::raise();
	return r != 0;
}

//==================================================================================================================================================================
// Integer

integer Integer::newobj(int value) throw()
{
	return integer::from(PyInt_FromLong(value));
}

int Integer::value(Integer* o) throw()
{
	return o ? PyInt_AS_LONG(o) : 0;
}

int Integer::value(PyObject* o) throw(...)
{
	if (is<Null>(o))
		return 0;
	if (is<String>(o))
		return atoi(str(static_cast<String*>(o)));
	if (is<Unicode>(o))
		return _wtoi(str(static_cast<Unicode*>(o)));
	int r = PyInt_AsLong(o);
	if (r == -1 && PyErr_Occurred())
		Exception::raise();
	return r;
}

//==================================================================================================================================================================
// Numeric

numeric Numeric::newobj(__int64 value) throw()
{
	return numeric::from(PyLong_FromLongLong(value));
}

__int64 Numeric::value(PyObject* o) throw(...)
{
	if (is<Null>(o))
		return 0;
	if (is<String>(o))
		return _atoi64(str(static_cast<String*>(o)));
	if (is<Unicode>(o))
		return _wtoi64(str(static_cast<Unicode*>(o)));
	__int64 r = PyLong_AsLongLong(o);
	if (r == -1 && PyErr_Occurred())
		Exception::raise();
	return r;
}

//==================================================================================================================================================================
// Real

double Real::value(PyObject* o) throw(...)
{
	if (is<Null>(o))
		return 0.0;
	if (is<String>(o))
		return atof(str(static_cast<String*>(o)));
	if (is<Unicode>(o))
		return _wtof(str(static_cast<Unicode*>(o)));
	return PyFloat_AsDouble(o);
}

//==================================================================================================================================================================
// String

PCSTR String::value(PyObject* o) throw()
{
	if (!o)
		return null;
	else if (o == Py_None)
		return "";
	else
		return PyString_AsString(o);
}

string String::newobj(PyObject* o) throw(...)
{
	if (!o)
		return null;
	if (PyString_Check(o))
		return string((String*)o);
	if (o == Py_None)
		return string::from(PyString_InternFromString(""));
	if (PyObject* r = PyObject_Str(o))
		return string::from(r);
	Exception::raise();
}

string String::newobj(PCSTR value) throw()
{
	if (empty(value))
		return string::from(PyString_InternFromString(""));
	return string::from(PyString_FromString(value));
}

string String::newobj(PCSTR value, ssize_t size) throw()
{
	if (empty(value))
		return string::from(PyString_InternFromString(""));
	return string::from(PyString_FromStringAndSize(value, size));
}

string String::newobj(PSTR* value, ssize_t size) throw()
{
	PyObject* r = PyString_FromStringAndSize("", size);
	*value = PyString_AS_STRING(r);
	return string::from(r);
}

string String::newobj(PCWSTR value) throw()
{
	if (empty(value))
		return string::from(PyString_InternFromString(""));
	size_t ln = len(value) * 2 + 1;
	PSTR s = (PSTR) alloca(ln);
	WideCharToMultiByte(CP_ACP, 0, value, -1, s, (int) ln, null, null);
	return string::from(PyString_FromString(s));
}

//==================================================================================================================================================================
// Unicode

PCWSTR Unicode::value(PyObject* o) throw()
{
	if (!o)
		return null;
	else if (o == Py_None)
		return L"";
	else
		return PyUnicode_AsUnicode(o);
}

unicode Unicode::newobj(PCWSTR value, ssize_t size) throw()
{
	PyObject* r = PyUnicode_FromWideChar(value, size);
	return unicode::from(r);
}

unicode Unicode::newobj(PWSTR* value, ssize_t size) throw()
{
	PyObject* r = PyUnicode_FromWideChar(L"", size);
	*value = PyUnicode_AS_UNICODE(r);
	return unicode::from(r);
}

unicode Unicode::newobj(PyObject* o) throw(...)
{
	if (!o)
		return null;
	if (PyUnicode_Check(o))
		return unicode((Unicode*)o);
	if (o == Py_None)
		return unicode::from(PyUnicode_FromUnicode(L"", 0));
	PyObject* r = PyObject_Unicode(o);
	if (!r)
	{
		PyErr_Clear();
		if (PyString_Check(o))
		{
			r = PyUnicode_FromObject(o);
		}
		else if (PyObject* a = PyObject_Str(o))
		{
			r = PyUnicode_FromObject(a);
			Py_DECREF(a);
		}
		if (!r)
			Exception::raise();
	}
	return unicode::from(r);
}

unicode Unicode::newobj(PyObject* o, PCSTR encoding) throw(...)
{
	if (PyUnicode_Check(o))
		return unicode(static_cast<Unicode*>(o));
	PyObject* r = PyUnicode_FromEncodedObject(o, encoding, "strict");
	if (!r)
		Exception::raise();
	return unicode::from(r);
}

unicode Unicode::newobj(Unicode* left, Unicode* right)
{
	if (PyObject* r = PyUnicode_Concat(left, right))
		return unicode::from(r);
	Exception::raise();
}

ssize_t Unicode::find(Unicode* substr, ssize_t start, ssize_t end, int direction)
{
	return PyUnicode_Find(this, substr, start, end ,direction);
}

ssize_t Unicode::count(Unicode* substr, ssize_t start, ssize_t end)
{
	return PyUnicode_Count(this, substr, start, end);
}

unicode Unicode::replace(Unicode* from, Unicode* to, ssize_t count)
{
	if (PyObject* r = PyUnicode_Replace(this, from, to, count))
		return unicode::from(r);
	Exception::raise();
}

//==================================================================================================================================================================
// Iterator

iterator Iterator::newobj(PyObject* o) throw(...)
{
	if (!o)
		throw ValueError("Iterator::newobj()");
	if (PyObject* i = PyObject_GetIter(o))
		return iterator::from(i);
	else
		Exception::raise();
}

iterator Iterator::newobj_nothrow(PyObject* o) throw()
{
	if (!o)
		return null;
	if (PyObject* i = PyObject_GetIter(o))
		return iterator::from(i);
	PyErr_Clear();
	return null;
}

object Iterator::next() throw(...)
{
	if (PyObject* r = PyIter_Next(this))
		return object::from(r);
	// error or finished
	if (PyErr_Occurred())
		Exception::raise();
	else
		return null;
}

//==================================================================================================================================================================
// Array

array Array::newobj(PyObject* o) throw(...)
{
	if (!o)
		throw ValueError("Array::newobj()");
	if (PyObject* i = PySequence_Fast(o, "Array::newobj()"))
		return array::from(i);
	else
		Exception::raise();
}

//==================================================================================================================================================================
// Tuple

tuple Tuple::newobj(PCSTR format, va_list va) throw(...)
{
	if (empty(format))
		return Tuple::newobj(0);
	PyObject* r = Py_VaBuildValue(const_cast<PSTR>(format), va);
	if (!r)
		Exception::raise();
	if (!PyTuple_Check(r))
	{
		tuple t = Tuple::newobj(1);
		t->attach(0, r);
		r = t.detach();
	}
	return tuple::from(r);
}

tuple Tuple::newobj(PyObject* o) throw(...)
{
	if (len(o) <= 0)
		return newobj(0);
	if (PyObject* r = PySequence_Tuple(o))
		return tuple::from(r);
	Exception::raise();
}

tuple Tuple::newobj(List* o) throw()
{
	return o ? tuple::from(PyList_AsTuple(o)) : null;
}

tuple Tuple::newobj(POINT pt) throw()
{
	PyObject* r = PyTuple_New(2);
	PyTuple_SET_ITEM(r, 0, PyInt_FromLong(pt.x));
	PyTuple_SET_ITEM(r, 1, PyInt_FromLong(pt.y));
	return tuple::from(r);
}

//==================================================================================================================================================================
// List

list List::newobj(PyObject* o) throw(...)
{
	if (len(o) <= 0)
		return newobj(0);
	if (PyObject* r = PySequence_List(o))
		return list::from(r);
	Exception::raise();
}

//==================================================================================================================================================================
// Module

ref<Module> Module::newobj(PCSTR name, PCSTR doc, const PyMethodDef methods[]) throw()
{
	static PyMethodDef EMPTY_METHODS[] = { { null } };

	if (!methods)
		methods = EMPTY_METHODS;
	PyObject* r = Py_InitModule3(const_cast<PSTR>(name), const_cast<PyMethodDef*>(methods), const_cast<PSTR>(doc));
	return ref<Module>::from(r);
}

ref<Module> Module::import(PCSTR name)
{
	PyObject* r = PyImport_ImportModule(const_cast<PSTR>(name));
	return ref<Module>::from(r);
}

object brownie::getattr(PyObject* self, PCSTR name) throw()
{
	ASSERT(self);
	PyObject* r = PyObject_GetAttrString(self, const_cast<PSTR>(name));
	if (!r)
		PyErr_Clear();
	return object::from(r);
}

void brownie::delattr(PyObject* self, PCSTR name) throw()
{
	ASSERT(self);
	if (PyObject_DelAttrString(self, name) < 0)
		PyErr_Clear();
}

object brownie::getitem(PyObject* self, ssize_t index) throw(...)
{
	ASSERT(self);
	PyObject* r = PySequence_GetItem(self, index);
	if (!r)
		Exception::raise();
	return object::from(r);
}

void brownie::setitem(PyObject* self, ssize_t index, PyObject* value) throw(...)
{
	ASSERT(self);
	if (PySequence_SetItem(self, index, value) < 0)
		Exception::raise();
}

object brownie::getitem(Array* self, ssize_t index) throw()
{
	ASSERT(self);
	ASSERT(index < len(self));
	PyObject* r = PySequence_Fast_GET_ITEM(self, index);
	return object(r);
}

object brownie::getitem(Tuple* self, ssize_t index) throw()
{
	ASSERT(self);
	ASSERT(index < len(self));
	PyObject* r = PyTuple_GET_ITEM(self, index);
	return object(r);
}

void brownie::setitem(Tuple* self, ssize_t index, PyObject* value) throw()
{
	ASSERT(self);
	ASSERT(0 <= index);
	ASSERT(index < len(self));
	if (!value)
		value = Py_None;
	PyTuple_SET_ITEM(self, index, value);
	incref(value);
}

object brownie::getitem(List* self, ssize_t index) throw()
{
	ASSERT(self);
	ASSERT(index < len(self));
	PyObject* r = PyList_GET_ITEM(self, index);
	return object(r);
}

void brownie::setitem(List* self, ssize_t index, PyObject* value) throw()
{
	ASSERT(self);
	ASSERT(0 <= index);
	ASSERT(index < len(self));
	if (!value)
		value = Py_None;
	PyList_SET_ITEM(self, index, value);
	incref(value);
}

void brownie::setitem(Dict* self, PCSTR name, PyObject* value) throw(...)
{
	ASSERT(self);
	if (!value)
		value = Py_None;
	if (PyDict_SetItemString(self, name, value) < 0)
		Exception::raise();
}

void brownie::detitem(Dict* self, PCSTR name) throw()
{
	ASSERT(self);
	if (PyDict_DelItemString(self, name) < 0)
		PyErr_Clear();
}

//==================================================================================================================================================================

object brownie::apply(PyObject* fn, Tuple* args, Dict* kw) throw(...)
{
	PyObject* r;
	if (args)
		r = PyObject_Call(fn, args, kw);
	else if (kw)
		r = PyObject_Call(fn, Tuple::newobj(0), kw);
	else
		r = PyObject_CallObject(fn, null);
	if (!r)
		Exception::raise();
	return object::from(r);
}

namespace
{
	object apply_va(PyObject* fn, PCSTR format, va_list va) throw(...)
	{
		tuple args = Tuple::newobj(format, va);
		PyObject* r = PyObject_Call(fn, args, null);
		if (!r)
			Exception::raise();
		return object::from(r);
	}

	void event_va(PyObject* self, PCSTR name, PCSTR format, va_list va)
	{
		try
		{
			if (object m = getattr(self, name))
				apply_va(m, format, va);
		}
		catch (Exception&)
		{
			Exception::clear();
		}
	}
}

object brownie::apply(PyObject* fn, PCSTR format, ...) throw(...)
{
	va_list va;
	va_start(va, format);
	try
	{
		object r = apply_va(fn, format, va);
		va_end(va);
		return r;
	}
	catch (...)
	{
		va_end(va);
		throw;
	}
}

//==================================================================================================================================================================

void Object::Event(PCSTR name, PCSTR format, ...) throw()
{
	py::Lock	lock;
	va_list		va;
	va_start(va, format);
	event_va(self, name, format, va);
	va_end(va);
}

void Object::event(PCSTR name, PCSTR format, ...) throw()
{
	va_list		va;
	va_start(va, format);
	event_va(self, name, format, va);
	va_end(va);
}

//==================================================================================================================================================================

void Module::def(PCSTR name, int value) throw()
{
	PyModule_AddIntConstant(this, const_cast<PSTR>(name), value);
}

void Module::def(PCSTR name, PCSTR value) throw()
{
	PyModule_AddStringConstant(this, const_cast<PSTR>(name), value);
}

void Module::def(PCSTR name, PCWSTR value) throw()
{
	PyModule_AddObject(this, const_cast<PSTR>(name), Unicode::newobj(value).detach());
}

void Module::def(PCSTR name, PyObject* value) throw()
{
	ASSERT(value);
	Py_INCREF(value);
	PyModule_AddObject(this, const_cast<PSTR>(name), value);
}

void Module::def(PyTypeObject* type)
{
	ASSERT(type);
	PyType_Ready(type);
	Py_INCREF(type);
	// __name__  module.class `ł邱Ƃ肷
	PCSTR name = type->tp_name;
	for (PCSTR i = name; *i; ++i)
	{
		if (*i == '.')
			name = i+1;
	}
	PyModule_AddObject(this, name, (PyObject*)type);
}

//==================================================================================================================================================================

void brownie::parse(Tuple* args, PCSTR format, ...) throw(...)
{
	va_list va;
	va_start(va, format);
	int r = PyArg_VaParse(args, const_cast<PSTR>(format), va);
	va_end(va);
	if (!r)
		Exception::raise();
}

void brownie::parse(Tuple* args, const PCSTR formats[], ...) throw(...)
{
	va_list va;
	va_start(va, formats);
	if (PyArg_VaParse(args, const_cast<PSTR>(formats[0]), va))
	{
		va_end(va);
		return;
	}
	for (size_t i = 1; formats[i]; ++i)
	{
		PyErr_Clear();
		if (PyArg_VaParse(args, const_cast<PSTR>(formats[i]), va))
		{
			va_end(va);
			return;
		}
	}
	Exception::raise();
}

void brownie::parse(Tuple* args, Dict* kw, PCSTR format, const PCSTR kwds[], ...) throw(...)
{
	va_list va;
	va_start(va, kwds);
	int r = PyArg_VaParseTupleAndKeywords(args, kw,
		const_cast<PSTR>(format), const_cast<char**>(kwds), va);
	va_end(va);
	if (!r)
		Exception::raise();
}

object brownie::build(PCSTR format, ...) throw(...)
{
	va_list va;
	va_start(va, format);
	PyObject* r = Py_VaBuildValue(const_cast<PSTR>(format), va);
	va_end(va);
	if (!r)
		Exception::raise();
	return object::from(r);
}

void brownie::abstract(PyTypeObject* type, PyObject* o) throw(...)
{
	if (o->ob_type == type)
		throw TypeError(_("Error.AbstractClass"));
}
