// stream.cpp

#include "stdafx.h"
#include "stream.hpp"
#include "buffer.hpp"

using namespace brownie;

ref<IStream> brownie::StreamCreate(HGLOBAL* mem) throw(HRESULT)
{
	ASSERT(mem);
	*mem = null;
	HRESULT hr;
	HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, 0);
	if (!hGlobal)
		throw HRESULT_FROM_WIN32(GetLastError());
	ref<IStream> stream;
	if SUCCEEDED(hr = ::CreateStreamOnHGlobal(hGlobal, false, &stream))
	{
		*mem = hGlobal;
		return stream;
	}
	::GlobalFree(hGlobal);
	throw hr;
}

ref<IStream> brownie::StreamCreate(PCWSTR filename, DWORD flags) throw(HRESULT)
{
	ref<IStream> stream;
	HRESULT hr;
	if FAILED(hr = ::SHCreateStreamOnFileW(filename, flags, &stream))
		throw hr;
	return stream;
}

ref<IStream> brownie::StreamCreate(PCSTR filename, DWORD flags) throw(HRESULT)
{
	ref<IStream> stream;
	HRESULT hr;
	if FAILED(hr = ::SHCreateStreamOnFileA(filename, flags, &stream))
		throw hr;
	return stream;
}

void brownie::StreamRead(IStream* stream, void* data, size_t size) throw(HRESULT)
{
	ULONG done = 0;
	HRESULT hr = stream->Read(data, (ULONG)size, &done);
	if FAILED(hr)
		throw hr;
	if (done != (ULONG)size)
		throw STG_E_READFAULT;
}

void brownie::StreamWrite(IStream* stream, const void* data, size_t size) throw(HRESULT)
{
	ULONG done = 0;
	HRESULT hr = stream->Write(data, (ULONG)size, &done);
	if FAILED(hr)
		throw hr;
	if (done != (ULONG)size)
		throw STG_E_MEDIUMFULL;
}

UINT64 brownie::StreamSeek(IStream* stream, INT64 off, DWORD whence)
{
	ULARGE_INTEGER done;
	HRESULT hr = stream->Seek((LARGE_INTEGER&)off, whence, &done);
	if FAILED(hr)
		throw hr;
	return done.QuadPart;
}

IStream& operator >> (IStream& stream, PCWSTR& v) throw(HRESULT)
{
	UINT32 ln;
	stream >> ln;
	if (ln > 0)
	{
		PWSTR s = new WCHAR[ln+1];
		try
		{
			StreamReadT(&stream, s, ln);
			s[ln] = L'\0';
			v = s;
		}
		catch (...)
		{
			delete[] s;
			throw;
		}
	}
	else
		v = null;
	return stream;
}

IStream& operator << (IStream& stream, PCWSTR v) throw(HRESULT)
{
	UINT32 ln = (UINT32)len(v);
	stream << ln;
	StreamWriteT(&stream, v, ln);
	return stream;
}

//==================================================================================================================================================================

BufferStream::BufferStream(BufferT<BYTE>& buffer) : m_buffer(buffer), m_pos(0)
{
}

HRESULT __stdcall BufferStream::Read(void* data, ULONG size, ULONG* done)
{
	size_t r = m_buffer.read(m_pos, (BYTE*)data, size);
	m_pos += r;
	if (done)
		*done = (ULONG)r;
	return (r > 0 ? S_OK : S_FALSE);
}

HRESULT __stdcall BufferStream::Write(const void* data, ULONG size, ULONG* done)
{
	m_buffer.write(m_pos, (const BYTE*)data, size);
	m_pos += size;
	if (done)
		*done = size;
	return S_OK;
}

HRESULT __stdcall BufferStream::SetSize(ULARGE_INTEGER sz)
{
	m_buffer.resize((size_t)sz.QuadPart);
	return S_OK;
}

DWORD BufferStream::GetAccessMode() const		{ return STGM_READWRITE; }
UINT64 BufferStream::GetSize() const			{ return m_buffer.size(); }
UINT64 BufferStream::GetPosition() const		{ return m_pos; }

HRESULT BufferStream::SeekTo(UINT64 pos)
{
	if (pos > (size_t)-1)
		return E_FAIL;
	m_pos = (size_t)pos;
	return S_OK;
}
