// system.cpp

#include "stdafx.h"
#include "alert.hpp"
#include "system.hpp"
#include "thread.hpp"
#include "window.hpp"

using namespace brownie;

namespace
{
	// GlobalKeymap
	static PyObject*			m_keymap;
	static std::vector<ATOM>	m_hotkeys;

	// Threads
	const DWORD		MAX_THREADS = MAXIMUM_WAIT_OBJECTS - 1;
	DWORD			theSystemThreadID;
	static DWORD	theThreadCount;
	static DWORD	theThreadIDs[MAX_THREADS];
	static HANDLE	theThreadHandles[MAX_THREADS];
	static HWND		theThreadWindows[MAX_THREADS];
	static HWND		theAnotherWindow;

	template < typename T >
	inline void RemoveAt(T* array, DWORD count, DWORD index)
	{
		memmove(array + index, array + index + 1, sizeof(T) * (count - index));
	}

	bool IsDesktopWindow(HWND hwnd)
	{
		WCHAR cls[MAX_PATH];
		::GetClassName(hwnd, cls, MAX_PATH);
		return wcscmp(cls, L"Shell_TrayWnd") == 0 || wcscmp(cls, L"Progman") == 0;
	}

#ifdef _DEBUG
	bool	THREADS_LOCKED;
	void	THREADS_ACCESS()	{ ASSERT(!THREADS_LOCKED); }
	void	THREADS_LOCK()		{ ASSERT(!THREADS_LOCKED); THREADS_LOCKED = true; }
	void	THREADS_UNLOCK()	{ ASSERT(THREADS_LOCKED); THREADS_LOCKED = false; }
#else
#		define	THREADS_ACCESS	__noop
#		define	THREADS_LOCK	__noop
#		define	THREADS_UNLOCK	__noop
#endif
}

HINSTANCE System::Module;
HWND System::Window;

//==================================================================================================================================================================

void System::Init(WNDPROC wndproc)
{
	theSystemThreadID = ::GetCurrentThreadId();

#ifdef _DEBUG
	THREADS_LOCKED = false;
#endif

	OnActivate(0, 0);

	ASSERT(!System::Window);
	WNDCLASSEX wc = { sizeof(WNDCLASSEX) };
	wc.lpfnWndProc = wndproc;
	wc.hInstance = ::GetModuleHandle(null);
	wc.lpszClassName = APPNAME;
	::RegisterClassEx(&wc);
	System::Window = ::CreateWindowEx(0, wc.lpszClassName, null, WS_POPUP, 0, 0, 0, 0, HWND_MESSAGE, null, wc.hInstance, null);
}

void System::Term()
{
	if (::IsWindow(System::Window))
		::DestroyWindow(System::Window);
	System::Window = null;
}

//==================================================================================================================================================================

DWORD System::Fork(ThreadProc proc, PyObject* fn)
{
	THREADS_ACCESS();
	ASSERT(::GetCurrentThreadId() == theSystemThreadID);

	PCWSTR error;
	unsigned id;
	if (theThreadCount >= MAX_THREADS)
	{
		error = _W("Error.NoMoreWindow");
	}
	else if (HANDLE handle = (HANDLE)_beginthreadex(null, 0, proc, fn, CREATE_SUSPENDED, &id))
	{
		theThreadIDs[theThreadCount] = id;
		theThreadHandles[theThreadCount] = handle;
		theThreadWindows[theThreadCount] = null;
		++theThreadCount;
		::ResumeThread(handle);
		return id;
	}
	else
	{
		error = _W("Error.NoMoreThread");
	}

	// on error
	Alert(error, null, MB_OK | MB_ICONERROR);
	py::Lock lock;
	decref(fn);
	return 0;
}

void System::ActivateWindow(int dir)
{
	if (theThreadCount <= 0)
		return;

	int current = -1;
	if (HWND hwnd = ::GetAncestor(::GetForegroundWindow(), GA_ROOT))
		for (DWORD i = 0; i < theThreadCount; ++i)
			if (theThreadWindows[i] == hwnd)
				current = i;

	HWND hwnd;
	if (current >= 0 && (current + dir < 0 || current + dir >= (int)theThreadCount))
		hwnd = theAnotherWindow;
	else if (current >= 0)
		hwnd = theThreadWindows[(current + theThreadCount + dir) % theThreadCount];
	else
		hwnd = theThreadWindows[dir > 0 ? 0: theThreadCount-1];

	if (::IsWindow(hwnd))
	{
		::AttachThreadInput(::GetCurrentThreadId(), ::GetWindowThreadProcessId(hwnd, NULL), true);
		if (::IsIconic(hwnd))
			::ShowWindow(hwnd, SW_RESTORE);
		if (!::IsWindowVisible(hwnd))
			::ShowWindow(hwnd, SW_SHOW);
		::BringWindowToTop(hwnd);
		::SetActiveWindow(hwnd);
		::SetForegroundWindow(hwnd);
	}
	else if (current >= 0)
	{
		::ShowWindow(theThreadWindows[current], SW_MINIMIZE);
	}
}

void System::Broadcast(UINT msg, WPARAM wParam, LPARAM lParam) throw()
{
	THREADS_ACCESS();
	switch (msg)
	{
	case WM_DESTROY:
	case WM_NCDESTROY:
		for (DWORD i = 0; i < theThreadCount; ++i)
		{
			if (::IsWindow(theThreadWindows[i]))
				::DestroyWindow(theThreadWindows[i]);
		}
		break;
	default:
		for (DWORD i = 0; i < theThreadCount; ++i)
		{
			if (::IsWindow(theThreadWindows[i]))
				::PostMessage(theThreadWindows[i], msg, wParam, lParam);
		}
		break;
	}
}

HWND System::GetWindow() throw()
{
	if (Thread* thread = Thread::Get())
		return thread->get_Hwnd();
	return null;
}

void System::MessageLoop(HWND hwnd) throw()
{
	Thread* thread = Thread::Get();
	if (!thread)
		return;

	DWORD	tid = ::GetCurrentThreadId();
	for (DWORD i = 0; i < theThreadCount; ++i)
	{
		if (theThreadIDs[i] == tid)
		{
			theThreadWindows[i] = hwnd;
			break;
		}
	}
	thread->MessageLoop(hwnd);

	BoardcastMessage(WM_CLOSE);		// ܂ WM_CLOSE ł̏I₢킹B
	PumpMessage();
	BoardcastMessage(WM_DESTROY);	// łcĂ狭IɔjB
	PumpMessage();
}

void System::Run()
{
	do
	{
		while (theThreadCount > 0)
		{
			if (!::IsWindow(System::Window))
			{	// TuXbhIOɁAVXeEBhEĂiD܂Ԃł͂Ȃj
				Broadcast(WM_DESTROY);
			}

			THREADS_LOCK();
			DWORD n = ::MsgWaitForMultipleObjects(theThreadCount, theThreadHandles, false, INFINITE, QS_ALLINPUT);
			if (WAIT_OBJECT_0 <= n && n < WAIT_OBJECT_0 + theThreadCount)
			{	// thread exit
				DWORD index = n - WAIT_OBJECT_0;
				::CloseHandle(theThreadHandles[index]);
				RemoveAt(theThreadIDs, theThreadCount, index);
				RemoveAt(theThreadHandles, theThreadCount, index);
				RemoveAt(theThreadWindows, theThreadCount, index);
				--theThreadCount;
			}
			THREADS_UNLOCK();

			PumpMessage();
		}

		if (::IsWindow(System::Window))
		{	// TuXbhׂďIAVXeEBhÊ݂cĂ
			PumpMessage(); // cĂ郁bZ[WŐVȃTuXbh\B
			if (theThreadCount == 0)
				::DestroyWindow(System::Window);
		}
	} while (::IsWindow(System::Window));
}

//==================================================================================================================================================================

void System::activate(Tuple* args) throw()
{
	if (System::Window)
	{
		int dir = 1;
		parse(static_cast<Tuple*>(args), "|i", &dir);
		WPARAM wParam = (dir > 0 ? +1 : -1);
		::PostMessage(System::Window, WM_ACTIVATEWINDOW, wParam, 0);
	}
}

object System::window()
{
	return object(Window::From(System::GetWindow())->self);
}

void System::fork(PyObject* o)
{
	if (!is_callable(o))
		throw ValueError("The argument passed to System.fork() must be callable");
	// VXeXbhɃVACYB
	if (System::Window)
	{
		incref(o);
		::PostMessage(System::Window, WM_FORK, 0, (LPARAM)o);
	}
}

void System::restart() throw()
{
	if (System::Window)
		::PostMessage(System::Window, WM_RESTART, 0, 0);
}

void System::quit() throw()
{
	if (System::Window)
		::PostMessage(System::Window, WM_CLOSE, 0, 0);
}

namespace
{
	static UINT MK2MOD(UINT mods)
	{
		UINT ret = 0;
		if(mods & MK_CONTROL)	ret |= MOD_CONTROL;
		if(mods & MK_SHIFT)		ret |= MOD_SHIFT;
		if(mods & MK_ALTKEY)	ret |= MOD_ALT;
		if(mods & MK_WINKEY)	ret |= MOD_WIN;
		return ret;
	}

	static UINT16 MOD2MK(UINT mods)
	{
		UINT16 ret = 0;
		if(mods & MOD_CONTROL)	ret |= MK_CONTROL;
		if(mods & MOD_SHIFT)	ret |= MK_SHIFT;
		if(mods & MOD_ALT)		ret |= MK_ALTKEY;
		if(mods & MOD_WIN)		ret |= MK_WINKEY;
		return ret;
	}
}

void System::OnActivate(WPARAM wParam, LPARAM lParam)
{
	HWND	another = (HWND)lParam;
	if (!another)
		another = ::GetForegroundWindow();
	DWORD pid;
	if (another && ::IsWindowEnabled(another) &&
		GetWindowThreadProcessId(another, &pid) && pid != ::GetCurrentProcessId() &&
		!IsDesktopWindow(another))
		theAnotherWindow = another;
}

void System::OnHotKey(WPARAM wParam, LPARAM lParam) throw()
{
	if (!m_keymap)
		return;
	UINT8	vkey = (UINT8)HIWORD(lParam);
	UINT16	mod = MOD2MK(LOWORD(lParam));
	{
		py::Lock lock;
		try
		{
			apply(m_keymap, "Oii", Py_None, vkey, mod);
		}
		catch (Exception&)
		{
			Exception::notify();
		}
	}
}

extern void Menu_Reset();

namespace
{
	const UINT	TIMER_SETTING = 1000;
	const UINT	DELAY_SETTING = 500; // ms
	void CALLBACK OnSystemSetting(HWND hwnd, UINT msg, UINT_PTR idEvent, DWORD dwTime)
	{
		::KillTimer(hwnd, idEvent);
		Menu_Reset();
	}
}

void System::OnSettingChange(WPARAM wParam, LPARAM lParam)
{
	// 񑗐M邱Ƃ邽߁AЂƂɂ܂Ƃ߂ړIŒxB
	::SetTimer(System::Window, TIMER_SETTING, DELAY_SETTING, OnSystemSetting);
}

object System::get_keymap()
{
	return object(m_keymap);
}

void System::set_keymap(PyObject* o)
{
	if (m_keymap == o)
		return;

	// clear
	size_t count = m_hotkeys.size();
	for(size_t i = 0; i < count; ++i)
	{
		ATOM atom = m_hotkeys[i];
		::UnregisterHotKey(NULL, (int)atom);
		::GlobalDeleteAtom(atom);
	}
	m_hotkeys.clear();
	decref(m_keymap);

	// set
	incref(o);
	m_keymap = o;

	if (m_keymap && System::Window)
	{
		ssize_t ln = len(PATH_ROOT);
		if (ln > 230)
			ln -= 230; // ATOM ɂ255邽߁Aꍇ͍B
		foreach (object i, m_keymap)
		{
			try
			{
				UINT8	vkey = getitem<UINT8>(i, 0);
				UINT16	mod  = getitem<UINT16>(i, 1);
				WCHAR	name[256];
				swprintf_s(name, L"%s:%04X:%02X", PATH_ROOT + ln, mod, vkey);
				if (ATOM atom = ::GlobalAddAtom(name))
					if (::RegisterHotKey(System::Window, (int)atom, MK2MOD(mod), vkey))
						m_hotkeys.push_back(atom);
			}
			catch (Exception&)
			{
				Exception::clear();
			}
		}
	}
}

//==================================================================================================================================================================

PYTHON_METHODS(System)
	PYTHON_METHOD(activate)
	PYTHON_METHOD(fork)
	PYTHON_METHOD(restart)
	PYTHON_METHOD(quit)
	PYTHON_METHOD(window)
PYTHON_END()
