// xp_explorer.cpp

#include "stdafx.h"
#include "alert.hpp"
#include "deque.hpp"
#include "hwnd.hpp"
#include "shell.hpp"
#include "unknown.hpp"
#include "xp_class.hpp"

#pragma comment(lib, "winmm")

using namespace brownie;

namespace
{
	//==================================================================================================================================================================

	const size_t MAX_HISTORY = 20;

	class XpExplorerHistory
	{
	private:
		std::deque<ITEMIDLIST*>	m_items;
		int						m_index;

	public:
		XpExplorerHistory() : m_index(0)
		{
		}
		~XpExplorerHistory()
		{
			Clear();
		}
		void Clear(int after = 0) throw()
		{
			if (after >= (int)m_items.size())
				return;
			for (int i = after; i < (int)m_items.size(); ++i)
				ILDelete(m_items[i]);
			m_items.erase(m_items.begin() + after, m_items.end());
			m_index = (int)m_items.size() - 1;;
		}
		void Push(const ITEMIDLIST* item) throw()
		{
			Clear(m_index+1);
			m_items.push_back(ILClone(item));
			if (m_items.size() > MAX_HISTORY)
			{
				ILDelete(m_items.front());
				m_items.pop_front();
			}
			m_index = (int)m_items.size() - 1;
		}
		int NormalizedIndex(int mv) const throw()
		{
			int i = m_index + mv;
			if (i < 0 || i >= (int)m_items.size())
				return -1; // out of range
			return i;
		}
		const ITEMIDLIST* Get(int mv = 0) const throw()
		{
			if (m_items.empty())
				return null;
			int i = NormalizedIndex(mv);
			if (i < 0)
				return null;
			return m_items[i];
		}
		void Seek(int mv) throw()
		{
			m_index = NormalizedIndex(mv);
		}
	};

	//==================================================================================================================================================================

	class XpExplorerBrowser :
		public Unknown<implements<IExplorerBrowserXp, IExplorerBrowser, IObjectWithSite, IOleWindow, IShellBrowser, ICommDlgBrowser, ICommDlgBrowser2/*, ICommDlgBrowser3*/> >,
		public Hwnd
	{
	private:
		ref<IServiceProvider>			m_site;
		refs<IExplorerBrowserEvents>	m_handlers;
		XpExplorerHistory				m_history;
		ref<IShellFolder>				m_folder;
		ref<IShellView>					m_view;
		Hwnd							m_defview;
		EXPLORER_BROWSER_OPTIONS		m_options;
		FOLDERSETTINGS					m_settings;
		RECT							m_bounds;
		ref<IStream>					m_streamViewState;

	public:
		XpExplorerBrowser()	{}

	public: // IExplorerBrowserXp
		STDMETHODIMP GetHistory(IShellItem** pp, int mv);
		STDMETHODIMP EnumHistory(IEnumShellItems** pp, int dir);
		STDMETHODIMP BrowseToHistory(int mv);
		STDMETHODIMP SaveViewState()
		{
			if (const ITEMIDLIST* current = Current())
				return FolderStateSave(current, m_view, &m_streamViewState);
			else
				return E_FAIL;
		}

	public: // IExplorerBrowser
        STDMETHODIMP Initialize(HWND hwndParent, const RECT* rc, const FOLDERSETTINGS* fs)
		{
			m_hwnd = hwndParent;
			if (rc)
				m_bounds = *rc;
			else
				SetRectEmpty(&m_bounds);
			if (fs)
				m_settings = *fs;
			return S_OK;
		}
        
        STDMETHODIMP Destroy()
		{
			SaveViewState();
			m_site = null;
			m_handlers.clear();
			m_defview = null;
			SetShellView(null);
			return S_OK;
		}
        
        STDMETHODIMP SetRect(HDWP* phdwp, RECT rc)
		{
			m_bounds = rc;
			if (m_defview)
			{
				if (phdwp)
					*phdwp = ::DeferWindowPos(*phdwp, m_defview, null, RECT_XYWH(rc), SWP_NOZORDER | SWP_NOACTIVATE);
				else
					m_defview.SetPos(RECT_XYWH(rc));
			}
			return S_OK;
		}
        
        STDMETHODIMP SetPropertyBag(PCWSTR pszPropertyBag)
		{
			if (wcscmp(pszPropertyBag, PATH_FOLDER_STATE) != 0)
				return E_UNEXPECTED;
			return S_OK;
		}

        STDMETHODIMP SetEmptyText(PCWSTR pszEmptryText)
		{
			return E_NOTIMPL;
		}
        
        STDMETHODIMP SetFolderSettings(const FOLDERSETTINGS *pfs)
		{
			// TODO: handle pfs->fFlags, maybe requires re-create view.
			m_settings = *pfs;
			ref<IFolderView> view;
			if SUCCEEDED(m_view.copyto(&view))
				return view->SetCurrentViewMode(pfs->ViewMode);
			return S_OK;
		}
        
        STDMETHODIMP Advise(IExplorerBrowserEvents* handler, DWORD* cookie);
        STDMETHODIMP Unadvise(DWORD cookie);
        
        STDMETHODIMP SetOptions(EXPLORER_BROWSER_OPTIONS dwFlag)
		{
			m_options = dwFlag;
			return S_OK;
		}
        
        STDMETHODIMP GetOptions(EXPLORER_BROWSER_OPTIONS *pdwFlag)
		{
			*pdwFlag = m_options;
			return S_OK;
		}
        
        STDMETHODIMP BrowseToIDList(const ITEMIDLIST* pidl, UINT uFlags)
		{
			if (uFlags & SBSP_NAVIGATEBACK)
				return BrowseToHistory(-1);
			else if (uFlags & SBSP_NAVIGATEFORWARD)
				return BrowseToHistory(+1);
			else if (uFlags & SBSP_PARENT)
				return BrowseToParent();
			else if (uFlags & SBSP_RELATIVE)
				return BrowseToRelative(pidl);
			else
				return BrowseToAbsolute(pidl, true);
		}
        
        STDMETHODIMP BrowseToObject(IUnknown *punk, UINT uFlags)
		{
			if (uFlags & SBSP_NAVIGATEBACK)
				return BrowseToHistory(-1);
			else if (uFlags & SBSP_NAVIGATEFORWARD)
				return BrowseToHistory(+1);
			else if (uFlags & SBSP_PARENT)
				return BrowseToParent();
			else
				return E_NOTIMPL;
		}
        
        STDMETHODIMP FillFromObject(IUnknown *punk, EXPLORER_BROWSER_FILL_FLAGS dwFlags)
		{
			return E_NOTIMPL;
		}
        
        STDMETHODIMP RemoveAll()
		{
			return E_NOTIMPL;
		}
       
        STDMETHODIMP GetCurrentView(REFIID riid, void **ppv)
		{
			if (!m_view)
			{
				*ppv = null;
				return E_UNEXPECTED;
			}
			return m_view->QueryInterface(riid, ppv);
		}

	public: // IObjectWithSite
		STDMETHODIMP SetSite(IUnknown* site);
	    STDMETHODIMP GetSite(REFIID iid, void** pp);

		void SetShellView(IShellView* view)
		{
			if (m_view == view)
				return;
			if (m_view)
			{
				ref<IShellView> view(m_view);
				m_view = null;
				m_folder = null;
				view->UIActivate(SVUIA_DEACTIVATE);
				view->DestroyViewWindow();
			}
			m_view = view;
		}

	public: // IOleWindow
		STDMETHODIMP GetWindow(HWND* pHwnd);
		STDMETHODIMP ContextSensitiveHelp(BOOL fEnterMode)	{ return E_NOTIMPL; }

	public: // ICommDlgBrowser
		STDMETHODIMP OnDefaultCommand(IShellView* view);
		STDMETHODIMP OnStateChange(IShellView* view, ULONG uChange);
		STDMETHODIMP IncludeObject(IShellView* view, const ITEMIDLIST* pidl);

	public: // ICommDlgBrowser2
		STDMETHODIMP Notify(IShellView* view, DWORD dwNotifyType)	{ return S_OK; }
		STDMETHODIMP GetDefaultMenuText(IShellView* view, WCHAR* pszText, int cchMax);
		STDMETHODIMP GetViewFlags(DWORD* pdwFlags);

	//public:	// ICommDlgBrowser3
	//	STDMETHODIMP OnColumnClicked(IShellView* view, int iColumn)				{ return E_NOTIMPL; }
	//	STDMETHODIMP GetCurrentFilter(PWSTR pszFileSpec, int cchFileSpec)		{ return E_NOTIMPL; }
	//	STDMETHODIMP OnPreViewCreated(IShellView* view)							{ return E_NOTIMPL; }

	public: // IShellBrowser
		STDMETHODIMP InsertMenusSB(HMENU hmenuShared, LPOLEMENUGROUPWIDTHS lpMenuWidths)				{ return E_NOTIMPL; }
		STDMETHODIMP SetMenuSB(HMENU hmenuShared, HOLEMENU holemenuReserved,HWND hwndActiveObject)		{ return E_NOTIMPL; }
		STDMETHODIMP RemoveMenusSB(HMENU hmenuShared)													{ return E_NOTIMPL; }
		STDMETHODIMP SetStatusTextSB(LPCOLESTR text)													{ return E_NOTIMPL; }
		STDMETHODIMP EnableModelessSB(BOOL fEnable)														{ return E_NOTIMPL; }
		STDMETHODIMP SetToolbarItems(LPTBBUTTON lpButtons, UINT nButtons,UINT uFlags)					{ return E_NOTIMPL; }
		STDMETHODIMP TranslateAcceleratorSB(LPMSG lpmsg, WORD wID)										{ return E_NOTIMPL; }
		STDMETHODIMP OnViewWindowActive(IShellView* pShellView)											{ return E_NOTIMPL; }
		STDMETHODIMP GetControlWindow(UINT id, HWND* pHwnd);
		STDMETHODIMP SendControlMsg(UINT id, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT* lResult);
		STDMETHODIMP GetViewStateStream(DWORD stgm, IStream** pp);
		STDMETHODIMP QueryActiveShellView(IShellView** pp);
		STDMETHODIMP BrowseObject(const ITEMIDLIST* pidl, UINT wFlags);

	private:
		HRESULT	QueryService(REFGUID sid, REFINTF pp) const;
		void	InvokeViewCreated(IShellView* view) const;
		void	InvokeNavigationComplete(const ITEMIDLIST* folder) const;
		void	InvokeNavigationFailed(const ITEMIDLIST* folder) const;

	private:
		const ITEMIDLIST* Current() const throw()			{ return m_history.Get(); }
		CIDA* GetContents(STATUS status) const throw();
		HRESULT BrowseToAbsolute(const ITEMIDLIST* item, bool addToHistory) throw();
		HRESULT BrowseToRelative(const ITEMIDLIST* relative) throw();
		HRESULT BrowseToParent() throw();
	};
}

//==================================================================================================================================================================
// IExplorerBrowser

STDMETHODIMP XpExplorerBrowser::SetSite(IUnknown* site)
{
	m_site = null;
	return site ? site->QueryInterface(&m_site) : S_OK;
}

STDMETHODIMP XpExplorerBrowser::GetSite(REFIID iid, void** pp)
{
	return m_site.copyto(iid, pp);
}

STDMETHODIMP XpExplorerBrowser::Advise(IExplorerBrowserEvents* handler, DWORD* pdwCookie)
{
	for (DWORD i = 0; i < m_handlers.size(); ++i)
	{
		if (m_handlers[i])
			continue;
		m_handlers[i] = handler;
		*pdwCookie = i + 1;
		return S_OK;
	}

	m_handlers.push_back(ref<IExplorerBrowserEvents>(handler));
	*pdwCookie = (DWORD)m_handlers.size();
	return S_OK;
}

STDMETHODIMP XpExplorerBrowser::Unadvise(DWORD dwCookie)
{
	if (dwCookie > m_handlers.size())
		return E_INVALIDARG;
	m_handlers[dwCookie - 1] = null;
	return S_OK;
}

//==================================================================================================================================================================
// IOleWindow

STDMETHODIMP XpExplorerBrowser::GetWindow(HWND* pHwnd)
{
	if (m_defview)
	{
		*pHwnd = m_defview.Parent;
		return S_OK;
	}
	ref<IOleWindow> site;
	if SUCCEEDED(QueryService(SID_SExplorerBrowserFrame, &site))
		return site->GetWindow(pHwnd);
	return E_FAIL;
}

//==================================================================================================================================================================
// ICommDlgBrowser

// handle double click and ENTER key if needed
STDMETHODIMP XpExplorerBrowser::OnDefaultCommand(IShellView* view)
{
	ref<ICommDlgBrowser> site;
	if SUCCEEDED(QueryService(SID_SExplorerBrowserFrame, &site))
		return site->OnDefaultCommand(view);
	return S_FALSE;
}

STDMETHODIMP XpExplorerBrowser::OnStateChange(IShellView* view, ULONG uChange)
{	//handle selection, rename, focus if needed
	ref<ICommDlgBrowser> site;
	if SUCCEEDED(QueryService(SID_SExplorerBrowserFrame, &site))
		return site->OnStateChange(view, uChange);
	return E_NOTIMPL;
}

STDMETHODIMP XpExplorerBrowser::IncludeObject(IShellView* view, const ITEMIDLIST* pidl)
{	// \t@C̃tB^OB
	// S_OK: show, S_FALSE: hide
	return S_OK;
}

//==================================================================================================================================================================
// ICommDlgBrowser2

STDMETHODIMP XpExplorerBrowser::GetDefaultMenuText(IShellView* view, WCHAR* pszText, int cchMax)
{
	ref<ICommDlgBrowser2> site;
	if SUCCEEDED(QueryService(SID_SExplorerBrowserFrame, &site))
		return site->GetDefaultMenuText(view, pszText, cchMax);
	return E_NOTIMPL;
}

STDMETHODIMP XpExplorerBrowser::GetViewFlags(DWORD* pdwFlags)
{	// ׂẴt@C\ׂۂ̖₢킹B
	// CDB2GVF_SHOWALLFILES: Bt@C\B
	// 0: Bt@C\ȂB
	ref<ICommDlgBrowser2> site;
	if SUCCEEDED(QueryService(SID_SExplorerBrowserFrame, &site))
		return site->GetViewFlags(pdwFlags);
	*pdwFlags = 0;
	return S_OK;
}

//==================================================================================================================================================================
// ICommDlgBrowser3

//==================================================================================================================================================================
// IShellBrowser

STDMETHODIMP  XpExplorerBrowser::GetControlWindow(UINT id, HWND* pHwnd)
{
	ref<IShellBrowser> site;
	if SUCCEEDED(QueryService(SID_SShellBrowser, &site))
		return site->GetControlWindow(id, pHwnd);
	return E_NOTIMPL;
}

STDMETHODIMP XpExplorerBrowser::SendControlMsg(UINT id, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT* lResult)
{
	ref<IShellBrowser> site;
	if SUCCEEDED(QueryService(SID_SShellBrowser, &site))
		return site->SendControlMsg(id, msg, wParam, lParam, lResult);
	return E_NOTIMPL;
}

STDMETHODIMP XpExplorerBrowser::GetViewStateStream(DWORD stgm, IStream** pp)
{
	if (m_streamViewState)
		return m_streamViewState.copyto(pp);

	TRACE(L"GetViewStateStream() NoStream (%08X)", stgm);
	*pp = null;
	return E_FAIL;
}

STDMETHODIMP XpExplorerBrowser::QueryActiveShellView(IShellView** pp)
{
	return m_view.copyto(pp);
}

STDMETHODIMP XpExplorerBrowser::BrowseObject(const ITEMIDLIST* pidl, UINT wFlags)
{
	return BrowseToIDList(pidl, wFlags);
}

//==================================================================================================================================================================

HRESULT XpExplorerBrowser::QueryService(REFGUID sid, REFINTF pp) const
{
	if (!m_site)
		return E_FAIL;
	return m_site->QueryService(sid, pp.iid, pp.pp);
}

void XpExplorerBrowser::InvokeViewCreated(IShellView* view) const
{
	m_handlers.each(&IExplorerBrowserEvents::OnViewCreated, view);
}

void XpExplorerBrowser::InvokeNavigationComplete(const ITEMIDLIST* folder) const
{
	m_handlers.each(&IExplorerBrowserEvents::OnNavigationComplete, folder);
}

void XpExplorerBrowser::InvokeNavigationFailed(const ITEMIDLIST* folder) const
{
	m_handlers.each(&IExplorerBrowserEvents::OnNavigationFailed, folder);
}

//==================================================================================================================================================================

CIDA* XpExplorerBrowser::GetContents(STATUS status) const throw()
{
	if (!m_view)
		return null;
	int svgio = 0;
	switch (status)
	{
	case NORMAL:
		svgio = SVGIO_ALLVIEW | SVGIO_FLAG_VIEWORDER;
		break;
	case FOCUSED:
		svgio = SVGIO_SELECTION;
		break;
	case SELECTED:
		svgio = SVGIO_SELECTION | SVGIO_FLAG_VIEWORDER;
		break;
	case CHECKED:
		svgio = SVGIO_CHECKED | SVGIO_FLAG_VIEWORDER;
		break;
	default:
		ASSERT(0);
		return null;
	}
	return CIDACreate(m_view, svgio);
}

namespace
{
	void PlayNavigateSound(HWND hwnd)
	{
		if (!empty(PATH_SOUND_NAVIGATION) && ::IsWindowVisible(hwnd))
			Beep(PATH_SOUND_NAVIGATION);
	}
}

HRESULT XpExplorerBrowser::BrowseToAbsolute(const ITEMIDLIST* item, bool addToHistory) throw()
{
	if (!item)
		return E_INVALIDARG;

	if (!::IsWindow(m_hwnd))
		return E_UNEXPECTED;

	const ITEMIDLIST* current = Current();
	ILPtr focus; // qeֈړꍇ́AtH[JXB

	if (current)
	{
		if (ILIsEqual(current, item))
			return S_FALSE; // ړ̕KvȂ
		if (ILIsParent(item, current, FALSE))
		{
			size_t size = ILGetSize(item);
			focus.attach(ILCloneFirst((ITEMIDLIST*)(((BYTE*)current) + size - 2)));
		}
	}

	HRESULT hr;
	ref<IShellFolder> folder;
	ref<IShellView> view;
	HWND hwndDefView = null;

	if (FAILED(hr = ILFolder(&folder, item))
	|| FAILED(hr = folder->CreateViewObject(m_hwnd, IID_IShellView, (void**)&view)))
		return hr;

	bool hasFocus = HasFocus(m_defview);

	if (m_defview)
		m_defview.GetPos(&m_bounds); // Ƀr[쐬ς݂̏ꍇ́ÃTCYgB

	// ȉAGetViewStateStream() Ăяo^C~O̊֌WAώGɂȂB
	// FolderStateSave/Load  stream ́AGetViewStateStream() ŕԂ̂Ɠ̂gƁB

	// ړȌԂoB
	if (current)
		FolderStateSave(current, m_view, &m_streamViewState);

	// ړ̏ԂǂݍށB
	FolderState state;
	if FAILED(hr = FolderStateLoad(item, &state, &m_streamViewState))
		return hr;
	hr = view->CreateViewWindow(m_view, &state, this, &m_bounds, &hwndDefView);
	m_streamViewState = null; // ̃^C~O m_streamViewState B

	if FAILED(hr)
		return hr;

	// 쐬ɐ̂ŁAoϐXVB
	SetShellView(view);
	if (addToHistory)
		m_history.Push(item);
	m_folder = folder;
	m_defview = hwndDefView;

	// R[obNB
	InvokeViewCreated(m_view);

	m_view->UIActivate(hasFocus ? SVUIA_ACTIVATE_FOCUS : SVUIA_ACTIVATE_NOFOCUS);

	if (HWND hwndListView = ::FindWindowEx(m_defview, null, WC_LISTVIEW, null))
	{	// ListViewH ̃ACe[̏ꍇA܂ACe̎擾ĂȂ\B
		if (ListView_GetItemCount(hwndListView) > 0)
		{	// eւ̈ړ́A܂łqIԂɂBȂ΃tH[JXԂ̃ACeIB
			if (focus)
				m_view->SelectItem(focus, SVSI_FOCUSED | SVSI_SELECT | SVSI_ENSUREVISIBLE);
			else if (IsKeyPressed(VK_RETURN))
			{
				// L[{[hł̈ړ̂悤Ȃ΁Aŏ̍ڂIB
				// ܂ĂȂꍇ̂ŁASVSI_ENSUREVISIBLE ͕tȂB
				ref<IFolderView> view;
				if SUCCEEDED(m_view.copyto(&view))
					view->SelectItem(0, SVSI_FOCUSED | SVSI_SELECT);
			}
		}
	}

	// ړ̏ꍇ̓irQ[gTEh炷B
	if (current)
		PlayNavigateSound(m_hwnd);

	InvokeNavigationComplete(item);

	// tH_̓eXVĂꍇA̎NGXgB
	WIN32_FIND_DATA info;
	if (SUCCEEDED(ILInfo(item, &info)) && state.mtime == info.ftLastWriteTime)
		return S_OK;
	return S_UPDATED;
}

HRESULT XpExplorerBrowser::BrowseToRelative(const ITEMIDLIST* relative) throw()
{
	if (!relative)
		return E_INVALIDARG;
	if (const ITEMIDLIST* item = Current())
		return BrowseToAbsolute(ILPtr(ILCombine(item, relative)), true);
	else
		return E_UNEXPECTED;
}

HRESULT XpExplorerBrowser::BrowseToHistory(int mv) throw()
{
	const ITEMIDLIST* item = m_history.Get(mv);
	if (!item)
		return S_FALSE;
	HRESULT hr;
	if SUCCEEDED(hr = BrowseToAbsolute(item, false))
	{
		// ŁÃ݂tH_ƉߋÕtH_͓ł܂ȂB
		// [h͔ȂAO̕\ʒüړ̂ݍsB
		// ̍ہAirQ[gTEh͖炷B
		if (hr == S_FALSE)
			PlayNavigateSound(m_hwnd);
		m_history.Seek(mv);
	}
	return hr;
}

HRESULT XpExplorerBrowser::BrowseToParent() throw()
{
	ILPtr parent = ILCloneParent(Current());
	if (!parent)
		return S_FALSE;
	return BrowseToAbsolute(parent, true);
}

//==================================================================================================================================================================

namespace
{
	class XpEnumHistory : public Unknown<implements<IEnumShellItems> >
	{
	private:
		const XpExplorerHistory&	m_history;
		int							m_index;
		int							m_dir;

	public:
		XpEnumHistory(const XpExplorerHistory& history, int dir) : m_history(history), m_dir(dir)
		{
			m_index = m_dir;
		}

		HRESULT STDMETHODCALLTYPE Next(ULONG celt, IShellItem** rgelt, ULONG *pceltFetched)
		{
			ULONG fetched = 0;
			for (ULONG i = 0; i < celt; ++i)
			{
				if (const ITEMIDLIST* item = m_history.Get(m_index))
				{
					if SUCCEEDED(ItemCreate(&rgelt[fetched], item))
						fetched++;
					m_index += m_dir;
				}
			}
			if (pceltFetched)
				*pceltFetched = fetched;
			return (fetched == celt) ? S_OK : S_FALSE;
		}
        
        HRESULT STDMETHODCALLTYPE Skip(ULONG celt)
		{
			m_index += celt * m_dir;
			return S_OK;
		}
        
        HRESULT STDMETHODCALLTYPE Reset()
		{
			m_index = m_dir;
			return S_OK;
		}
        
        HRESULT STDMETHODCALLTYPE Clone(IEnumShellItems** pp)
		{
			XpEnumHistory* p = new XpEnumHistory(m_history, m_dir);
			p->m_index = m_index;
			*pp = p;
			return S_OK;
		}

	private:
		XpEnumHistory& operator = (const XpEnumHistory&);
	};
}

//==================================================================================================================================================================

HRESULT XpExplorerBrowser::GetHistory(IShellItem** pp, int mv)
{
	if (const ITEMIDLIST* item = m_history.Get(mv))
		return ItemCreate(pp, item);
	return E_FAIL;
}

HRESULT XpExplorerBrowser::EnumHistory(IEnumShellItems** pp, int dir)
{
	*pp = new XpEnumHistory(m_history, dir);
	return S_OK;
}

//==================================================================================================================================================================

HRESULT brownie::XpExplorerBrowserCreate(REFINTF pp)
{
	return newobj<XpExplorerBrowser>(pp);
}
