// autoptr.hpp
#pragma once

#include "brownie.hpp"

namespace brownie
{
	struct CrtMem
	{
		static void*	Alloc(size_t sz)			{ return malloc(sz); }
		static void*	Realloc(void* p, size_t sz)	{ return realloc(p, sz); }
		static void		Free(void* p)				{ free(p); }
	};

	struct CoTaskMem
	{
		static void*	Alloc(size_t sz)			{ return ::CoTaskMemAlloc(sz); }
		static void*	Realloc(void* p, size_t sz)	{ return ::CoTaskMemRealloc(p, sz); }
		static void		Free(void* p)				{ ::CoTaskMemFree(p); }
	};

	template < typename T, typename Mem = CrtMem >
	class AutoPtr
	{
	private:
		T*	m_ptr;

	public:
		AutoPtr(T* ptr = null) : m_ptr(ptr)	{}
		explicit AutoPtr(size_t sz) : m_ptr((T*)Mem::Alloc(sz))	{}
		~AutoPtr()							{ Mem::Free(m_ptr); }
		operator const T* () const throw()	{ return m_ptr; }
		operator       T* ()       throw()	{ return m_ptr; }
		T* operator -> () const throw()		{ ASSERT(m_ptr); return m_ptr; }
		T& operator * () const throw()		{ ASSERT(m_ptr); return *m_ptr; }
		T** operator & () throw()			{ ASSERT(!m_ptr); return &m_ptr; }

		T* get_ptr() const throw()		{ return m_ptr; }
		__declspec(property(get=get_ptr)) T* ptr;

		void clear() throw()
		{
			Mem::Free(m_ptr);
			m_ptr = null;
		}

		void attach(T* ptr) throw()
		{
			if (m_ptr != ptr)
			{
				Mem::Free(m_ptr);
				m_ptr = ptr;
			}
		}

		T* detach() throw()
		{
			T* tmp = m_ptr;
			m_ptr = null;
			return tmp;
		}

		void realloc(size_t sz) throw()
		{
			m_ptr = (T*)Mem::Realloc(m_ptr, sz);
		}

	private:
		AutoPtr(const AutoPtr&);
		AutoPtr& operator = (const AutoPtr&);
	};
}
