// buffer.hpp
#pragma once

#include "stream.hpp"
#include "unknown.hpp"

namespace brownie
{
	/// BufferT.
	template < class T >
	class BufferT
	{
	private:
		T*		m_buffer;
		size_t	m_length;
		size_t	m_capacity;

	public:
		BufferT() throw() : m_buffer(null), m_length(0), m_capacity(0) {}
		BufferT(const BufferT& rhs) throw() : m_buffer(null), m_length(0), m_capacity(0)
		{
			append(rhs.data(), rhs.size());
		}
		BufferT& operator = (const BufferT& rhs) throw()
		{
			clear();
			append(rhs, rhs.size());
			return *this;
		}
		~BufferT()
		{
			free(m_buffer);
		}
	public:
		T* data() throw()				{ return m_buffer; }
		const T* data() const throw()	{ return m_buffer; }
		void push_back(T c) throw()
		{
			reserve(m_length + 1);
			m_buffer[m_length] = c;
			++m_length;
		}
		void append(T c) throw()
		{
			push_back(c);
		}
		void append(const T* data, size_t ln) throw()
		{
			write(m_length, data, ln);
		}
		size_t read(size_t pos, T* data, size_t ln)
		{
			if (pos >= m_length)
				return 0;
			if (pos + ln >= m_length)
				ln = m_length - pos;
			memcpy(data, m_buffer + pos, ln * sizeof(T));
			return ln;
		}
		void write(size_t pos, const T* data, size_t ln)
		{
			if (ln == 0)
				return;
			if (m_length < pos + ln)
				resize(pos + ln);
			memcpy(m_buffer + pos, data, ln * sizeof(T));
		}
		bool empty() const throw()
		{
			return m_length == 0;
		}
		size_t size() const throw()
		{
			return m_length;
		}
		void resize(size_t sz) throw()
		{
			reserve(sz);
			m_length = sz;
		}
		void reserve(size_t capacity) throw()
		{
			if (m_capacity >= capacity)
				return;
			m_capacity *= 2;
			if (m_capacity < capacity)
				m_capacity = capacity;
			if (m_capacity < 16)
				m_capacity = 16;
			m_buffer = (T*)realloc(m_buffer, m_capacity * sizeof(T));
		}
		void clear() throw()
		{
			m_length = 0;
		}
	};

	/// StringBufferT.
	template < class T >
	class StringBufferT : public BufferT<T>
	{
		typedef BufferT<T> super;
	public:
		const T* str() const	{ return empty() ? 0 : data(); }
		using super::append;
		/// [null͒ǉȂ.
		void append(const T* str)
		{
			size_t length = len(str);
			size_t sz = size();
			resize(sz + length);
			for (size_t i = 0; i < length; ++i)
				data()[sz+i] = str[i];
		}
		void append_path(const T* path)
		{
			if (str::find(path, T(' ')))
			{	// Xy[X܂ޏꍇ
				append(T('"'));
				append(path);
				append(T('"'));
			}
			else
			{	// Xy[X܂܂Ȃꍇ
				append(path);
			}
		}
		void append_format(PCWSTR format, ...)
		{
			WCHAR buf[1024];
			va_list va;
			va_start(va, format);
			wvsprintf(buf, format, va);
			va_end(va);
			append(buf);
		}
		const T* terminate()
		{
			push_back(T('\0'));
			return str();
		}
	};
	typedef StringBufferT<WCHAR>	StringBuffer;

	//==================================================================================================================================================================

	class BufferStream : public Unknown< implements<IStream, ISequentialStream>, mixin<StreamImpl, DynamicLife> >
	{
	private:
		BufferT<BYTE>&	m_buffer;
		size_t			m_pos;

	private: // non-copyable
		BufferStream(const BufferStream&);
		BufferStream& operator = (const BufferStream&);

	public:
		BufferStream(BufferT<BYTE>& buffer);
		HRESULT __stdcall Read(void* data, ULONG size, ULONG* done);
		HRESULT __stdcall Write(const void* data, ULONG size, ULONG* done);
		HRESULT __stdcall SetSize(ULARGE_INTEGER sz);
		virtual DWORD	GetAccessMode() const;
		virtual UINT64	GetSize() const;
		virtual UINT64	GetPosition() const;
		virtual HRESULT	SeekTo(UINT64 pos);
	};
}

//==================================================================================================================================================================

template < typename T >
inline IStream& operator >> (IStream& stream, brownie::BufferT<T>& v) throw(HRESULT)
{
	UINT32 size;
	stream >> size;
	v.resize(size);
	StreamRead(&stream, v.data(), size * sizeof(T));
	return stream;
}

template < typename T >
inline IStream& operator << (IStream& stream, const brownie::BufferT<T>& v) throw(HRESULT)
{
	stream << (UINT32)v.size();
	StreamWrite(&stream, v.data(), v.size() * sizeof(T));
	return stream;
}

//==================================================================================================================================================================

template < typename T, typename U >
brownie::StringBufferT<T>& operator << (brownie::StringBufferT<T>& stream, const U& o)
{
	stream.append(o);
	return stream;
}
