// py.hpp
#pragma once

#include "ref.hpp"
#include "preprocessor.hpp"

#if PY_VERSION_HEX >= 0x03000000
#	define PyInt_Type		PyLong_Type
#else
#	define PyVarObject_HEAD_INIT(type, size)	PyObject_HEAD_INIT(type) (size),
#endif

namespace brownie
{
	//==================================================================================================================================================================

	object	getattr(PyObject* self, PCSTR name) throw();
	void	setattr(PyObject* self, PCSTR name, PyObject* value) throw(...);
	void	delattr(PyObject* self, PCSTR name) throw();

	object	getitem(PyObject* self, ssize_t index) throw(...);
	void	setitem(PyObject* self, ssize_t index, PyObject* value) throw(...);

	object	getitem(Array* self, ssize_t index) throw();

	object	getitem(Tuple* self, ssize_t index) throw();
	void	setitem(Tuple* self, ssize_t index, PyObject* value) throw();

	object	getitem(List* self, ssize_t index) throw();
	void	setitem(List* self, ssize_t index, PyObject* value) throw();

	object	getitem(Dict* self, PCSTR name) throw();
	void	setitem(Dict* self, PCSTR name, PyObject* value) throw(...);
	void	detitem(Dict* self, PCSTR name) throw();

	//==================================================================================================================================================================
	// Boolean
	
	template <> class ValueT<bool> : public PyObject
	{
	public:
		static boolean	newobj(bool value) throw()	{ return boolean(static_cast<Boolean*>(value ? Py_True : Py_False)); }
		static bool		value(PyObject* o) throw(...);
	};

	//==================================================================================================================================================================
	// Integer
	
	template <> class ValueT<int> : public PyObject
	{
	public:
		static integer	newobj(int value) throw();
		static int		value(Integer* o) throw();
		static int		value(PyObject* o) throw(...);
	};

	//==================================================================================================================================================================
	// Numeric

	template <> class ValueT<__int64> : public PyObject
	{
	public:
		static numeric	newobj(__int64 value) throw();
		static __int64	value(Numeric* o) throw()		{ return o ? PyLong_AsLongLong(o) : 0; }
		static __int64	value(PyObject* o) throw(...);
	};

	//==================================================================================================================================================================
	// Real
	
	template <> class ValueT<double> : public PyObject
	{
	public:
		static double	value(Real* o) throw()			{ return o ? PyFloat_AS_DOUBLE(o) : 0; }
		static double	value(PyObject* o) throw(...);
	};

	//==================================================================================================================================================================
	// String

	template <> class ValueT<CHAR> : public PyObject
	{
	public:
		static string	newobj(PCSTR value) throw();
		static string	newobj(PCWSTR value) throw();
		static string	newobj(PCSTR value, ssize_t size) throw();
		static string	newobj(PSTR* value, ssize_t size) throw();
		static string	newobj(PyObject* o) throw(...);
		static PCSTR	value(PyObject* o) throw();
		static PCSTR	value(const String* o) throw()	{ return o ? PyString_AS_STRING(o) : null; }
		PCSTR	str() const throw()						{ return value(this); }
		ssize_t	length() const throw()					{ return PyString_GET_SIZE(this); }
	};

	//==================================================================================================================================================================
	// Unicode

	template <> class ValueT<WCHAR> : public PyObject
	{
	public:
		static unicode	newobj(PCWSTR value) throw()		{ return newobj(value, (int)wcslen(value)); }
		static unicode	newobj(PCWSTR value, ssize_t size) throw();
		static unicode	newobj(PWSTR* value, ssize_t size) throw();
		static unicode	newobj(PyObject* o) throw(...);
		static unicode	newobj(PyObject* o, PCSTR encoding) throw(...);
		static unicode	newobj(Unicode* left, Unicode* right);
		static PCWSTR	value(PyObject* o) throw();
		static PCWSTR	value(const Unicode* o) throw()	{ return o ? PyUnicode_AS_UNICODE(o) : null; }
		PCWSTR	str() const throw()						{ return value(this); }
		ssize_t	length() const throw()					{ return PyUnicode_GET_SIZE(this); }
		ssize_t	find(Unicode* substr, ssize_t start = 0, ssize_t end = -1, int direction = 1);
		ssize_t	count(Unicode* substr, ssize_t start = 0, ssize_t end = -1);
		unicode	replace(Unicode* from, Unicode* to, ssize_t count = -1);

		// PyObject* PyUnicode_Split( PyObject *s, PyObject *sep, Py_ssize_t maxsplit) 
		// PyObject* PyUnicode_Splitlines( PyObject *s, int keepend) 
		// PyObject* PyUnicode_Translate( PyObject *str, PyObject *table, const char *errors) 
		// PyObject* PyUnicode_Join( PyObject *separator, PyObject *seq) 
		//int PyUnicode_Compare( PyObject *left, PyObject *right) 
		//int PyUnicode_RichCompare( PyObject *left, PyObject *right, int op) 
		//PyObject* PyUnicode_Format( PyObject *format, PyObject *args) 
		//int PyUnicode_Contains( PyObject *container, PyObject *element) 
	};

	//==================================================================================================================================================================
	// Iterator

	class Iterator : public PyObject
	{
	public:
		static iterator newobj(PyObject* o) throw(...);
		static iterator newobj_nothrow(PyObject* o) throw();
		object	next() throw(...);
	};

#define foreach_iter(ITEM, LIST, ITER)								\
	if (iterator ITER = brownie::Iterator::newobj_nothrow(LIST))	\
		while (ITEM = (ITER)->next())

#define foreach(ITEM, LIST)		foreach_iter(ITEM, LIST, PP_ID(__iter__))

	//==================================================================================================================================================================
	// Sequence

	class Sequence : public PyObject
	{
	public:
		ssize_t	length() throw()				{ return PySequence_Size(this); }
		bool	contains(PyObject* o) throw()	{ return PySequence_Contains(this, o) > 0; }
		ssize_t	index(PyObject* o) throw()		{ return PySequence_Index(this, o); }
	};

	//==================================================================================================================================================================
	// Array

	class Array : public Sequence
	{
	public:
		static array		newobj(PyObject* o) throw(...);
		static PyObject**	value(Array* o) throw()	{ return o ? PySequence_Fast_ITEMS(o) : null; }
	};

	//==================================================================================================================================================================
	// Tuple

	class Tuple : public Array
	{
	public:
		static tuple newobj(PyObject* o) throw(...);
		static tuple newobj(Tuple* o) throw()					{ return tuple(o); }
		static tuple newobj(List* o) throw();
		static tuple newobj(PCSTR format, va_list va) throw(...);
		static tuple newobj(ssize_t n) throw()					{ return tuple::from(PyTuple_New(n)); }
		static tuple newobj(POINT pt) throw();
		static tuple newobj(SIZE sz) throw()					{ return newobj((POINT&)sz); }
		ssize_t	length() const throw()							{ return PyTuple_GET_SIZE(this); }
		tuple	slice(ssize_t low, ssize_t high) throw()		{ return tuple::from(PyTuple_GetSlice(this, low, high)); }
		void	attach(ssize_t index, PyObject* value) throw()	{ PyTuple_SET_ITEM(this, index, value); }
		PyObject*	get_item(ssize_t index) throw()				{ return PyTuple_GET_ITEM(this, index); }
		__declspec(property(get=get_item)) PyObject* item[];
	};

	//==================================================================================================================================================================
	// List

	class List : public Array
	{
	public:
		static list newobj(PyObject* o) throw(...);
		static list newobj(List* o) throw()				{ return list(o); }
		static list newobj(ssize_t n) throw()			{ return list::from(PyList_New(n)); }
		ssize_t	length() throw()						{ return PyList_GET_SIZE(this); }
		bool	insert(int index, PyObject* o) throw()	{ return PyList_Insert(this, index, o) == 0; }
		bool	append(PyObject* o) throw()				{ return PyList_Append(this, o) == 0; }
		PyObject*	get_item(ssize_t index) throw()		{ return PyList_GET_ITEM(this, index); }
		__declspec(property(get=get_item)) PyObject* item[];
	};

	//==================================================================================================================================================================
	// Mapping

	class Mapping : public PyObject
	{
	public:
		list	items() throw()							{ return list::from(PyMapping_Items(this)); }
	};

	//==================================================================================================================================================================
	// Dict

	class Dict : public Mapping
	{
	public:
		static dict newobj() throw()						{ return dict::from(PyDict_New()); }
		ssize_t	length() throw()							{ return PyDict_Size(this); }
		list	items() throw()								{ return list::from(PyDict_Items(this)); }
		list	keys() throw()								{ return list::from(PyDict_Keys(this)); }
		list	values() throw()							{ return list::from(PyDict_Values(this)); }
		object	get_item(PCSTR name) throw()				{ return getitem(this, name); }
		void	set_item(PCSTR name, PyObject* o) throw()	{ setitem(this, name, o); }
		__declspec(property(get=get_item, put=set_item)) object item[];
	};
	
	//==================================================================================================================================================================
	// Set

	class Set : public PyObject
	{
	public:
		ssize_t	length() throw()				{ return PySet_GET_SIZE(this); }
		bool	contains(PyObject* key) throw()	{ return PySet_Contains(this, key) > 0; }
	};

	class FrozenSet : public Set
	{
	public:
		static ref<FrozenSet> newobj(PyObject* iterable)	{ return ref<FrozenSet>::from(PyFrozenSet_New(iterable)); }
	};

	class MutableSet : public Set
	{
	public:
		static ref<MutableSet> newobj(PyObject* iterable)	{ return ref<MutableSet>::from(PySet_New(iterable)); }
		void	add(PyObject* key) throw(...)				{ PySet_Add(this, key); /* 0 if ok */ }
		void	remove(PyObject* key) throw(...)			{ PySet_Discard(this, key); /* 1 if found and removed, 0 if not found  */ }
		object	pop() throw()								{ return object::from(PySet_Pop(this)); }
		void	clear() throw()								{ PySet_Clear(this); }
	};

	//==================================================================================================================================================================

	class File : public PyObject
	{
	public:
		static object		newobj(IStream* stream, bool text);

		void	close();
		object	read(ssize_t size = -1);
		object	readline(ssize_t size = 0);
		void	write(PyObject* obj, int flags = 0);
		void	write(PCSTR s);
		void	seek(ssize_t offset, int whence = SEEK_SET);
		int		tell();
		void	truncate(int size = -1);

		//closed;
		//encoding;
		//mode;
		//newlines;
		//softspace;

		__declspec(property(get=get_handle))	FILE* handle;
		__declspec(property(get=get_name))		String* name;
	};

	//==================================================================================================================================================================

	class Module : public PyObject
	{
	public:
		static ref<Module> newobj(PCSTR name, PCSTR doc, const PyMethodDef methods[] = null) throw();
		static ref<Module> import(PCSTR name);
		void def(PCSTR name, int value) throw();
		void def(PCSTR name, PCSTR value) throw();
		void def(PCSTR name, PCWSTR value) throw();
		void def(PCSTR name, PyObject* value) throw();
		void def(PyTypeObject* type) throw();
		template < typename T > void def() throw()	{ def(typeof<T>()); }
	};

	//==========================================================================================================================
	// typeof

	template < typename T > inline PyTypeObject* typeof() throw()	{ return detail::ObjectImpl<T>::__type__(); }
	template <> inline PyTypeObject* typeof<void   >() throw()		{ return null;            }
	template <> inline PyTypeObject* typeof<Integer>() throw()		{ return &PyInt_Type;     }
	template <> inline PyTypeObject* typeof<Numeric>() throw()		{ return &PyLong_Type;    }
	template <> inline PyTypeObject* typeof<Real   >() throw()		{ return &PyFloat_Type;   }
	template <> inline PyTypeObject* typeof<String >() throw()		{ return &PyString_Type;  }
	template <> inline PyTypeObject* typeof<Unicode>() throw()		{ return &PyUnicode_Type; }
	template <> inline PyTypeObject* typeof<Tuple  >() throw()		{ return &PyTuple_Type;   }
	template <> inline PyTypeObject* typeof<List   >() throw()		{ return &PyList_Type;    }
	template <> inline PyTypeObject* typeof<Dict   >() throw()		{ return &PyDict_Type;    }

	//==========================================================================================================================
	// is

	template < typename T > inline bool is(PyObject* o) throw()
	{
		return o && PyObject_TypeCheck(o, typeof<T>()) == 1;
	}
	template <> inline bool is<Null    >(PyObject* o) throw()	{ return !o || o == Py_None; }
	template <> inline bool is<Iterator>(PyObject* o) throw()	{ return o && PyIter_Check(o) == 1; }
	template <> inline bool is<Sequence>(PyObject* o) throw()	{ return o && PySequence_Check(o) == 1; }
	template <> inline bool is<Array   >(PyObject* o) throw()	{ return is<Tuple>(o) || is<List>(o); }
	template <> inline bool is<Mapping >(PyObject* o) throw()	{ return o && PyMapping_Check(o) == 1; }

	inline bool is_callable(PyObject* o) throw()	{ return o && PyCallable_Check(o) == 1; }
	inline bool is_iterable(PyObject* o) throw()	{ return is<Iterator>(o) || is<Sequence>(o); }

	//==========================================================================================================================
	// ln

	inline ssize_t len(PyObject* o) throw()	{ return is<Null>(o) ? 0 : PyObject_Size(o); }
	inline ssize_t len(Sequence* o) throw()	{ return !o ? 0 : PySequence_Size(o); }
	inline ssize_t len(Array*    o) throw()	{ return !o ? 0 : PySequence_Fast_GET_SIZE(o); }
	inline ssize_t len(Tuple*    o) throw()	{ return !o ? 0 : PyTuple_GET_SIZE(o); }
	inline ssize_t len(List*     o) throw()	{ return !o ? 0 : PyList_GET_SIZE(o); }
	inline ssize_t len(Dict*     o) throw()	{ return !o ? 0 : PyMapping_Size(o); }
	inline ssize_t len(String*   o) throw()	{ return !o ? 0 : PyString_GET_SIZE(o); }
	inline ssize_t len(Unicode*  o) throw()	{ return !o ? 0 : PyUnicode_GET_SIZE(o); }

	//==========================================================================================================================
	// str

	inline PCSTR	str(String* o)	{ return o->str(); }
	inline PCWSTR	str(Unicode* o)	{ return o->str(); }

	//==========================================================================================================================
	// apply

	object	apply(PyObject* fn, Tuple* args = null, Dict* kw = null) throw(...);
	object	apply(PyObject* fn, PCSTR format, ...) throw(...);

	//==================================================================================================================================================================
	// parse & build

	void	parse(Tuple* args, PCSTR format, ...) throw(...);
	void	parse(Tuple* args, const PCSTR formats[], ...) throw(...);
	void	parse(Tuple* args, Dict* kw, PCSTR format, const PCSTR kwds[], ...) throw(...);
	object	build(PCSTR format, ...) throw(...);

	//==================================================================================================================================================================
	// etc

	void	abstract(PyTypeObject* type, PyObject* o) throw(...);

	//==================================================================================================================================================================

	namespace py
	{
		void		Enter() throw();
		void		Leave() throw();
		void		CollectGarbage() throw();
		bool		HasLock() throw();
		PyObject*	GenericNew(PyTypeObject* type) throw();
		void		GenericDelete(PyObject* self) throw();
		int			GenericClear(PyObject* self) throw();
		int			GenericTraverse(PyObject* self, visitproc visit, void* arg) throw();
		int			GenericTypeFlags(const PyGetSetDef members[], PyTypeObject* base) throw();
		reprfunc	GenericStr(const PyMethodDef methods[]) throw();
		void		AddRef(PyObject* o) throw();
		void		Release(PyObject* o) throw();

		class Lock
		{
		public:
			Lock() throw()	{ Enter(); }
			~Lock() throw()	{ Leave(); }
		};

		class Unlock
		{
		public:
			Unlock() throw()	{ Leave(); }
			~Unlock() throw()	{ Enter(); }
		};
	}
}

//#ifndef _DEBUG

inline void brownie::py::AddRef(PyObject* o) throw()
{
	ASSERT(!o || HasLock());
	Py_XINCREF(o);
}

inline void brownie::py::Release(PyObject* o) throw()
{
	ASSERT(!o || HasLock());
	Py_XDECREF(o);
}

//#endif
