// ref.hpp
#pragma once

#include "brownie.hpp"

namespace brownie
{
	class Object;

	//========================================================================================================================
	// ^tꂽC^tF[X|C^̎Q.
	struct REFINTF
	{
		REFIID iid;
		void** const pp;
		template < class T > REFINTF(T** ppT)              : iid(__uuidof(T)), pp((void**)ppT)	{ ASSERT(pp); }
		template < class T > REFINTF(REFIID riid, T** ppT) : iid(riid), pp((void**)ppT)			{ ASSERT(pp); }
		                     REFINTF(const REFINTF& r)     : iid(r.iid), pp(r.pp)				{ ASSERT(pp); }

	private:
		REFINTF& operator = (const REFINTF&);
	};

	//========================================================================================================================

	HRESULT	CreateInstance(REFCLSID clsid, DWORD clsctx, REFINTF pp);
	bool	same(IUnknown* lhs, IUnknown* rhs) throw();

	//========================================================================================================================
	// ref<T>

	template < typename T >
	inline void incref(T* p)
	{
		__if_exists(T::AddRef)		{ if (p) p->AddRef(); }
		__if_not_exists(T::AddRef)	{ py::AddRef(p); }
	}

	template < typename T >
	inline void decref(T* p)
	{
		__if_exists(T::Release)		{ if (p) p->Release(); }
		__if_not_exists(T::Release)	{ py::Release(p); }
	}

	template < typename T >
	class ref
	{
	public:
		typedef T*	pointer;

	private:
		pointer	m_ptr;

		static T* extract(T* p) throw()			{ return p; }
		static T* extract(Object* p) throw()	{ return p->self; }

	public:
		                        ref()                throw() : m_ptr(null)			{}
		                        ref(const ref& p)    throw() : m_ptr(p)				{ incref(m_ptr); }
		template < typename U > ref(const ref<U>& p) throw() : m_ptr(extract(p))	{ incref(m_ptr); }
		explicit                ref(T* p)            throw() : m_ptr(p)				{ incref(m_ptr); }
		                        ref(const Null&)     throw() : m_ptr(null)			{}

		~ref() throw()	{ decref(m_ptr); }

		                        ref& operator = (const ref& p) throw()		{ assign(p); return *this; }
		template < typename U > ref& operator = (const ref<U>& p) throw()	{ assign(extract(p)); return *this; }
		                        ref& operator = (T* p) throw()				{ assign(p); return *this; }
		                        ref& operator = (Object* p) throw()			{ assign(extract(p)); return *this; }
		                        ref& operator = (const Null&) throw()		{ decref(m_ptr); m_ptr = null; return *this; }

		void assign(T* p) throw()
		{
			incref(p);
			attach(p);
		}

		void attach(T* p) throw()
		{
			decref(m_ptr);
			m_ptr = p;
		}

		T* detach() throw()
		{
			T* tmp = m_ptr;
			m_ptr = null;
			return tmp;
		}

		template < typename U >
		static ref from(U* o) throw()
		{
			ref r;
			r.m_ptr = static_cast<T*>(o);
			return r;
		}

		HRESULT copyto(REFIID iid, void** pp) throw()
		{
			if (m_ptr)
				return m_ptr->QueryInterface(iid, pp);
			*pp = null;
			return E_POINTER;
		}

		HRESULT copyto(REFINTF pp) throw()
		{
			return copyto(pp.iid, pp.pp);
		}

		HRESULT newobj(REFCLSID clsid, DWORD clsctx = CLSCTX_INPROC) throw()
		{
			ASSERT(!m_ptr);
			return CreateInstance(clsid, clsctx, &m_ptr);
		}

		operator T* () const throw()	{ return m_ptr; }
		T* operator -> () const throw()	{ ASSERT(m_ptr); return m_ptr; }
		T& operator * () const throw()	{ ASSERT(m_ptr); return *m_ptr; }
		T** operator & () throw()		{ ASSERT(!m_ptr); return &m_ptr; }

		T* get_ptr() const throw()		{ return m_ptr; }
		__declspec(property(get=get_ptr)) T* ptr;

		object operator [] (ssize_t index) throw()
		{
			ASSERT(m_ptr);
			return getitem(m_ptr, index);
		}
	};

	template < typename T, typename U >
	inline bool operator == (const ref<T>& lhs, const ref<U>& rhs) throw()	{ return same(lhs, rhs); }
	template < typename T, typename U >
	inline bool operator == (T* lhs, const ref<U>& rhs) throw()				{ return same(lhs, rhs); }
	template < typename T, typename U >
	inline bool operator == (const ref<T>& lhs, U* rhs) throw()				{ return same(lhs, rhs); }

	template < typename T, typename U >
	inline brownie::ref<T>& operator >> (brownie::ref<T>& r, U& v)			{ *r >> v; return r; }
	template < typename T, typename U >
	inline brownie::ref<T>& operator << (brownie::ref<T>& r, const U& v)	{ *r << v; return r; }

	//========================================================================================================================
	// X}[g|C^̔z.

	template
	<
		class T,
		template < class T, class TAlloc > class TSequence = std::vector,
		template < class T > class TAlloc = std::allocator
	>
	class refs : public TSequence< ref<T>, TAlloc< ref<T> > >
	{
	public: // algorithm extension
		template < class A1 >
		void each(HRESULT (__stdcall T::*m)(A1 a1), A1 a1) const
		{
			for (const_iterator i = begin(); i != end(); ++i)
				if (*i)
					((*i)->*m)(a1);
		}
	};
}
