// stream.hpp
#pragma once

#include "ref.hpp"
#include <utility>

//==================================================================================================================================================================
// Stream

namespace brownie
{
	ref<IStream>	StreamCreate(HGLOBAL* mem) throw(HRESULT);
	ref<IStream>	StreamCreate(PCWSTR filename, DWORD flags) throw(HRESULT);
	ref<IStream>	StreamCreate(PCSTR filename, DWORD flags) throw(HRESULT);
	ref<IStream>	StreamCreate(PyObject* o) throw(HRESULT);
	void			StreamRead(IStream* stream, void* data, size_t size) throw(HRESULT);
	void			StreamWrite(IStream* stream, const void* data, size_t size) throw(HRESULT);
	UINT64			StreamSeek(IStream* stream, INT64 off, DWORD whence = SEEK_SET);

	template < typename T >
	inline void StreamReadT(IStream* stream, T data[], size_t ln) throw(HRESULT)
	{
		StreamRead(stream, data, ln * sizeof(T));
	}

	template < typename T >
	inline void StreamWriteT(IStream* stream, const T data[], size_t ln) throw(HRESULT)
	{
		StreamWrite(stream, data, ln * sizeof(T));
	}
}

//==================================================================================================================================================================
// Stream : POD

#define BROWNIE_POD(type)													\
inline IStream& operator >> (IStream& stream, type& v) throw(HRESULT)		\
{ brownie::StreamRead(&stream, &v, sizeof(v)); return stream; }				\
inline IStream& operator << (IStream& stream, const type& v) throw(HRESULT)	\
{ brownie::StreamWrite(&stream, &v, sizeof(v)); return stream; }

BROWNIE_POD(__int8)
BROWNIE_POD(__int16)
BROWNIE_POD(__int32)
BROWNIE_POD(__int64)
BROWNIE_POD(unsigned __int8)
BROWNIE_POD(unsigned __int16)
BROWNIE_POD(unsigned __int32)
BROWNIE_POD(unsigned __int64)
BROWNIE_POD(long)
BROWNIE_POD(unsigned long)
BROWNIE_POD(POINT)
BROWNIE_POD(SIZE)
BROWNIE_POD(RECT)
BROWNIE_POD(GUID)

//==================================================================================================================================================================
// Stream : string

IStream& operator >> (IStream& stream, PCWSTR& v) throw(HRESULT);
IStream& operator << (IStream& stream, PCWSTR v) throw(HRESULT);

//==================================================================================================================================================================
// Stream : std::pair<First, Second>

template < typename First, typename Second >
inline IStream& operator >> (IStream& stream, std::pair<First, Second>& v) throw(HRESULT)
{
	return stream >> v.first >> v.second;
}

template < typename First, typename Second >
inline IStream& operator << (IStream& stream, const std::pair<First, Second>& v) throw(HRESULT)
{
	return stream << v.first << v.second;
}

//==================================================================================================================================================================
// StreamImpl

namespace brownie
{
	template < class TBase >
	class __declspec(novtable) StreamImpl : public TBase
	{
	public:
		HRESULT __stdcall Read(void* data, ULONG size, ULONG* done)			{ return E_NOTIMPL; }
		HRESULT __stdcall Write(const void* data, ULONG size, ULONG* done)	{ return E_NOTIMPL; }

		HRESULT __stdcall Seek(LARGE_INTEGER where, DWORD whence, ULARGE_INTEGER* pos)
		{
			UINT64 p;
			switch (whence)
			{
			case STREAM_SEEK_SET:
				p = where.QuadPart;
				break;
			case STREAM_SEEK_CUR:
				p = GetPosition() + where.QuadPart;
				break;
			case STREAM_SEEK_END:
				p = GetSize() + where.QuadPart;
				break;
			default:
				return E_INVALIDARG;
			}
			HRESULT hr = SeekTo(p);
			if (SUCCEEDED(hr) && pos)
				pos->QuadPart = p;
			return hr;
		}
		HRESULT __stdcall Stat(STATSTG* stat, DWORD grfStatFlag)
		{
			if (!stat)
				return E_INVALIDARG;
			switch (grfStatFlag)
			{
			case STATFLAG_DEFAULT:
				if (stat->pwcsName)
					stat->pwcsName[0] = L'\0';
				break;
			default:
				stat->pwcsName = null;
				break;
			}
			FILETIME time;
			CoFileTimeNow(&time);
			stat->type = STGTY_STREAM;
			stat->cbSize.QuadPart = GetSize();
			stat->grfMode = GetAccessMode();
			stat->grfLocksSupported = 0;
			stat->clsid = GUID_NULL;
			stat->mtime = stat->ctime = stat->atime = time;
			return S_OK;
		}

		HRESULT __stdcall CopyTo(IStream* pStream, ULARGE_INTEGER size, ULARGE_INTEGER* read, ULARGE_INTEGER* written)	{ return E_NOTIMPL; }
		HRESULT __stdcall Commit(DWORD grfCommitFlags)			{ return E_NOTIMPL; }
		HRESULT __stdcall Revert()								{ return E_NOTIMPL; }
		HRESULT __stdcall Clone(IStream** ppClone)				{ return E_NOTIMPL; }
		HRESULT __stdcall SetSize(ULARGE_INTEGER sz)			{ return E_NOTIMPL; }
		HRESULT __stdcall LockRegion(ULARGE_INTEGER where, ULARGE_INTEGER size, DWORD dwLockType)	{ return E_NOTIMPL; }
		HRESULT __stdcall UnlockRegion(ULARGE_INTEGER where, ULARGE_INTEGER size, DWORD dwLockType)	{ return E_NOTIMPL; }

	protected:
		virtual DWORD	GetAccessMode() const = 0;
		virtual UINT64	GetSize() const = 0;
		virtual UINT64	GetPosition() const = 0;
		virtual HRESULT	SeekTo(UINT64 pos) = 0;
	};
}
