// struct.hpp
#pragma once

#include "config.hpp"

//==================================================================================================================================================================
// POINT

#define POINT_XY(o)			(o).x, (o).y

inline bool operator == (const POINT& lhs, const POINT& rhs) throw()	{ return memcmp(&lhs, &rhs, sizeof(POINT)) == 0; }
inline bool operator != (const POINT& lhs, const POINT& rhs) throw()	{ return !(lhs == rhs); }

inline POINT operator + (const POINT& lhs, const POINT& rhs)
{
	POINT result = { lhs.x + rhs.x, lhs.y + rhs.y };
	return result;
}

inline POINT operator - (const POINT& lhs, const POINT& rhs)
{
	POINT result = { lhs.x - rhs.x, lhs.y - rhs.y };
	return result;
}

inline POINT POINT_FROM_LPARAM(LPARAM lParam)	{ POINT o = { GET_XY_LPARAM(lParam) }; return o; }

//==================================================================================================================================================================
// SIZE

#define SIZE_W(o)			((o).cx)
#define SIZE_H(o)			((o).cy)
#define SIZE_WH(o)			(o).cx, (o).cy

inline bool operator == (const SIZE& lhs, const SIZE& rhs) throw()	{ return memcmp(&lhs, &rhs, sizeof(SIZE)) == 0; }
inline bool operator != (const SIZE& lhs, const SIZE& rhs) throw()	{ return !(lhs == rhs); }

inline SIZE operator + (const SIZE& lhs, const SIZE& rhs)
{
	SIZE result = { lhs.cx + rhs.cx, lhs.cy + rhs.cy };
	return result;
}

inline SIZE operator - (const SIZE& lhs, const SIZE& rhs)
{
	SIZE result = { lhs.cx - rhs.cx, lhs.cy - rhs.cy };
	return result;
}

//==================================================================================================================================================================
// RECT

#define RECT_X(o)			((o).left)
#define RECT_Y(o)			((o).top)
#define RECT_W(o)			((o).right - (o).left)
#define RECT_H(o)			((o).bottom - (o).top)
#define RECT_XY(o)			RECT_X(o), RECT_Y(o)
#define RECT_WH(o)			RECT_W(o), RECT_H(o)
#define RECT_XYWH(o)		RECT_XY(o), RECT_WH(o)
#define RECT_LTRB(o)		(o).left, (o).top, (o).right, (o).bottom
#define RECT_LOCATION(o)	((POINT&)(o))
inline POINT	RECT_NW(const RECT& o)		{ POINT pt = { (o).left, (o).top }; return pt; }
inline POINT	RECT_NE(const RECT& o)		{ POINT pt = { (o).right, (o).top }; return pt; }
inline POINT	RECT_SW(const RECT& o)		{ POINT pt = { (o).left, (o).bottom }; return pt; }
inline POINT	RECT_SE(const RECT& o)		{ POINT pt = { (o).right, (o).bottom }; return pt; }
inline SIZE		RECT_SIZE(const RECT& o)	{ SIZE sz = { RECT_WH(o) }; return sz; }
#define RECT_CENTER_W(o)	(((o).left + (o).right)/2) 
#define RECT_CENTER_H(o)	(((o).top + (o).bottom)/2)
#define RECT_CENTER(o)		RECT_CENTER_W(o), RECT_CENTER_H(o)

inline RECT RECT_FROM_LTRB(int l, int t, int r, int b)	{ RECT o = { l, t, r, b }; return o; }
inline RECT RECT_FROM_XYWH(int x, int y, int w, int h)	{ return RECT_FROM_LTRB(x, y, x+w, y+h); }

inline bool operator == (const RECT& lhs, const RECT& rhs) throw()	{ return memcmp(&lhs, &rhs, sizeof(RECT)) == 0; }
inline bool operator != (const RECT& lhs, const RECT& rhs) throw()	{ return !(lhs == rhs); }

//========================================================================================================================

template < typename Coord >
struct PointT : POINT
{
	PointT() {}
	PointT(INT xx, INT yy) { assign(xx, yy); }

	void assign(INT xx, INT yy)	{ x = xx; y = yy; }
	void offset(INT dx, INT dy)	{ x += dx; y += dy; }

	bool operator == (const PointT& rhs) const	{ return memcmp(this, &rhs, sizeof(PointT)) == 0; }
	bool operator != (const PointT& rhs) const	{ return !operator == (rhs); }

	friend PointT operator - (const PointT& lhs, const PointT& rhs)	{ return Point(lhs.x-rhs.x, lhs.y-rhs.y); }
};

//========================================================================================================================

template < typename Coord >
struct SizeT : SIZE
{
	SizeT()	{}
	SizeT(INT ww, INT hh) { assign(ww, hh); }

	INT get_w() const	{ return cx; }
	INT get_h() const	{ return cy; }

	void set_w(INT value)	{ cx = value; }
	void set_h(INT value)	{ cy = value; }

	__declspec(property(get=get_w, put=set_w)) INT w;
	__declspec(property(get=get_h, put=set_h)) INT h;

	void assign(INT ww, INT hh)		{ cx = ww; cy = hh; }
	void inflate(INT dw, INT dh)	{ cx += dw; cy += dh; }

	bool operator == (const SizeT& rhs) const	{ return memcmp(this, &rhs, sizeof(SizeT)) == 0; }
	bool operator != (const SizeT& rhs) const	{ return !operator == (rhs); }
};

//========================================================================================================================

template < typename Coord >
struct RectT : RECT
{
	typedef PointT<Coord>	Point;
	typedef SizeT<Coord>	Size;

	RectT()	{}
	RectT(INT xx, INT yy, INT ww, INT hh)	{ assign(xx, yy, ww, hh); }
	RectT(Point pt, Size sz)				{ assign(pt.x, pt.y, sz.w, sz.h); }
	explicit RectT(Size sz)					{ assign(0, 0, sz.w, sz.h); }

	INT get_x() const	{ return left; }
	INT get_y() const	{ return top; }
	INT get_w() const	{ return right-left; }
	INT get_h() const	{ return bottom-top; }

	void set_x(INT value)	{ right += value - left; left = value; }
	void set_y(INT value)	{ bottom += value - top; top = value; }
	void set_w(INT value)	{ right = left + value; }
	void set_h(INT value)	{ bottom = top + value; }

	__declspec(property(get=get_x, put=set_x)) INT x;
	__declspec(property(get=get_y, put=set_y)) INT y;
	__declspec(property(get=get_w, put=set_w)) INT w;
	__declspec(property(get=get_h, put=set_h)) INT h;

	Point get_location() const			{ return Point(x, y); }
	void set_location(const Point& p)	{ x = p.x; y = p.y; }
	__declspec(property(get=get_location, put=set_location)) Point location;

	Size get_size() const			{ return Size(w, h); }
	void set_size(const Size& s)	{ w = s.w; h = s.h; }
	__declspec(property(get=get_size, put=set_size)) Size size;

	Point get_center() const		{ return Point(x+w/2, y+h/2); }
	void set_center(const Point& p)	{ x = p.x - w/2; y = p.y - h/2; }
	__declspec(property(get=get_center, put=set_center)) Point center;

	void assign(INT xx, INT yy, INT ww, INT hh)
	{
		left = xx;
		top = yy;
		set_w(ww);
		set_h(hh);
	}

	void inflate(INT dx, INT dy)		{ ::InflateRect(this, dx, dy); }

	bool contains(const Point& p) const
	{
		return left <= p.x && top <= p.y && p.x < right && p.y < bottom;
	}
	bool overlap(INT x, INT y, INT w, INT h) const
	{
		return x < right && y < bottom && left <= x+w && top <= y+h;
	}

	bool operator == (const RectT& rhs) const	{ return memcmp(this, &rhs, sizeof(RectT)) == 0; }
	bool operator != (const RectT& rhs) const	{ return !operator == (rhs); }

	RectT& operator |= (const RectT& rhs)
	{
		left   = __min(left  , rhs.left  );
		right  = __max(right , rhs.right );
		top    = __min(top   , rhs.top   );
		bottom = __max(bottom, rhs.bottom);
		return *this;
	}
};
